@@ -11,7 +11,7 @@
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
  */
 
-static char sccsid[] = "@(#) fixstrtod.c 1.3 95/06/27 16:19:56";
+#include <stdio.h>
 
 #undef strtod
 
@@ -29,7 +29,7 @@ fixstrtod(string, endPtr)
 {
     double d;
     d = strtod(string, endPtr);
-    if ((*endPtr != string) && ((*endPtr)[-1] == 0)) {
+    if ((endPtr != NULL) && (*endPtr != string) && ((*endPtr)[-1] == 0)) {
 	*endPtr -= 1;
     }
     return d;
