@@ -8,16 +8,10 @@
  *
  * See the file "license.terms" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
- *
- * SCCS: @(#) strtod.c 1.8 96/02/15 12:08:23
  */
 
-#include "tcl.h"
-#ifdef NO_STDLIB_H
-#   include "compat/stdlib.h"
-#else
-#   include <stdlib.h>
-#endif
+#include "tclInt.h"
+#include "tclPort.h"
 #include <ctype.h>
 
 #ifndef TRUE
@@ -108,7 +102,7 @@ strtod(string, endPtr)
      */
 
     p = string;
-    while (isspace(*p)) {
+    while (isspace(UCHAR(*p))) {
 	p += 1;
     }
     if (*p == '-') {
@@ -206,7 +200,11 @@ strtod(string, endPtr)
 	    }
 	    expSign = FALSE;
 	}
-	while (isdigit(*p)) {
+	if (!isdigit(UCHAR(*p))) {
+	    p = pExp;
+	    goto done;
+	}
+	while (isdigit(UCHAR(*p))) {
 	    exp = exp * 10 + (*p - '0');
 	    p += 1;
 	}
@@ -232,6 +230,7 @@ strtod(string, endPtr)
     }
     if (exp > maxExponent) {
 	exp = maxExponent;
+	errno = ERANGE;
     }
     dblExp = 1.0;
     for (d = powersOf10; exp != 0; exp >>= 1, d += 1) {
