@@ -14,26 +14,36 @@
  *
  * See the file "license.terms" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
- *
- * RCS: @(#) $Id: tclAlloc.c,v 1.9 2000/04/21 04:04:12 hobbs Exp $
  */
 
+/*
+ * Windows and Unix use an alternative allocator when building with threads
+ * that has significantly reduced lock contention.
+ */
+
+#if !defined(TCL_THREADS) || !defined(USE_THREAD_ALLOC) || defined(TCL_MEM_DEBUG)
+
 #include "tclInt.h"
 #include "tclPort.h"
 
 #if USE_TCLALLOC
 
-#ifdef TCL_DEBUG
-#   define DEBUG
-/* #define MSTATS */
-#   define RCHECK
+/*
+ * We should really make use of AC_CHECK_TYPE(caddr_t)
+ * here, but it can wait until Tcl uses config.h properly.
+ */
+#if defined(_MSC_VER) || defined(__MINGW32__) || defined(__BORLANDC__)
+typedef unsigned long caddr_t;
 #endif
 
 /*
- * On Unix this will already be defined
+ * Alignment for allocated memory.
  */
-#if defined(WIN32) || defined(MAC_TCL)
-typedef unsigned long caddr_t;
+
+#if defined(__APPLE__)
+#define ALLOCALIGN	16
+#else
+#define ALLOCALIGN	8
 #endif
 
 /*
@@ -48,14 +58,14 @@ typedef unsigned long caddr_t;
  */
 
 union overhead {
-    union overhead *ov_next;	/* when free */
-    unsigned char ov_padding[8]; /* Ensure the structure is 8-byte aligned. */
+    union overhead *ov_next;		/* when free */
+    unsigned char ov_padding[ALLOCALIGN];/* align struct to ALLOCALIGN bytes */
     struct {
 	unsigned char	ovu_magic0;	/* magic number */
 	unsigned char	ovu_index;	/* bucket # */
 	unsigned char	ovu_unused;	/* unused */
 	unsigned char	ovu_magic1;	/* other magic number */
-#ifdef RCHECK
+#ifndef NDEBUG
 	unsigned short	ovu_rmagic;	/* range magic number */
 	unsigned long	ovu_size;	/* actual block size */
 	unsigned short  ovu_unused2;    /* padding to 8-byte align */
@@ -72,7 +82,7 @@ union overhead {
 #define MAGIC		0xef		/* magic # on accounting info */
 #define RMAGIC		0x5555		/* magic # on range info */
 
-#ifdef RCHECK
+#ifndef NDEBUG
 #define	RSLOP		sizeof (unsigned short)
 #else
 #define	RSLOP		0
@@ -82,11 +92,12 @@ union overhead {
 
 /*
  * nextf[i] is the pointer to the next free block of size 2^(i+3).  The
- * smallest allocatable block is 8 bytes.  The overhead information
+ * smallest allocatable block is MINBLOCK bytes. The overhead information
  * precedes the data area returned to the user.
  */
 
-#define NBUCKETS	13
+#define MINBLOCK	((sizeof(union overhead) + (ALLOCALIGN-1)) & ~(ALLOCALIGN-1))
+#define NBUCKETS	(13 - (MINBLOCK >> 4))
 #define MAXMALLOC	(1<<(NBUCKETS+2))
 static	union overhead *nextf[NBUCKETS];
 
@@ -131,7 +142,7 @@ static	unsigned int nmalloc[NBUCKETS+1];
 #include <stdio.h>
 #endif
 
-#if defined(DEBUG) || defined(RCHECK)
+#if !defined(NDEBUG)
 #define	ASSERT(p)   if (!(p)) panic(# p)
 #define RANGE_ASSERT(p) if (!(p)) panic(# p)
 #else
@@ -200,7 +211,7 @@ TclInitAlloc()
 void
 TclFinalizeAllocSubsystem()
 {
-    int i;
+    unsigned int i;
     struct block *blockPtr, *nextPtr;
 
     Tcl_MutexLock(allocMutexPtr);
@@ -253,7 +264,7 @@ TclpAlloc(nbytes)
     register union overhead *op;
     register long bucket;
     register unsigned amt;
-    struct block *bigBlockPtr;
+    struct block *bigBlockPtr = NULL;
 
     if (!allocInit) {
 	/*
@@ -267,9 +278,11 @@ TclpAlloc(nbytes)
     /*
      * First the simple case: we simple allocate big blocks directly
      */
-    if (nbytes + OVERHEAD >= MAXMALLOC) {
-	bigBlockPtr = (struct block *) TclpSysAlloc((unsigned) 
-		(sizeof(struct block) + OVERHEAD + nbytes), 0);
+    if (nbytes >= MAXMALLOC - OVERHEAD) {
+	if (nbytes <= UINT_MAX - OVERHEAD - sizeof(struct block)) {
+	    bigBlockPtr = (struct block *) TclpSysAlloc((unsigned) 
+		    (sizeof(struct block) + OVERHEAD + nbytes), 0);
+	}
 	if (bigBlockPtr == NULL) {
 	    Tcl_MutexUnlock(allocMutexPtr);
 	    return NULL;
@@ -285,7 +298,7 @@ TclpAlloc(nbytes)
 #ifdef MSTATS
 	nmalloc[NBUCKETS]++;
 #endif
-#ifdef RCHECK
+#ifndef NDEBUG
 	/*
 	 * Record allocated size of block and
 	 * bound space with magic numbers.
@@ -302,13 +315,10 @@ TclpAlloc(nbytes)
      * stored in hash buckets which satisfies request.
      * Account for space used per block for accounting.
      */
-#ifndef RCHECK
-    amt = 8;	/* size of first bucket */
-    bucket = 0;
-#else
-    amt = 16;	/* size of first bucket */
-    bucket = 1;
-#endif
+
+    amt = MINBLOCK;		/* size of first bucket */
+    bucket = MINBLOCK >> 4;
+
     while (nbytes + OVERHEAD > amt) {
 	amt <<= 1;
 	if (amt == 0) {
@@ -339,7 +349,7 @@ TclpAlloc(nbytes)
 #ifdef MSTATS
     nmalloc[bucket]++;
 #endif
-#ifdef RCHECK
+#ifndef NDEBUG
     /*
      * Record allocated size of block and
      * bound space with magic numbers.
@@ -551,7 +561,7 @@ TclpRealloc(cp, nbytes)
 #ifdef MSTATS
 	nmalloc[NBUCKETS]++;
 #endif
-#ifdef RCHECK
+#ifndef NDEBUG
 	/*
 	 * Record allocated size of block and update magic number bounds.
 	 */
@@ -590,7 +600,7 @@ TclpRealloc(cp, nbytes)
     /*
      * Ok, we don't have to copy, it fits as-is
      */
-#ifdef RCHECK
+#ifndef NDEBUG
     op->ov_size = (nbytes + RSLOP - 1) & ~(RSLOP - 1);
     *(unsigned short *)((caddr_t)(op + 1) + op->ov_size) = RMAGIC;
 #endif
@@ -720,3 +730,4 @@ TclpRealloc(cp, nbytes)
 }
 
 #endif /* !USE_TCLALLOC */
+#endif /* !TCL_THREADS */
