@@ -9,8 +9,6 @@
  *
  * See the file "license.terms" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
- *
- * RCS: @(#) $Id: tclBinary.c,v 1.13.2.4 2005/10/23 22:01:29 msofer Exp $
  */
 
 #include "tclInt.h"
@@ -771,6 +769,10 @@ Tcl_BinaryObjCmd(dummy, interp, objc, objv)
 	     */
 
 	    resultPtr = Tcl_GetObjResult(interp);
+	    if (Tcl_IsShared(resultPtr)) {
+		TclNewObj(resultPtr);
+		Tcl_SetObjResult(interp, resultPtr);
+	    }
 	    buffer = Tcl_SetByteArrayLength(resultPtr, length);
 	    memset((VOID *) buffer, 0, (size_t) length);
 
@@ -789,7 +791,9 @@ Tcl_BinaryObjCmd(dummy, interp, objc, objv)
 		    break;
 		}
 		if ((count == 0) && (cmd != '@')) {
-		    arg++;
+		    if (cmd != 'x') {
+			arg++;
+		    }
 		    continue;
 		}
 		switch (cmd) {
@@ -1156,7 +1160,7 @@ Tcl_BinaryObjCmd(dummy, interp, objc, objv)
 			char *dest;
 			unsigned char *src;
 			int i;
-			static char hexdigit[] = "0123456789abcdef";
+			static CONST char hexdigit[] = "0123456789abcdef";
 
 			if (arg >= objc) {
 			    DeleteScanNumberCache(numberCachePtr);
