@@ -8,8 +8,6 @@
  *
  * See the file "license.terms" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
- *
- * RCS: @(#) $Id: tclCompExpr.c,v 1.13.2.5 2008/04/17 19:47:34 dgp Exp $
  */
 
 #include "tclInt.h"
@@ -103,19 +101,19 @@ static int opTableInitialized = 0; /* 0 means not yet initialized. */
 TCL_DECLARE_MUTEX(opMutex)
 
 typedef struct OperatorDesc {
-    char *name;			/* Name of the operator. */
+    CONST char *name;			/* Name of the operator. */
     int numOperands;		/* Number of operands. 0 if the operator
 				 * requires special handling. */
     int instruction;		/* Instruction opcode for the operator.
 				 * Ignored if numOperands is 0. */
 } OperatorDesc;
 
-static OperatorDesc operatorTable[] = {
+static CONST OperatorDesc operatorTable[] = {
     {"*",   2,  INST_MULT},
     {"/",   2,  INST_DIV},
     {"%",   2,  INST_MOD},
-    {"+",   0}, 
-    {"-",   0},
+    {"+",   0,  0},
+    {"-",   0,  0},
     {"<<",  2,  INST_LSHIFT},
     {">>",  2,  INST_RSHIFT},
     {"<",   2,  INST_LT},
@@ -127,14 +125,14 @@ static OperatorDesc operatorTable[] = {
     {"&",   2,  INST_BITAND},
     {"^",   2,  INST_BITXOR},
     {"|",   2,  INST_BITOR},
-    {"&&",  0},
-    {"||",  0},
-    {"?",   0},
+    {"&&",  0,  0},
+    {"||",  0,  0},
+    {"?",   0,  0},
     {"!",   1,  INST_LNOT},
     {"~",   1,  INST_BITNOT},
     {"eq",  2,  INST_STR_EQ},
     {"ne",  2,  INST_STR_NEQ},
-    {NULL}
+    {NULL,  0,  0}
 };
 
 /*
@@ -345,7 +343,7 @@ CompileSubExpr(exprTokenPtr, infoPtr, envPtr)
     Tcl_Interp *interp = infoPtr->interp;
     Tcl_Token *tokenPtr, *endPtr = NULL; /* silence gcc 4 warning */
     Tcl_Token *afterSubexprPtr;
-    OperatorDesc *opDescPtr;
+    CONST OperatorDesc *opDescPtr;
     Tcl_HashEntry *hPtr;
     CONST char *operator;
     Tcl_DString opBuf;
@@ -389,8 +387,8 @@ CompileSubExpr(exprTokenPtr, infoPtr, envPtr)
 	    break;
 	    
         case TCL_TOKEN_BS:
-	    length = Tcl_UtfBackslash(tokenPtr->start, (int *) NULL,
-		    buffer);
+	    length = TclParseBackslash(tokenPtr->start, tokenPtr->size,
+		    (int *) NULL, buffer);
 	    if (length > 0) {
 		objIndex = TclRegisterNewLiteral(envPtr, buffer, length);
 	    } else {
