@@ -8,10 +8,14 @@
  *
  * See the file "license.terms" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
- *
- * RCS: @(#) $Id: tclFCmd.c,v 1.20.2.2 2005/08/17 17:46:36 hobbs Exp $
  */
 
+#ifndef _WIN64
+/* See [Bug 3354324]: file mtime sets wrong time */
+#   define _USE_32BIT_TIME_T
+#endif
+
+#include <sys/stat.h>
 #include "tclInt.h"
 #include "tclPort.h"
 
@@ -512,20 +516,23 @@ CopyRenameOneFile(interp, source, target, copyFlag, force)
 	    goto done;
 	}
 
-        /* 
-         * Prevent copying or renaming a file onto itself.  Under Windows, 
-         * stat always returns 0 for st_ino.  However, the Windows-specific 
-         * code knows how to deal with copying or renaming a file on top of
-         * itself.  It might be a good idea to write a stat that worked.
-         */
-     
-        if ((sourceStatBuf.st_ino != 0) && (targetStatBuf.st_ino != 0)) {
-            if ((sourceStatBuf.st_ino == targetStatBuf.st_ino) &&
-            	    (sourceStatBuf.st_dev == targetStatBuf.st_dev)) {
-            	result = TCL_OK;
-            	goto done;
-            }
-        }
+	/*
+	 * Prevent copying or renaming a file onto itself. On Windows since
+	 * 8.5 we do get an inode number, however the unsigned short field is
+	 * insufficient to accept the Win32 API file id so it is truncated to
+	 * 16 bits and we get collisions. See bug #2015723.
+	 */
+
+#if !defined(WIN32) && !defined(__CYGWIN__)
+	if ((sourceStatBuf.st_ino != 0) && (targetStatBuf.st_ino != 0)) {
+	    if ((sourceStatBuf.st_ino == targetStatBuf.st_ino) &&
+		    (sourceStatBuf.st_dev == targetStatBuf.st_dev)) {
+		result = TCL_OK;
+		goto done;
+	    }
+	}
+#endif
+
 
 	/*
 	 * Prevent copying/renaming a file onto a directory and
@@ -684,15 +691,14 @@ CopyRenameOneFile(interp, source, target, copyFlag, force)
 	     * so it should be quite clear 
 	     */
 	    errfile = target;
-	    /* 
-	     * We now need to reset the result, because the above call,
-	     * if it failed, may have put an error message in place.
-	     * (Ideally we would prefer not to pass an interpreter in
-	     * above, but the channel IO code used by
-	     * TclCrossFilesystemCopy currently requires one)
-	     */
-	    Tcl_ResetResult(interp);
 	}
+	/* 
+	 * We now need to reset the result, because the above call,
+	 * may have left set it.  (Ideally we would prefer not to pass
+	 * an interpreter in above, but the channel IO code used by
+	 * TclCrossFilesystemCopy currently requires one)
+	 */
+	Tcl_ResetResult(interp);
     }
     if ((copyFlag == 0) && (result == TCL_OK)) {
 	if (S_ISDIR(sourceStatBuf.st_mode)) {
@@ -988,6 +994,11 @@ TclFileAttrsCmd(interp, objc, objv)
 		"option", 0, &index) != TCL_OK) {
 	    goto end;
 	}
+	if (numObjStrings != -1 && objv[0]->typePtr != NULL
+		&& objv[0]->typePtr->freeIntRepProc != NULL) {
+	    objv[0]->typePtr->freeIntRepProc(objv[0]);
+	    objv[0]->typePtr = NULL;
+	}
 	if (Tcl_FSFileAttrsGet(interp, index, filePtr,
 		&objPtr) != TCL_OK) {
 	    goto end;
@@ -1012,6 +1023,11 @@ TclFileAttrsCmd(interp, objc, objv)
 		    "option", 0, &index) != TCL_OK) {
 		goto end;
     	    }
+	    if (numObjStrings != -1 && objv[i]->typePtr != NULL
+		    && objv[i]->typePtr->freeIntRepProc != NULL) {
+		objv[i]->typePtr->freeIntRepProc(objv[i]);
+		objv[i]->typePtr = NULL;
+	    }
 	    if (i + 1 == objc) {
 		Tcl_AppendResult(interp, "value for \"",
 			Tcl_GetString(objv[i]), "\" missing",
