@@ -9,59 +9,16 @@
  *
  * See the file "license.terms" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
- *
- * RCS: @(#) $Id: tclFileName.c,v 1.40.2.7 2004/03/09 12:56:59 vincentdarley Exp $
  */
 
+#include <sys/stat.h>
 #include "tclInt.h"
 #include "tclPort.h"
 #include "tclRegexp.h"
 
-/* 
- * This define is used to activate Tcl's interpretation of Unix-style
- * paths (containing forward slashes, '.' and '..') on MacOS.  A 
- * side-effect of this is that some paths become ambiguous.
- */
-#define MAC_UNDERSTANDS_UNIX_PATHS
-
-#ifdef MAC_UNDERSTANDS_UNIX_PATHS
-/*
- * The following regular expression matches the root portion of a Macintosh
- * absolute path.  It will match degenerate Unix-style paths, tilde paths,
- * Unix-style paths, and Mac paths.  The various subexpressions in this
- * can be summarised as follows: ^(/..|~user/unix|~user:mac|/unix|mac:dir).
- * The subexpression indices which match the root portions, are as follows:
- * 
- * degenerate unix-style: 2
- * unix-tilde: 5
- * mac-tilde: 7
- * unix-style: 9 (or 10 to cut off the irrelevant header).
- * mac: 12
- * 
- */
-
-#define MAC_ROOT_PATTERN "^((/+([.][.]?/+)*([.][.]?)?)|(~[^:/]*)(/[^:]*)?|(~[^:]*)(:.*)?|/+([.][.]?/+)*([^:/]+)(/[^:]*)?|([^:]+):.*)$"
-
-/*
- * The following variables are used to hold precompiled regular expressions
- * for use in filename matching.
- */
-
-typedef struct ThreadSpecificData {
-    int initialized;
-    Tcl_Obj *macRootPatternPtr;
-} ThreadSpecificData;
-
-static Tcl_ThreadDataKey dataKey;
-
-static void		FileNameCleanup _ANSI_ARGS_((ClientData clientData));
-static void		FileNameInit _ANSI_ARGS_((void));
-
-#endif
-
 /*
  * The following variable is set in the TclPlatformInit call to one
- * of: TCL_PLATFORM_UNIX, TCL_PLATFORM_MAC, or TCL_PLATFORM_WINDOWS.
+ * of: TCL_PLATFORM_UNIX, or TCL_PLATFORM_WINDOWS.
  */
 
 TclPlatformType tclPlatform = TCL_PLATFORM_UNIX;
@@ -77,64 +34,8 @@ static CONST char *	ExtractWinRoot _ANSI_ARGS_((CONST char *path,
 			    Tcl_PathType *typePtr));
 static int		SkipToChar _ANSI_ARGS_((char **stringPtr,
 			    char *match));
-static Tcl_Obj*		SplitMacPath _ANSI_ARGS_((CONST char *path));
 static Tcl_Obj*		SplitWinPath _ANSI_ARGS_((CONST char *path));
 static Tcl_Obj*		SplitUnixPath _ANSI_ARGS_((CONST char *path));
-#ifdef MAC_UNDERSTANDS_UNIX_PATHS
-
-/*
- *----------------------------------------------------------------------
- *
- * FileNameInit --
- *
- *	This procedure initializes the patterns used by this module.
- *
- * Results:
- *	None.
- *
- * Side effects:
- *	Compiles the regular expressions.
- *
- *----------------------------------------------------------------------
- */
-
-static void
-FileNameInit()
-{
-    ThreadSpecificData *tsdPtr = TCL_TSD_INIT(&dataKey);
-    if (!tsdPtr->initialized) {
-	tsdPtr->initialized = 1;
-	tsdPtr->macRootPatternPtr = Tcl_NewStringObj(MAC_ROOT_PATTERN, -1);
-	Tcl_CreateThreadExitHandler(FileNameCleanup, NULL);
-    }
-}
-
-/*
- *----------------------------------------------------------------------
- *
- * FileNameCleanup --
- *
- *	This procedure is a Tcl_ExitProc used to clean up the static
- *	data structures used in this file.
- *
- * Results:
- *	None.
- *
- * Side effects:
- *	Deallocates storage used by the procedures in this file.
- *
- *----------------------------------------------------------------------
- */
-
-static void
-FileNameCleanup(clientData)
-    ClientData clientData;	/* Not used. */
-{
-    ThreadSpecificData *tsdPtr = TCL_TSD_INIT(&dataKey);
-    Tcl_DecrRefCount(tsdPtr->macRootPatternPtr);
-    tsdPtr->initialized = 0;
-}
-#endif
 
 /*
  *----------------------------------------------------------------------
@@ -244,37 +145,47 @@ ExtractWinRoot(path, resultPtr, offset, typePtr)
 	}
     } else {
 	int abs = 0;
-	if (path[0] == 'c' && path[1] == 'o') {
-	    if (path[2] == 'm' && path[3] >= '1' && path[3] <= '9') {
-		/* May have match for 'com[1-9]:?', which is a serial port */
-	        if (path[4] == '\0') {
-	            abs = 4;
-	        } else if (path [4] == ':' && path[5] == '\0') {
+	if ((path[0] == 'c' || path[0] == 'C') 
+	    && (path[1] == 'o' || path[1] == 'O')) {
+	    if ((path[2] == 'm' || path[2] == 'M')
+		&& path[3] >= '1' && path[3] <= '4') {
+		/* May have match for 'com[1-4]:?', which is a serial port */
+		if (path[4] == '\0') {
+		    abs = 4;
+		} else if (path [4] == ':' && path[5] == '\0') {
 		    abs = 5;
-	        }
-	    } else if (path[2] == 'n' && path[3] == '\0') {
+		}
+	    } else if ((path[2] == 'n' || path[2] == 'N') && path[3] == '\0') {
 		/* Have match for 'con' */
 		abs = 3;
 	    }
-	} else if (path[0] == 'l' && path[1] == 'p' && path[2] == 't') {
-	    if (path[3] >= '1' && path[3] <= '9') {
-		/* May have match for 'lpt[1-9]:?' */
+	} else if ((path[0] == 'l' || path[0] == 'L')
+		   && (path[1] == 'p' || path[1] == 'P')
+		   && (path[2] == 't' || path[2] == 'T')) {
+	    if (path[3] >= '1' && path[3] <= '3') {
+		/* May have match for 'lpt[1-3]:?' */
 		if (path[4] == '\0') {
 		    abs = 4;
 		} else if (path [4] == ':' && path[5] == '\0') {
 		    abs = 5;
 		}
 	    }
-	} else if (path[0] == 'p' && path[1] == 'r' 
-		   && path[2] == 'n' && path[3] == '\0') {
+	} else if ((path[0] == 'p' || path[0] == 'P')
+		   && (path[1] == 'r' || path[1] == 'R')
+		   && (path[2] == 'n' || path[2] == 'N')
+		   && path[3] == '\0') {
 	    /* Have match for 'prn' */
 	    abs = 3;
-	} else if (path[0] == 'n' && path[1] == 'u' 
-		   && path[2] == 'l' && path[3] == '\0') {
+	} else if ((path[0] == 'n' || path[0] == 'N')
+		   && (path[1] == 'u' || path[1] == 'U')
+		   && (path[2] == 'l' || path[2] == 'L')
+		   && path[3] == '\0') {
 	    /* Have match for 'nul' */
 	    abs = 3;
-	} else if (path[0] == 'a' && path[1] == 'u' 
-		   && path[2] == 'x' && path[3] == '\0') {
+	} else if ((path[0] == 'a' || path[0] == 'A')
+		   && (path[1] == 'u' || path[1] == 'U')
+		   && (path[2] == 'x' || path[2] == 'X')
+		   && path[3] == '\0') {
 	    /* Have match for 'aux' */
 	    abs = 3;
 	}
@@ -402,87 +313,6 @@ TclpGetNativePathType(pathObjPtr, driveNameLengthPtr, driveNameRef)
 		}
 		break;
 	    }
-	    case TCL_PLATFORM_MAC:
-		if (path[0] == ':') {
-		    type = TCL_PATH_RELATIVE;
-		} else {
-#ifdef MAC_UNDERSTANDS_UNIX_PATHS
-		    ThreadSpecificData *tsdPtr;
-		    Tcl_RegExp re;
-
-		    tsdPtr = TCL_TSD_INIT(&dataKey);
-
-		    /*
-		     * Since we have eliminated the easy cases, use the
-		     * root pattern to look for the other types.
-		     */
-
-		    FileNameInit();
-		    re = Tcl_GetRegExpFromObj(NULL, tsdPtr->macRootPatternPtr,
-			    REG_ADVANCED);
-
-		    if (!Tcl_RegExpExec(NULL, re, path, path)) {
-			type = TCL_PATH_RELATIVE;
-		    } else {
-			CONST char *root, *end;
-			Tcl_RegExpRange(re, 2, &root, &end);
-			if (root != NULL) {
-			    type = TCL_PATH_RELATIVE;
-			} else {
-			    if (driveNameLengthPtr != NULL) {
-				Tcl_RegExpRange(re, 0, &root, &end);
-				*driveNameLengthPtr = end - root;
-			    }
-			    if (driveNameRef != NULL) {
-				if (*root == '/') {
-				    char *c;
-				    int gotColon = 0;
-				    *driveNameRef = Tcl_NewStringObj(root + 1,
-					    end - root -1);
-				    c = Tcl_GetString(*driveNameRef);
-				    while (*c != '\0') {
-					if (*c == '/') {
-					    gotColon++;
-					    *c = ':';
-					}
-					c++;
-				    }
-				    /* 
-				     * If there is no colon, we have just a
-				     * volume name so we must add a colon so
-				     * it is an absolute path.
-				     */
-				    if (gotColon == 0) {
-				        Tcl_AppendToObj(*driveNameRef, ":", 1);
-				    } else if ((gotColon > 1) &&
-					    (*(c-1) == ':')) {
-					/* We have an extra colon */
-				        Tcl_SetObjLength(*driveNameRef, 
-					  c - Tcl_GetString(*driveNameRef) - 1);
-				    }
-				}
-			    }
-			}
-		    }
-#else
-		    if (path[0] == '~') {
-		    } else if (path[0] == ':') {
-			type = TCL_PATH_RELATIVE;
-		    } else {
-			char *colonPos = strchr(path,':');
-			if (colonPos == NULL) {
-			    type = TCL_PATH_RELATIVE;
-			} else {
-			}
-		    }
-		    if (type == TCL_PATH_ABSOLUTE) {
-			if (driveNameLengthPtr != NULL) {
-			    *driveNameLengthPtr = strlen(path);
-			}
-		    }
-#endif
-		}
-		break;
 	    
 	    case TCL_PLATFORM_WINDOWS: {
 		Tcl_DString ds;
@@ -550,9 +380,6 @@ TclpNativeSplitPath(pathPtr, lenPtr)
 	    resultPtr = SplitWinPath(Tcl_GetString(pathPtr));
 	    break;
 	    
-	case TCL_PLATFORM_MAC:
-	    resultPtr = SplitMacPath(Tcl_GetString(pathPtr));
-	    break;
     }
 
     /*
@@ -782,8 +609,9 @@ SplitWinPath(path)
     Tcl_DStringFree(&buf);
     
     /*
-     * Split on slashes.  Embedded elements that start with tilde will be
-     * prefixed with "./" so they are not affected by tilde substitution.
+     * Split on slashes.  Embedded elements that start with tilde 
+     * or a drive letter will be prefixed with "./" so they are not 
+     * affected by tilde substitution.
      */
 
     do {
@@ -794,7 +622,10 @@ SplitWinPath(path)
 	length = p - elementStart;
 	if (length > 0) {
 	    Tcl_Obj *nextElt;
-	    if ((elementStart[0] == '~') && (elementStart != path)) {
+	    if ((elementStart != path)
+		&& ((elementStart[0] == '~')
+		    || (isalpha(UCHAR(elementStart[0]))
+			&& elementStart[1] == ':'))) {
 		nextElt = Tcl_NewStringObj("./",2);
 		Tcl_AppendToObj(nextElt, elementStart, length);
 	    } else {
@@ -807,246 +638,6 @@ SplitWinPath(path)
     return result;
 }
 
-/*
- *----------------------------------------------------------------------
- *
- * SplitMacPath --
- *
- *	This routine is used by Tcl_(FS)SplitPath to handle splitting
- *	Macintosh paths.
- *
- * Results:
- *	Returns a newly allocated Tcl list object.
- *
- * Side effects:
- *	None.
- *
- *----------------------------------------------------------------------
- */
-
-static Tcl_Obj*
-SplitMacPath(path)
-    CONST char *path;		/* Pointer to string containing a path. */
-{
-    int isMac = 0;		/* 1 if is Mac-style, 0 if Unix-style path. */
-    int length;
-    CONST char *p, *elementStart;
-    Tcl_Obj *result;
-#ifdef MAC_UNDERSTANDS_UNIX_PATHS
-    Tcl_RegExp re;
-    int i;
-    ThreadSpecificData *tsdPtr = TCL_TSD_INIT(&dataKey);
-#endif
-    
-    result = Tcl_NewObj();
-    
-#ifdef MAC_UNDERSTANDS_UNIX_PATHS
-    /*
-     * Initialize the path name parser for Macintosh path names.
-     */
-
-    FileNameInit();
-
-    /*
-     * Match the root portion of a Mac path name.
-     */
-
-    i = 0;			/* Needed only to prevent gcc warnings. */
-
-    re = Tcl_GetRegExpFromObj(NULL, tsdPtr->macRootPatternPtr, REG_ADVANCED);
-
-    if (Tcl_RegExpExec(NULL, re, path, path) == 1) {
-	CONST char *start, *end;
-	Tcl_Obj *nextElt;
-
-	/*
-	 * Treat degenerate absolute paths like / and /../.. as
-	 * Mac relative file names for lack of anything else to do.
-	 */
-
-	Tcl_RegExpRange(re, 2, &start, &end);
-	if (start) {
-	    Tcl_Obj *elt = Tcl_NewStringObj(":", 1);
-	    Tcl_RegExpRange(re, 0, &start, &end);
-	    Tcl_AppendToObj(elt, path, end - start);
-	    Tcl_ListObjAppendElement(NULL, result, elt);
-	    return result;
-	}
-
-	Tcl_RegExpRange(re, 5, &start, &end);
-	if (start) {
-	    /*
-	     * Unix-style tilde prefixed paths.
-	     */
-
-	    isMac = 0;
-	    i = 5;
-	} else {
-	    Tcl_RegExpRange(re, 7, &start, &end);
-	    if (start) {
-		/*
-		 * Mac-style tilde prefixed paths.
-		 */
-
-		isMac = 1;
-		i = 7;
-	    } else {
-		Tcl_RegExpRange(re, 10, &start, &end);
-		if (start) {
-		    /*
-		     * Normal Unix style paths.
-		     */
-
-		    isMac = 0;
-		    i = 10;
-		} else {
-		    Tcl_RegExpRange(re, 12, &start, &end);
-		    if (start) {
-			/*
-			 * Normal Mac style paths.
-			 */
-
-			isMac = 1;
-			i = 12;
-		    }
-		}
-	    }
-	}
-	Tcl_RegExpRange(re, i, &start, &end);
-	length = end - start;
-
-	/*
-	 * Append the element and terminate it with a : 
-	 */
-
-	nextElt = Tcl_NewStringObj(start, length);
-	Tcl_AppendToObj(nextElt, ":", 1);
-	Tcl_ListObjAppendElement(NULL, result, nextElt);
-	p = end;
-    } else {
-	isMac = (strchr(path, ':') != NULL);
-	p = path;
-    }
-#else
-    if ((path[0] != ':') && (path[0] == '~' || (strchr(path,':') != NULL))) {
-	CONST char *end;
-	Tcl_Obj *nextElt;
-
-	isMac = 1;
-	
-	end = strchr(path,':');
-	if (end == NULL) {
-	    length = strlen(path);
-	} else {
-	    length = end - path;
-	}
-
-	/*
-	 * Append the element and terminate it with a :
-	 */
-
-	nextElt = Tcl_NewStringObj(path, length);
-	Tcl_AppendToObj(nextElt, ":", 1);
-	Tcl_ListObjAppendElement(NULL, result, nextElt);
-	p = path + length;
-    } else {
-	isMac = (strchr(path, ':') != NULL);
-	isMac = 1;
-	p = path;
-    }
-#endif
-    
-    if (isMac) {
-
-	/*
-	 * p is pointing at the first colon in the path.  There
-	 * will always be one, since this is a Mac-style path.
-	 * (This is no longer true if MAC_UNDERSTANDS_UNIX_PATHS 
-	 * is false, so we must check whether 'p' points to the
-	 * end of the string.)
-	 */
-	elementStart = p;
-	if (*p == ':') {
-	    p++;
-	}
-	
-	while ((p = strchr(p, ':')) != NULL) {
-	    length = p - elementStart;
-	    if (length == 1) {
-		while (*p == ':') {
-		    Tcl_ListObjAppendElement(NULL, result,
-			    Tcl_NewStringObj("::", 2));
-		    elementStart = p++;
-		}
-	    } else {
-		/*
-		 * If this is a simple component, drop the leading colon.
-		 */
-
-		if ((elementStart[1] != '~')
-			&& (strchr(elementStart+1, '/') == NULL)) {
-		    elementStart++;
-		    length--;
-		}
-		Tcl_ListObjAppendElement(NULL, result, 
-			Tcl_NewStringObj(elementStart, length));
-		elementStart = p++;
-	    }
-	}
-	if (elementStart[0] != ':') {
-	    if (elementStart[0] != '\0') {
-		Tcl_ListObjAppendElement(NULL, result, 
-			Tcl_NewStringObj(elementStart, -1));
-	    }
-	} else {
-	    if (elementStart[1] != '\0' || elementStart == path) {
-		if ((elementStart[1] != '~') && (elementStart[1] != '\0')
-			&& (strchr(elementStart+1, '/') == NULL)) {
-		    elementStart++;
-		}
-		Tcl_ListObjAppendElement(NULL, result, 
-			Tcl_NewStringObj(elementStart, -1));
-	    }
-	}
-    } else {
-
-	/*
-	 * Split on slashes, suppress extra /'s, and convert .. to ::. 
-	 */
-
-	for (;;) {
-	    elementStart = p;
-	    while ((*p != '\0') && (*p != '/')) {
-		p++;
-	    }
-	    length = p - elementStart;
-	    if (length > 0) {
-		if ((length == 1) && (elementStart[0] == '.')) {
-		    Tcl_ListObjAppendElement(NULL, result, 
-					     Tcl_NewStringObj(":", 1));
-		} else if ((length == 2) && (elementStart[0] == '.')
-			&& (elementStart[1] == '.')) {
-		    Tcl_ListObjAppendElement(NULL, result, 
-					     Tcl_NewStringObj("::", 2));
-		} else {
-		    Tcl_Obj *nextElt;
-		    if (*elementStart == '~') {
-			nextElt = Tcl_NewStringObj(":",1);
-			Tcl_AppendToObj(nextElt, elementStart, length);
-		    } else {
-			nextElt = Tcl_NewStringObj(elementStart, length);
-		    }
-		    Tcl_ListObjAppendElement(NULL, result, nextElt);
-		}
-	    }
-	    if (*p++ == '\0') {
-		break;
-	    }
-	}
-    }
-    return result;
-}
-
 /*
  *---------------------------------------------------------------------------
  *
@@ -1115,23 +706,25 @@ TclpNativeJoinPath(prefix, joining)
     start = Tcl_GetStringFromObj(prefix, &length);
 
     /*
-     * Remove the ./ from tilde prefixed elements unless
-     * it is the first component.
+     * Remove the ./ from tilde prefixed elements, and drive-letter
+     * prefixed elements on Windows, unless it is the first component.
      */
     
     p = joining;
     
     if (length != 0) {
-	if ((p[0] == '.') && (p[1] == '/') && (p[2] == '~')) {
+	if ((p[0] == '.') && (p[1] == '/')
+	    && ((p[2] == '~')
+		|| ((tclPlatform == TCL_PLATFORM_WINDOWS)
+		    && isalpha(UCHAR(p[2]))
+		    && (p[3] == ':')))) {
 	    p += 2;
 	}
     }
-       
     if (*p == '\0') {
 	return;
     }
 
-
     switch (tclPlatform) {
         case TCL_PLATFORM_UNIX:
 	    /*
@@ -1140,7 +733,7 @@ TclpNativeJoinPath(prefix, joining)
 
 	    if (length > 0 && (start[length-1] != '/')) {
 		Tcl_AppendToObj(prefix, "/", 1);
-		length++;
+		Tcl_GetStringFromObj(prefix, &length);
 	    }
 	    needsSep = 0;
 	    
@@ -1179,7 +772,7 @@ TclpNativeJoinPath(prefix, joining)
 	    if ((length > 0) && 
 		(start[length-1] != '/') && (start[length-1] != ':')) {
 		Tcl_AppendToObj(prefix, "/", 1);
-		length++;
+		Tcl_GetStringFromObj(prefix, &length);
 	    }
 	    needsSep = 0;
 	    
@@ -1207,85 +800,6 @@ TclpNativeJoinPath(prefix, joining)
 	    Tcl_SetObjLength(prefix, length);
 	    break;
 
-	case TCL_PLATFORM_MAC: {
-	    int newLength;
-	    
-	    /*
-	     * Sort out separators.  We basically add the object we've
-	     * been given, but we have to make sure that there is
-	     * exactly one separator inbetween (unless the object we're
-	     * adding contains multiple contiguous colons, all of which
-	     * we must add).  Also if an object is just ':' we don't
-	     * bother to add it unless it's the very first element.
-	     */
-
-#ifdef MAC_UNDERSTANDS_UNIX_PATHS
-	    int adjustedPath = 0;
-	    if ((strchr(p, ':') == NULL) && (strchr(p, '/') != NULL)) {
-		char *start = p;
-		adjustedPath = 1;
-		while (*start != '\0') {
-		    if (*start == '/') {
-		        *start = ':';
-		    }
-		    start++;
-		}
-	    }
-#endif
-	    if (length > 0) {
-		if ((p[0] == ':') && (p[1] == '\0')) {
-		    return;
-		}
-		if (start[length-1] != ':') {
-		    if (*p != '\0' && *p != ':') {
-			Tcl_AppendToObj(prefix, ":", 1);
-			length++;
-		    }
-		} else if (*p == ':') {
-		    p++;
-		}
-	    } else {
-		if (*p != '\0' && *p != ':') {
-		    Tcl_AppendToObj(prefix, ":", 1);
-		    length++;
-		}
-	    }
-	    
-	    /*
-	     * Append the element
-	     */
-
-	    newLength = strlen(p);
-	    /* 
-	     * It may not be good to just do 'Tcl_AppendToObj(prefix,
-	     * p, newLength)' because the object may contain duplicate
-	     * colons which we want to get rid of.
-	     */
-	    Tcl_AppendToObj(prefix, p, newLength);
-	    
-	    /* Remove spurious trailing single ':' */
-	    dest = Tcl_GetString(prefix) + length + newLength;
-	    if (*(dest-1) == ':') {
-		if (dest-1 > Tcl_GetString(prefix)) {
-		    if (*(dest-2) != ':') {
-		        Tcl_SetObjLength(prefix, length + newLength -1);
-		    }
-		}
-	    }
-#ifdef MAC_UNDERSTANDS_UNIX_PATHS
-	    /* Revert the path to what it was */
-	    if (adjustedPath) {
-		char *start = joining;
-		while (*start != '\0') {
-		    if (*start == ':') {
-			*start = '/';
-		    }
-		    start++;
-		}
-	    }
-#endif
-	    break;
-	}
     }
     return;
 }
@@ -1444,18 +958,6 @@ TclGetExtension(name)
 	    lastSep = strrchr(name, '/');
 	    break;
 
-	case TCL_PLATFORM_MAC:
-#ifdef MAC_UNDERSTANDS_UNIX_PATHS
-	    if (strchr(name, ':') == NULL) {
-		lastSep = strrchr(name, '/');
-	    } else {
-		lastSep = strrchr(name, ':');
-	    }
-#else
-	    lastSep = strrchr(name, ':');
-#endif
-	    break;
-
 	case TCL_PLATFORM_WINDOWS:
 	    lastSep = NULL;
 	    for (p = name; *p != '\0'; p++) {
@@ -1684,9 +1186,6 @@ Tcl_GlobObjCmd(dummy, interp, objc, objv)
 	case TCL_PLATFORM_WINDOWS:
 	    separators = "/\\:";
 	    break;
-	case TCL_PLATFORM_MAC:
-	    separators = ":";
-	    break;
     }
     if (dir == PATH_GENERAL) {
 	int pathlength;
@@ -1758,6 +1257,9 @@ Tcl_GlobObjCmd(dummy, interp, objc, objv)
 	 * on an incompatible platform.
 	 */
 	Tcl_ListObjLength(interp, typePtr, &length);
+	if (length <= 0) {
+	    goto skipTypes;
+	}
 	globTypes = (Tcl_GlobTypeData*) ckalloc(sizeof(Tcl_GlobTypeData));
 	globTypes->type = 0;
 	globTypes->perm = 0;
@@ -1863,6 +1365,7 @@ Tcl_GlobObjCmd(dummy, interp, objc, objv)
 	}
     }
 
+  skipTypes:
     /* 
      * Now we perform the actual glob below.  This may involve joining
      * together the pattern arguments, dealing with particular file types
@@ -2021,17 +1524,6 @@ TclGlob(interp, pattern, unquotedPrefix, globFlags, types)
 	case TCL_PLATFORM_WINDOWS:
 	    separators = "/\\:";
 	    break;
-	case TCL_PLATFORM_MAC:
-#ifdef MAC_UNDERSTANDS_UNIX_PATHS
-	    if (unquotedPrefix == NULL) {
-		separators = (strchr(pattern, ':') == NULL) ? "/" : ":";
-	    } else {
-		separators = ":";
-	    }
-#else
-	    separators = ":";
-#endif
-	    break;
     }
 
     Tcl_DStringInit(&buffer);
@@ -2114,6 +1606,10 @@ TclGlob(interp, pattern, unquotedPrefix, globFlags, types)
 	     */
 	    if (globFlags & TCL_GLOBMODE_DIR) {
 		Tcl_DStringAppend(&buffer,separators,1);
+		/* Try to borrow that separator from the tail */
+		if (*tail == *separators) {
+		    tail++;
+		}
 	    }
 	    prefixLen++;
 	}
@@ -2160,16 +1656,6 @@ TclGlob(interp, pattern, unquotedPrefix, globFlags, types)
 
 	Tcl_ListObjGetElements(NULL, Tcl_GetObjResult(interp), 
 			       &objc, &objv);
-#ifdef MAC_TCL
-	/* adjust prefixLen if TclDoGlob prepended a ':' */
-	if ((prefixLen > 0) && (objc > 0)
-	&& (Tcl_DStringValue(&buffer)[0] != ':')) {
-	    char *str = Tcl_GetStringFromObj(objv[0],NULL);
-	    if (str[0] == ':') {
-		    prefixLen++;
-	    }
-	}
-#endif
 	for (i = 0; i< objc; i++) {
 	    Tcl_Obj* elt;
 	    if (globFlags & TCL_GLOBMODE_TAILS) {
@@ -2298,13 +1784,17 @@ TclDoGlob(interp, separators, headPtr, tail, types)
     int baseLength, quoted, count;
     int result = TCL_OK;
     char *name, *p, *openBrace, *closeBrace, *firstSpecialChar, savedChar;
+    /*
     char lastChar = 0;
+    */
     
     int length = Tcl_DStringLength(headPtr);
 
+    /*
     if (length > 0) {
 	lastChar = Tcl_DStringValue(headPtr)[length-1];
     }
+    */
 
     /*
      * Consume any leading directory separators, leaving tail pointing
@@ -2330,6 +1820,11 @@ TclDoGlob(interp, separators, headPtr, tail, types)
 	} else if (strchr(separators, *tail) == NULL) {
 	    break;
 	}
+	    if (*tail == '\\') {
+		Tcl_DStringAppend(headPtr, separators, 1);
+	    } else {
+		Tcl_DStringAppend(headPtr, tail, 1);
+	    }
 	count++;
     }
 
@@ -2340,38 +1835,12 @@ TclDoGlob(interp, separators, headPtr, tail, types)
      */
 
     switch (tclPlatform) {
-	case TCL_PLATFORM_MAC:
-#ifdef MAC_UNDERSTANDS_UNIX_PATHS
-	    if (*separators == '/') {
-		if (((length == 0) && (count == 0))
-			|| ((length > 0) && (lastChar != ':'))) {
-		    Tcl_DStringAppend(headPtr, ":", 1);
-		}
-	    } else {
-#endif
-		if (count == 0) {
-		    if ((length > 0) && (lastChar != ':')) {
-			Tcl_DStringAppend(headPtr, ":", 1);
-		    }
-		} else {
-		    if (lastChar == ':') {
-			count--;
-		    }
-		    while (count-- > 0) {
-			Tcl_DStringAppend(headPtr, ":", 1);
-		    }
-		}
-#ifdef MAC_UNDERSTANDS_UNIX_PATHS
-	    }
-#endif
-	    break;
 	case TCL_PLATFORM_WINDOWS:
 	    /*
 	     * If this is a drive relative path, add the colon and the
 	     * trailing slash if needed.  Otherwise add the slash if
 	     * this is the first absolute element, or a later relative
 	     * element.  Add an extra slash if this is a UNC path.
-	     */
 
 	    if (*name == ':') {
 		Tcl_DStringAppend(headPtr, ":", 1);
@@ -2387,13 +1856,13 @@ TclDoGlob(interp, separators, headPtr, tail, types)
 		    Tcl_DStringAppend(headPtr, "/", 1);
 		}
 	    }
+	     */
 	    
 	    break;
-	case TCL_PLATFORM_UNIX:
+	case TCL_PLATFORM_UNIX: {
 	    /*
 	     * Add a separator if this is the first absolute element, or
 	     * a later relative element.
-	     */
 
 	    if ((*tail != '\0')
 		    && (((length > 0)
@@ -2401,7 +1870,9 @@ TclDoGlob(interp, separators, headPtr, tail, types)
 			    || ((length == 0) && (count > 0)))) {
 		Tcl_DStringAppend(headPtr, "/", 1);
 	    }
+	     */
 	    break;
+	}
     }
 
     /*
@@ -2513,41 +1984,61 @@ TclDoGlob(interp, separators, headPtr, tail, types)
 	    ret = Tcl_FSMatchInDirectory(interp, Tcl_GetObjResult(interp), 
 					 head, tail, types);
 	} else {
-	    Tcl_Obj* resultPtr;
-
 	    /* 
 	     * We do the recursion ourselves.  This makes implementing
 	     * Tcl_FSMatchInDirectory for each filesystem much easier.
 	     */
 	    Tcl_GlobTypeData dirOnly = { TCL_GLOB_TYPE_DIR, 0, NULL, NULL };
 	    char save = *p;
-	    
-	    *p = '\0';
+	    Tcl_Obj *resultPtr;
+
 	    resultPtr = Tcl_NewListObj(0, NULL);
+	    Tcl_IncrRefCount(resultPtr);
+	    *p = '\0';
 	    ret = Tcl_FSMatchInDirectory(interp, resultPtr, 
 					 head, tail, &dirOnly);
 	    *p = save;
 	    if (ret == TCL_OK) {
-		int resLength;
+		int resLength, repair = -1;
 		ret = Tcl_ListObjLength(interp, resultPtr, &resLength);
 		if (ret == TCL_OK) {
 		    int i;
 		    for (i =0; i< resLength; i++) {
 			Tcl_Obj *elt;
 			Tcl_DString ds;
-			Tcl_ListObjIndex(interp, resultPtr, i, &elt);
+			Tcl_ListObjIndex(NULL, resultPtr, i, &elt);
 			Tcl_DStringInit(&ds);
-			Tcl_DStringAppend(&ds, Tcl_GetString(elt), -1);
-			if(tclPlatform == TCL_PLATFORM_MAC) {
-			    Tcl_DStringAppend(&ds, ":",1);
-			} else {			
-			    Tcl_DStringAppend(&ds, "/",1);
+			if (Tcl_GetString(elt)[0] == '~') {
+			    Tcl_Obj *paths = Tcl_GetObjResult(interp);
+
+			    Tcl_ListObjLength(NULL, paths, &repair);
+			    Tcl_DStringAppend(&ds, "./", 2);
 			}
+			Tcl_DStringAppend(&ds, Tcl_GetString(elt), -1);
+			    Tcl_DStringAppend(&ds, "/",1);
 			ret = TclDoGlob(interp, separators, &ds, p+1, types);
 			Tcl_DStringFree(&ds);
 			if (ret != TCL_OK) {
 			    break;
 			}
+			if (repair >= 0) {
+			    Tcl_Obj *paths = Tcl_GetObjResult(interp);
+			    int end;
+
+			    Tcl_ListObjLength(NULL, paths, &end);
+			    while (repair < end) {
+				CONST char *bytes;
+				int numBytes;
+				Tcl_Obj *fixme, *newObj;
+				Tcl_ListObjIndex(NULL, paths, repair, &fixme);
+				bytes = Tcl_GetStringFromObj(fixme, &numBytes);
+				newObj = Tcl_NewStringObj(bytes+2, numBytes-2);
+				Tcl_ListObjReplace(NULL, paths, repair, 1,
+					1, &newObj);
+				repair++;
+			    }
+			    repair = -1;
+			}
 		    }
 		}
 	    }
@@ -2575,12 +2066,6 @@ TclDoGlob(interp, separators, headPtr, tail, types)
 	Tcl_Obj *nameObj;
 
 	switch (tclPlatform) {
-	    case TCL_PLATFORM_MAC: {
-		if (strchr(Tcl_DStringValue(headPtr), ':') == NULL) {
-		    Tcl_DStringAppend(headPtr, ":", 1);
-		}
-		break;
-	    }
 	    case TCL_PLATFORM_WINDOWS: {
 		if (Tcl_DStringLength(headPtr) == 0) {
 		    if (((*name == '\\') && (name[1] == '/' || name[1] == '\\'))
@@ -2590,17 +2075,6 @@ TclDoGlob(interp, separators, headPtr, tail, types)
 			Tcl_DStringAppend(headPtr, ".", 1);
 		    }
 		}
-#if defined(__CYGWIN__) && defined(__WIN32__)
-		{
-		extern int cygwin_conv_to_win32_path 
-		    _ANSI_ARGS_((CONST char *, char *));
-		char winbuf[MAX_PATH+1];
-
-		cygwin_conv_to_win32_path(Tcl_DStringValue(headPtr), winbuf);
-		Tcl_DStringFree(headPtr);
-		Tcl_DStringAppend(headPtr, winbuf, -1);
-		}
-#endif /* __CYGWIN__ && __WIN32__ */
 		/* 
 		 * Convert to forward slashes.  This is required to pass
 		 * some Tcl tests.  We should probably remove the conversions
@@ -2630,10 +2104,10 @@ TclDoGlob(interp, separators, headPtr, tail, types)
 	nameObj = Tcl_NewStringObj(name, Tcl_DStringLength(headPtr));
 
 	Tcl_IncrRefCount(nameObj);
-	Tcl_FSMatchInDirectory(interp, Tcl_GetObjResult(interp), nameObj, 
-			       NULL, types);
+	result = Tcl_FSMatchInDirectory(interp, Tcl_GetObjResult(interp),
+		nameObj, NULL, types);
 	Tcl_DecrRefCount(nameObj);
-	return TCL_OK;
+	return result;
     }
 }
 
@@ -2693,8 +2167,7 @@ TclFileDirname(interp, pathPtr)
 	splitResultPtr = Tcl_FSJoinPath(splitPtr, splitElements - 1);
     } else if (splitElements == 0 || 
       (Tcl_FSGetPathType(pathPtr) == TCL_PATH_RELATIVE)) {
-	splitResultPtr = Tcl_NewStringObj(
-		((tclPlatform == TCL_PLATFORM_MAC) ? ":" : "."), 1);
+	splitResultPtr = Tcl_NewStringObj(".", 1);
     } else {
 	Tcl_ListObjIndex(NULL, splitPtr, 0, &splitResultPtr);
     }
