@@ -10,13 +10,11 @@
  *
  * See the file "license.terms" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
- *
- * RCS: @(#) $Id: tclGet.c,v 1.8.2.1 2005/04/20 16:06:17 dgp Exp $
  */
 
 #include "tclInt.h"
 #include "tclPort.h"
-#include "tclMath.h"
+#include <math.h>
 
 
 /*
