@@ -7,8 +7,6 @@
  *
  * See the file "license.terms" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
- *
- * RCS: @(#) $Id: tclIOCmd.c,v 1.15.2.5 2008/04/10 20:53:48 andreas_kupries Exp $
  */
 
 #include "tclInt.h"
@@ -423,7 +421,7 @@ Tcl_SeekObjCmd(clientData, interp, objc, objv)
     static CONST char *originOptions[] = {
 	"start", "current", "end", (char *) NULL
     };
-    static int modeArray[] = {SEEK_SET, SEEK_CUR, SEEK_END};
+    static CONST int modeArray[] = {SEEK_SET, SEEK_CUR, SEEK_END};
 
     if ((objc != 3) && (objc != 4)) {
 	Tcl_WrongNumArgs(interp, 1, objv, "channelId offset ?origin?");
@@ -708,14 +706,6 @@ Tcl_ExecObjCmd(dummy, interp, objc, objv)
     int objc;				/* Number of arguments. */
     Tcl_Obj *CONST objv[];		/* Argument objects. */
 {
-#ifdef MAC_TCL
-
-    Tcl_AppendResult(interp, "exec not implemented under Mac OS",
-		(char *)NULL);
-    return TCL_ERROR;
-
-#else /* !MAC_TCL */
-
     /*
      * This procedure generates an argv array for the string arguments. It
      * starts out with stack-allocated space but uses dynamically-allocated
@@ -781,7 +771,7 @@ Tcl_ExecObjCmd(dummy, interp, objc, objv)
 
     argv = argStorage;
     argc = objc - skip;
-    if ((argc + 1) > sizeof(argv) / sizeof(argv[0])) {
+    if ((argc + 1) > (int)(sizeof(argv) / sizeof(argv[0]))) {
 	argv = (CONST char **) ckalloc((unsigned)(argc + 1) * sizeof(char *));
     }
 
@@ -856,7 +846,6 @@ Tcl_ExecObjCmd(dummy, interp, objc, objv)
     Tcl_SetObjResult(interp, resultPtr);
 
     return result;
-#endif /* !MAC_TCL */
 }
 
 /*
@@ -967,12 +956,6 @@ Tcl_OpenObjCmd(notUsed, interp, objc, objv)
     if (!pipeline) {
         chan = Tcl_FSOpenFileChannel(interp, objv[1], modeString, prot);
     } else {
-#ifdef MAC_TCL
-	Tcl_AppendResult(interp,
-		"command pipelines not supported on Macintosh OS",
-		(char *)NULL);
-	return TCL_ERROR;
-#else
 	int mode, seekFlag, cmdObjc;
 	CONST char **cmdArgv;
 
@@ -1002,7 +985,6 @@ Tcl_OpenObjCmd(notUsed, interp, objc, objv)
 	    chan = Tcl_OpenCommandChannel(interp, cmdObjc, cmdArgv, flags);
 	}
         ckfree((char *) cmdArgv);
-#endif
     }
     if (chan == (Tcl_Channel) NULL) {
         return TCL_ERROR;
