@@ -3,19 +3,17 @@
  *
  *	Common routines used by all socket based channel types.
  *
- * Copyright (c) 1995 Sun Microsystems, Inc.
+ * Copyright (c) 1995-1997 Sun Microsystems, Inc.
  *
  * See the file "license.terms" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
- *
- * SCCS: @(#) tclIOSock.c 1.18 96/08/16 07:09:27
  */
 
 #include "tclInt.h"
 #include "tclPort.h"
 
 /*
- *----------------------------------------------------------------------
+ *---------------------------------------------------------------------------
  *
  * TclSockGetPort --
  *
@@ -24,14 +22,14 @@
  *	registered service names to port numbers.
  *
  * Results:
- *	A standard Tcl result.  On success, the port number is
- *	returned in portPtr. On failure, an error message is left in
- *	interp->result.
+ *	A standard Tcl result.  On success, the port number is returned
+ *	in portPtr. On failure, an error message is left in the interp's
+ *	result.
  *
  * Side effects:
  *	None.
  *
- *----------------------------------------------------------------------
+ *---------------------------------------------------------------------------
  */
 
 int
@@ -41,10 +39,22 @@ TclSockGetPort(interp, string, proto, portPtr)
     char *proto;		/* "tcp" or "udp", typically */
     int *portPtr;		/* Return port number */
 {
-    struct servent *sp = getservbyname(string, proto);    
-    if (sp != NULL) {
-	*portPtr = ntohs((unsigned short) sp->s_port);
-	return TCL_OK;
+    struct servent *sp;		/* Protocol info for named services */
+    Tcl_DString ds;
+    CONST char *native;
+
+    if (Tcl_GetInt(NULL, string, portPtr) != TCL_OK) {
+	/*
+	 * Don't bother translating 'proto' to native.
+	 */
+	 
+	native = Tcl_UtfToExternalDString(NULL, string, -1, &ds);
+	sp = getservbyname(native, proto);		/* INTL: Native. */
+	Tcl_DStringFree(&ds);
+	if (sp != NULL) {
+	    *portPtr = ntohs((unsigned short) sp->s_port);
+	    return TCL_OK;
+	}
     }
     if (Tcl_GetInt(interp, string, portPtr) != TCL_OK) {
 	return TCL_ERROR;
@@ -73,24 +83,30 @@ TclSockGetPort(interp, string, proto, portPtr)
  *----------------------------------------------------------------------
  */
 
+#undef TclSockMinimumBuffers
+#if !defined(_WIN32) && !defined(__CYGWIN__)
+#   define SOCKET int
+#endif
+
 int
 TclSockMinimumBuffers(sock, size)
-    int sock;			/* Socket file descriptor */
+    void *sock;			/* Socket file descriptor */
     int size;			/* Minimum buffer size */
 {
     int current;
-    int len = sizeof(int);
+    socklen_t len;
 
-    getsockopt(sock, SOL_SOCKET, SO_SNDBUF, (char *)&current, (size_t *)&len);
+    len = sizeof(int);
+    getsockopt((SOCKET)(size_t)sock, SOL_SOCKET, SO_SNDBUF, (char *)&current, &len);
     if (current < size) {
 	len = sizeof(int);
-	setsockopt(sock, SOL_SOCKET, SO_SNDBUF, (char *)&size, len);
+	setsockopt((SOCKET)(size_t)sock, SOL_SOCKET, SO_SNDBUF, (char *)&size, len);
     }
     len = sizeof(int);
-    getsockopt(sock, SOL_SOCKET, SO_RCVBUF, (char *)&current, (size_t *)&len);
+    getsockopt((SOCKET)(size_t)sock, SOL_SOCKET, SO_RCVBUF, (char *)&current, &len);
     if (current < size) {
 	len = sizeof(int);
-	setsockopt(sock, SOL_SOCKET, SO_RCVBUF, (char *)&size, len);
+	setsockopt((SOCKET)(size_t)sock, SOL_SOCKET, SO_RCVBUF, (char *)&size, len);
     }
     return TCL_OK;
 }
