@@ -10,8 +10,6 @@
  *
  * See the file "license.terms" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
- *
- * RCS: @(#) $Id: tclListObj.c,v 1.13.4.1 2004/11/11 01:18:07 das Exp $
  */
 
 #include "tclInt.h"
@@ -657,6 +655,10 @@ Tcl_ListObjReplace(interp, listPtr, first, count, objc, objv)
 	count = 0;
     }
     
+    for (i = 0;  i < objc;  i++) {
+	Tcl_IncrRefCount(objv[i]);
+    }
+
     numRequired = (numElems - count + objc);
     if (numRequired <= listRepPtr->maxElemCount) {
 	/*
@@ -691,7 +693,6 @@ Tcl_ListObjReplace(interp, listPtr, first, count, objc, objv)
 
 	for (i = 0, j = first;  i < objc;  i++, j++) {
             elemPtrs[j] = objv[i];
-            Tcl_IncrRefCount(objv[i]);
         }
 
 	/*
@@ -747,7 +748,6 @@ Tcl_ListObjReplace(interp, listPtr, first, count, objc, objv)
 
 	for (i = 0, j = first;  i < objc;  i++, j++) {
 	    newPtrs[j] = objv[i];
-	    Tcl_IncrRefCount(objv[i]);
 	}
 
 	listRepPtr->elemCount = numRequired;
@@ -1623,6 +1623,12 @@ UpdateStringOfList(listPtr)
 	elem = Tcl_GetStringFromObj(listRepPtr->elements[i], &length);
 	listPtr->length += Tcl_ScanCountedElement(elem, length,
 		&flagPtr[i]) + 1;
+	/*
+	 * Check for continued sanity. [Bug 1267380]
+	 */
+	if (listPtr->length < 1) {
+	    Tcl_Panic("string representation size exceeds sane bounds");
+	}
     }
 
     /*
