@@ -12,8 +12,6 @@
  *
  * See the file "license.terms" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
- *
- * RCS: @(#) $Id: tclParse.c,v 1.25.2.2 2007/07/19 13:43:42 dgp Exp $
  */
 
 #include "tclInt.h"
@@ -618,7 +616,7 @@ TclParseBackslash(src, numBytes, readPtr, dst)
             result = 0xb;
             break;
         case 'x':
-	    count += TclParseHex(p+1, numBytes-1, &result);
+	    count += TclParseHex(p+1, numBytes-2, &result);
 	    if (count == 2) {
 		/* No hexadigits -> This is just "x". */
 		result = 'x';
@@ -628,7 +626,7 @@ TclParseBackslash(src, numBytes, readPtr, dst)
 	    }
             break;
         case 'u':
-	    count += TclParseHex(p+1, (numBytes > 5) ? 4 : numBytes-1, &result);
+	    count += TclParseHex(p+1, (numBytes > 5) ? 4 : numBytes-2, &result);
 	    if (count == 2) {
 		/* No hexadigits -> This is just "u". */
 		result = 'u';
@@ -1019,7 +1017,15 @@ TclExpandTokenArray(parsePtr)
     int newCount;
     Tcl_Token *newPtr;
 
+#define MAX_TOKENS (int)(UINT_MAX / sizeof(Tcl_Token))
+
+    if (parsePtr->tokensAvailable == MAX_TOKENS) {
+	Tcl_Panic("max # of tokens for a Tcl parse (%d) exceeded", MAX_TOKENS);
+    }
     newCount = parsePtr->tokensAvailable*2;
+    if (newCount > MAX_TOKENS) {
+	newCount = MAX_TOKENS;
+    }
     newPtr = (Tcl_Token *) ckalloc((unsigned) (newCount * sizeof(Tcl_Token)));
     memcpy((VOID *) newPtr, (VOID *) parsePtr->tokenPtr,
 	    (size_t) (parsePtr->tokensAvailable * sizeof(Tcl_Token)));
@@ -1441,7 +1447,7 @@ Tcl_ParseBraces(interp, string, numBytes, parsePtr, append, termPtr)
 	     *  by a '<whitespace>#' on the same line.
 	     */
 
-	    for (; src > string; src--) {
+	    while (--src > string) {
 		switch (*src) {
 		    case '{':
 			openBrace = 1;
