@@ -7,15 +7,14 @@
  *	code analysis, etc.
  *
  * Copyright (c) 1997 Sun Microsystems, Inc.
+ * Copyright (c) 1998-2000 by Scriptics Corporation.
+ * Contributions from Don Porter, NIST, 2002.  (not subject to US copyright)
  *
  * See the file "license.terms" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
- *
- * RCS: @(#) $Id: tclParseExpr.c,v 1.6 1999/12/04 06:15:42 hobbs Exp $
  */
 
 #include "tclInt.h"
-#include "tclCompile.h"
 
 /*
  * The stuff below is a bit of a hack so that this file can be used in
@@ -55,22 +54,24 @@ typedef struct ParseInfo {
     int lexeme;			/* Type of last lexeme scanned in expr.
 				 * See below for definitions. Corresponds to
 				 * size characters beginning at start. */
-    char *start;		/* First character in lexeme. */
+    CONST char *start;		/* First character in lexeme. */
     int size;			/* Number of bytes in lexeme. */
-    char *next;			/* Position of the next character to be
+    CONST char *next;		/* Position of the next character to be
 				 * scanned in the expression string. */
-    char *prevEnd;		/* Points to the character just after the
+    CONST char *prevEnd;	/* Points to the character just after the
 				 * last one in the previous lexeme. Used to
 				 * compute size of subexpression tokens. */
-    char *originalExpr;		/* Points to the start of the expression
+    CONST char *originalExpr;	/* Points to the start of the expression
 				 * originally passed to Tcl_ParseExpr. */
-    char *lastChar;		/* Points just after last byte of expr. */
+    CONST char *lastChar;	/* Points just after last byte of expr. */
 } ParseInfo;
 
 /*
  * Definitions of the different lexemes that appear in expressions. The
  * order of these must match the corresponding entries in the
  * operatorStrings array below.
+ *
+ * Basic lexemes:
  */
 
 #define LITERAL		0
@@ -84,62 +85,69 @@ typedef struct ParseInfo {
 #define COMMA		8
 #define END		9
 #define UNKNOWN		10
+#define UNKNOWN_CHAR	11
 
 /*
- * Binary operators:
+ * Binary numeric operators:
  */
 
-#define MULT		11
-#define DIVIDE		12
-#define MOD		13
-#define PLUS		14
-#define MINUS		15
-#define LEFT_SHIFT	16
-#define RIGHT_SHIFT	17
-#define LESS		18
-#define GREATER		19
-#define LEQ		20
-#define GEQ		21
-#define EQUAL		22
-#define NEQ		23
-#define BIT_AND		24
-#define BIT_XOR		25
-#define BIT_OR		26
-#define AND		27
-#define OR		28
-#define QUESTY		29
-#define COLON		30
+#define MULT		12
+#define DIVIDE		13
+#define MOD		14
+#define PLUS		15
+#define MINUS		16
+#define LEFT_SHIFT	17
+#define RIGHT_SHIFT	18
+#define LESS		19
+#define GREATER		20
+#define LEQ		21
+#define GEQ		22
+#define EQUAL		23
+#define NEQ		24
+#define BIT_AND		25
+#define BIT_XOR		26
+#define BIT_OR		27
+#define AND		28
+#define OR		29
+#define QUESTY		30
+#define COLON		31
 
 /*
  * Unary operators. Unary minus and plus are represented by the (binary)
  * lexemes MINUS and PLUS.
  */
 
-#define NOT		31
-#define BIT_NOT		32
+#define NOT		32
+#define BIT_NOT		33
+
+/*
+ * Binary string operators:
+ */
+
+#define STREQ		34
+#define STRNEQ		35
 
 /*
  * Mapping from lexemes to strings; used for debugging messages. These
  * entries must match the order and number of the lexeme definitions above.
  */
 
-#ifdef TCL_COMPILE_DEBUG
-static char *lexemeStrings[] = {
+static CONST char *CONST lexemeStrings[] = {
     "LITERAL", "FUNCNAME",
-    "[", "{", "(", ")", "$", "\"", ",", "END", "UNKNOWN",
+    "[", "{", "(", ")", "$", "\"", ",", "END", "UNKNOWN", "UNKNOWN_CHAR",
     "*", "/", "%", "+", "-",
     "<<", ">>", "<", ">", "<=", ">=", "==", "!=",
     "&", "^", "|", "&&", "||", "?", ":",
-    "!", "~"
+    "!", "~", "eq", "ne",
 };
-#endif /* TCL_COMPILE_DEBUG */
 
 /*
  * Declarations for local procedures to this file:
  */
 
 static int		GetLexeme _ANSI_ARGS_((ParseInfo *infoPtr));
-static void		LogSyntaxError _ANSI_ARGS_((ParseInfo *infoPtr));
+static void		LogSyntaxError _ANSI_ARGS_((ParseInfo *infoPtr,
+				CONST char *extraInfo));
 static int		ParseAddExpr _ANSI_ARGS_((ParseInfo *infoPtr));
 static int		ParseBitAndExpr _ANSI_ARGS_((ParseInfo *infoPtr));
 static int		ParseBitOrExpr _ANSI_ARGS_((ParseInfo *infoPtr));
@@ -148,14 +156,16 @@ static int		ParseCondExpr _ANSI_ARGS_((ParseInfo *infoPtr));
 static int		ParseEqualityExpr _ANSI_ARGS_((ParseInfo *infoPtr));
 static int		ParseLandExpr _ANSI_ARGS_((ParseInfo *infoPtr));
 static int		ParseLorExpr _ANSI_ARGS_((ParseInfo *infoPtr));
+static int		ParseMaxDoubleLength _ANSI_ARGS_((CONST char *string,
+				CONST char *end));
 static int		ParseMultiplyExpr _ANSI_ARGS_((ParseInfo *infoPtr));
 static int		ParsePrimaryExpr _ANSI_ARGS_((ParseInfo *infoPtr));
 static int		ParseRelationalExpr _ANSI_ARGS_((ParseInfo *infoPtr));
 static int		ParseShiftExpr _ANSI_ARGS_((ParseInfo *infoPtr));
 static int		ParseUnaryExpr _ANSI_ARGS_((ParseInfo *infoPtr));
-static void		PrependSubExprTokens _ANSI_ARGS_((char *op,
-			    int opBytes, char *src, int srcBytes,
-			    int firstIndex, ParseInfo *infoPtr));
+static void		PrependSubExprTokens _ANSI_ARGS_((CONST char *op,
+				int opBytes, CONST char *src, int srcBytes,
+				int firstIndex, ParseInfo *infoPtr));
 
 /*
  * Macro used to debug the execution of the recursive descent parser used
@@ -181,7 +191,8 @@ static void		PrependSubExprTokens _ANSI_ARGS_((char *op,
  *	Given a string, this procedure parses the first Tcl expression
  *	in the string and returns information about the structure of
  *	the expression. This procedure is the top-level interface to the
- *	the expression parsing module.
+ *	the expression parsing module.  No more that numBytes bytes will
+ *	be scanned.
  *
  * Results:
  *	The return value is TCL_OK if the command was parsed successfully
@@ -203,7 +214,7 @@ static void		PrependSubExprTokens _ANSI_ARGS_((char *op,
 int
 Tcl_ParseExpr(interp, string, numBytes, parsePtr)
     Tcl_Interp *interp;		/* Used for error reporting. */
-    char *string;		/* The source string to parse. */
+    CONST char *string;		/* The source string to parse. */
     int numBytes;		/* Number of bytes in string. If < 0, the
 				 * string consists of all bytes up to the
 				 * first null character. */
@@ -214,7 +225,6 @@ Tcl_ParseExpr(interp, string, numBytes, parsePtr)
 {
     ParseInfo info;
     int code;
-    char savedChar;
 
     if (numBytes < 0) {
 	numBytes = (string? strlen(string) : 0);
@@ -240,17 +250,6 @@ Tcl_ParseExpr(interp, string, numBytes, parsePtr)
     parsePtr->term = string;
     parsePtr->incomplete = 0;
 
-    /*
-     * Temporarily overwrite the character just after the end of the
-     * string with a 0 byte.  This acts as a sentinel and reduces the
-     * number of places where we have to check for the end of the
-     * input string.  The original value of the byte is restored at
-     * the end of the parse.
-     */
-
-    savedChar = string[numBytes];
-    string[numBytes] = 0;
-
     /*
      * Initialize the ParseInfo structure that holds state while parsing
      * the expression.
@@ -278,14 +277,12 @@ Tcl_ParseExpr(interp, string, numBytes, parsePtr)
 	goto error;
     }
     if (info.lexeme != END) {
-	LogSyntaxError(&info);
+	LogSyntaxError(&info, "extra tokens at end of expression");
 	goto error;
     }
-    string[numBytes] = (char) savedChar;
     return TCL_OK;
     
     error:
-    string[numBytes] = (char) savedChar;
     if (parsePtr->tokenPtr != parsePtr->staticTokens) {
 	ckfree((char *) parsePtr->tokenPtr);
     }
@@ -301,7 +298,7 @@ Tcl_ParseExpr(interp, string, numBytes, parsePtr)
  *	condExpr ::= lorExpr ['?' condExpr ':' condExpr]
  *
  *	Note that this is the topmost recursive-descent parsing routine used
- *	by TclParseExpr to parse expressions. This avoids an extra procedure
+ *	by Tcl_ParseExpr to parse expressions. This avoids an extra procedure
  *	call since such a procedure would only return the result of calling
  *	ParseCondExpr. Other recursive-descent procedures that need to parse
  *	complete expressions also call ParseCondExpr.
@@ -327,7 +324,7 @@ ParseCondExpr(infoPtr)
     Tcl_Parse *parsePtr = infoPtr->parsePtr;
     Tcl_Token *tokenPtr, *firstTokenPtr, *condTokenPtr;
     int firstIndex, numToMove, code;
-    char *srcStart;
+    CONST char *srcStart;
     
     HERE("condExpr", 1);
     srcStart = infoPtr->start;
@@ -384,7 +381,7 @@ ParseCondExpr(infoPtr)
 	    return code;
 	}
 	if (infoPtr->lexeme != COLON) {
-	    LogSyntaxError(infoPtr);
+	    LogSyntaxError(infoPtr, "missing colon from ternary conditional");
 	    return TCL_ERROR;
 	}
 	code = GetLexeme(infoPtr); /* skip over the ':' */
@@ -440,7 +437,7 @@ ParseLorExpr(infoPtr)
 {
     Tcl_Parse *parsePtr = infoPtr->parsePtr;
     int firstIndex, code;
-    char *srcStart, *operator;
+    CONST char *srcStart, *operator;
     
     HERE("lorExpr", 2);
     srcStart = infoPtr->start;
@@ -500,7 +497,7 @@ ParseLandExpr(infoPtr)
 {
     Tcl_Parse *parsePtr = infoPtr->parsePtr;
     int firstIndex, code;
-    char *srcStart, *operator;
+    CONST char *srcStart, *operator;
 
     HERE("landExpr", 3);
     srcStart = infoPtr->start;
@@ -560,7 +557,7 @@ ParseBitOrExpr(infoPtr)
 {
     Tcl_Parse *parsePtr = infoPtr->parsePtr;
     int firstIndex, code;
-    char *srcStart, *operator;
+    CONST char *srcStart, *operator;
 
     HERE("bitOrExpr", 4);
     srcStart = infoPtr->start;
@@ -621,7 +618,7 @@ ParseBitXorExpr(infoPtr)
 {
     Tcl_Parse *parsePtr = infoPtr->parsePtr;
     int firstIndex, code;
-    char *srcStart, *operator;
+    CONST char *srcStart, *operator;
 
     HERE("bitXorExpr", 5);
     srcStart = infoPtr->start;
@@ -682,7 +679,7 @@ ParseBitAndExpr(infoPtr)
 {
     Tcl_Parse *parsePtr = infoPtr->parsePtr;
     int firstIndex, code;
-    char *srcStart, *operator;
+    CONST char *srcStart, *operator;
 
     HERE("bitAndExpr", 6);
     srcStart = infoPtr->start;
@@ -720,7 +717,8 @@ ParseBitAndExpr(infoPtr)
  * ParseEqualityExpr --
  *
  *	This procedure parses a Tcl equality (inequality) expression:
- *	equalityExpr ::= relationalExpr {('==' | '!=') relationalExpr}
+ *	equalityExpr ::= relationalExpr
+ *		{('==' | '!=' | 'ne' | 'eq') relationalExpr}
  *
  * Results:
  *	The return value is TCL_OK on a successful parse and TCL_ERROR
@@ -742,7 +740,7 @@ ParseEqualityExpr(infoPtr)
 {
     Tcl_Parse *parsePtr = infoPtr->parsePtr;
     int firstIndex, lexeme, code;
-    char *srcStart, *operator;
+    CONST char *srcStart, *operator;
 
     HERE("equalityExpr", 7);
     srcStart = infoPtr->start;
@@ -754,9 +752,10 @@ ParseEqualityExpr(infoPtr)
     }
 
     lexeme = infoPtr->lexeme;
-    while ((lexeme == EQUAL) || (lexeme == NEQ)) {
+    while ((lexeme == EQUAL) || (lexeme == NEQ)
+	    || (lexeme == STREQ) || (lexeme == STRNEQ)) {
 	operator = infoPtr->start;
-	code = GetLexeme(infoPtr); /* skip over == or != */
+	code = GetLexeme(infoPtr); /* skip over ==, !=, 'eq' or 'ne'  */
 	if (code != TCL_OK) {
 	    return code;
 	}
@@ -766,7 +765,8 @@ ParseEqualityExpr(infoPtr)
 	}
 
 	/*
-	 * Generate tokens for the subexpression and '==' or '!=' operator.
+	 * Generate tokens for the subexpression and '==', '!=', 'eq' or 'ne'
+	 * operator.
 	 */
 
 	PrependSubExprTokens(operator, 2, srcStart,
@@ -804,7 +804,7 @@ ParseRelationalExpr(infoPtr)
 {
     Tcl_Parse *parsePtr = infoPtr->parsePtr;
     int firstIndex, lexeme, operatorSize, code;
-    char *srcStart, *operator;
+    CONST char *srcStart, *operator;
 
     HERE("relationalExpr", 8);
     srcStart = infoPtr->start;
@@ -872,7 +872,7 @@ ParseShiftExpr(infoPtr)
 {
     Tcl_Parse *parsePtr = infoPtr->parsePtr;
     int firstIndex, lexeme, code;
-    char *srcStart, *operator;
+    CONST char *srcStart, *operator;
 
     HERE("shiftExpr", 9);
     srcStart = infoPtr->start;
@@ -934,7 +934,7 @@ ParseAddExpr(infoPtr)
 {
     Tcl_Parse *parsePtr = infoPtr->parsePtr;
     int firstIndex, lexeme, code;
-    char *srcStart, *operator;
+    CONST char *srcStart, *operator;
 
     HERE("addExpr", 10);
     srcStart = infoPtr->start;
@@ -996,7 +996,7 @@ ParseMultiplyExpr(infoPtr)
 {
     Tcl_Parse *parsePtr = infoPtr->parsePtr;
     int firstIndex, lexeme, code;
-    char *srcStart, *operator;
+    CONST char *srcStart, *operator;
 
     HERE("multiplyExpr", 11);
     srcStart = infoPtr->start;
@@ -1058,7 +1058,7 @@ ParseUnaryExpr(infoPtr)
 {
     Tcl_Parse *parsePtr = infoPtr->parsePtr;
     int firstIndex, lexeme, code;
-    char *srcStart, *operator;
+    CONST char *srcStart, *operator;
 
     HERE("unaryExpr", 12);
     srcStart = infoPtr->start;
@@ -1123,7 +1123,7 @@ ParsePrimaryExpr(infoPtr)
     Tcl_Interp *interp = parsePtr->interp;
     Tcl_Token *tokenPtr, *exprTokenPtr;
     Tcl_Parse nested;
-    char *dollarPtr, *stringStart, *termPtr, *src;
+    CONST char *dollarPtr, *stringStart, *termPtr, *src;
     int lexeme, exprIndex, firstIndex, numToMove, code;
 
     /*
@@ -1142,7 +1142,8 @@ ParsePrimaryExpr(infoPtr)
 	    return code;
 	}
 	if (infoPtr->lexeme != CLOSE_PAREN) {
-	    goto syntaxError;
+	    LogSyntaxError(infoPtr, "looking for close parenthesis");
+	    return TCL_ERROR;
 	}
 	code = GetLexeme(infoPtr); /* skip over the ')' */
 	if (code != TCL_OK) {
@@ -1178,6 +1179,7 @@ ParsePrimaryExpr(infoPtr)
 	 * Int or double number.
 	 */
 	
+	tokenizeLiteral:
 	if (parsePtr->numTokens == parsePtr->tokensAvailable) {
 	    TclExpandTokenArray(parsePtr);
 	}
@@ -1192,7 +1194,7 @@ ParsePrimaryExpr(infoPtr)
 	exprTokenPtr->size = infoPtr->size;
 	exprTokenPtr->numComponents = 1;
 	break;
-	
+
     case DOLLAR:
 	/*
 	 * $var variable reference.
@@ -1284,10 +1286,23 @@ ParsePrimaryExpr(infoPtr)
 		return TCL_ERROR;
 	    }
 	    src = (nested.commandStart + nested.commandSize);
+
+	    /*
+	     * This is equivalent to Tcl_FreeParse(&nested), but
+	     * presumably inlined here for sake of runtime optimization
+	     */
+
 	    if (nested.tokenPtr != nested.staticTokens) {
 		ckfree((char *) nested.tokenPtr);
 	    }
-	    if ((src[-1] == ']') && !nested.incomplete) {
+
+	    /*
+	     * Check for the closing ']' that ends the command substitution.
+	     * It must have been the last character of the parsed command.
+	     */
+
+	    if ((nested.term < parsePtr->end) && (*nested.term == ']') 
+		    && !nested.incomplete) {
 		break;
 	    }
 	    if (src == parsePtr->end) {
@@ -1352,28 +1367,87 @@ ParsePrimaryExpr(infoPtr)
 	}
 	break;
 	
-    case FUNC_NAME:
+/*
+ *	Disable attempt to support functions named "eq" or "ne".  This
+ *	is unworkable in the Tcl 8.4.* releases.  See Tcl Bugs 1971879
+ *	and 1201589.
+ *
+    case STREQ:
+    case STRNEQ:
+*/
+    case FUNC_NAME: {
 	/*
 	 * math_func '(' expr {',' expr} ')'
 	 */
-	
-	if (parsePtr->numTokens == parsePtr->tokensAvailable) {
-	    TclExpandTokenArray(parsePtr);
-	}
-	tokenPtr = &parsePtr->tokenPtr[parsePtr->numTokens];
-	tokenPtr->type = TCL_TOKEN_OPERATOR;
-	tokenPtr->start = infoPtr->start;
-	tokenPtr->size = infoPtr->size;
-	tokenPtr->numComponents = 0;
-	parsePtr->numTokens++;
+
+	ParseInfo savedInfo = *infoPtr;
 	
 	code = GetLexeme(infoPtr); /* skip over function name */
 	if (code != TCL_OK) {
 	    return code;
 	}
 	if (infoPtr->lexeme != OPEN_PAREN) {
-	    goto syntaxError;
+	    int code;
+	    Tcl_DString functionName;
+	    Tcl_HashEntry *hPtr;
+	    Interp *iPtr = (Interp *) infoPtr->parsePtr->interp;
+	    Tcl_Obj *objPtr = Tcl_NewStringObj(savedInfo.start, savedInfo.size);
+
+	    /* Check for boolean literals (true, false, yes, no, on, off) */
+	    Tcl_IncrRefCount(objPtr);
+	    code = Tcl_ConvertToType(NULL, objPtr, &tclBooleanType);
+	    Tcl_DecrRefCount(objPtr);
+	    if (code == TCL_OK) {
+		*infoPtr = savedInfo;
+		goto tokenizeLiteral;
+	    }
+
+	    /*
+	     * Guess what kind of error we have by trying to tell
+	     * whether we have a function or variable name here.
+	     * Alas, this makes the parser more tightly bound with the
+	     * rest of the interpreter, but that is the only way to
+	     * give a sensible message here.  Still, it is not too
+	     * serious as this is only done when generating an error.
+	     */
+
+	    /*
+	     * Look up the name as a function name.  We need a writable
+	     * copy (DString) so we can terminate it with a NULL for
+	     * the benefit of Tcl_FindHashEntry which operates on
+	     * NULL-terminated string keys.
+	     */
+	    Tcl_DStringInit(&functionName);
+	    hPtr = Tcl_FindHashEntry(&iPtr->mathFuncTable, 
+	    	Tcl_DStringAppend(&functionName,
+			savedInfo.start, savedInfo.size));
+	    Tcl_DStringFree(&functionName);
+
+	    /*
+	     * Assume that we have an attempted variable reference
+	     * unless we've got a function name, as the set of
+	     * potential function names is typically much smaller.
+	     */
+	    if (hPtr != NULL) {
+		LogSyntaxError(infoPtr,
+			"expected parenthesis enclosing function arguments");
+	    } else {
+		LogSyntaxError(infoPtr,
+			"variable references require preceding $");
+	    }
+	    return TCL_ERROR;
 	}
+
+	if (parsePtr->numTokens == parsePtr->tokensAvailable) {
+	    TclExpandTokenArray(parsePtr);
+	}
+	tokenPtr = &parsePtr->tokenPtr[parsePtr->numTokens];
+	tokenPtr->type = TCL_TOKEN_OPERATOR;
+	tokenPtr->start = savedInfo.start;
+	tokenPtr->size = savedInfo.size;
+	tokenPtr->numComponents = 0;
+	parsePtr->numTokens++;
+	
 	code = GetLexeme(infoPtr); /* skip over '(' */
 	if (code != TCL_OK) {
 	    return code;
@@ -1391,7 +1465,9 @@ ParsePrimaryExpr(infoPtr)
 		    return code;
 		}
 	    } else if (infoPtr->lexeme != CLOSE_PAREN) {
-		goto syntaxError;
+		LogSyntaxError(infoPtr,
+			"missing close parenthesis at end of function call");
+		return TCL_ERROR;
 	    }
 	}
 
@@ -1399,9 +1475,38 @@ ParsePrimaryExpr(infoPtr)
 	exprTokenPtr->size = (infoPtr->next - exprTokenPtr->start);
 	exprTokenPtr->numComponents = parsePtr->numTokens - firstIndex;
 	break;
-	
-    default:
-	goto syntaxError;
+    }
+
+    case COMMA:
+	LogSyntaxError(infoPtr,
+		"commas can only separate function arguments");
+	return TCL_ERROR;
+    case END:
+	LogSyntaxError(infoPtr, "premature end of expression");
+	return TCL_ERROR;
+    case UNKNOWN:
+	LogSyntaxError(infoPtr, "single equality character not legal in expressions");
+	return TCL_ERROR;
+    case UNKNOWN_CHAR:
+	LogSyntaxError(infoPtr, "character not legal in expressions");
+	return TCL_ERROR;
+    case QUESTY:
+	LogSyntaxError(infoPtr, "unexpected ternary 'then' separator");
+	return TCL_ERROR;
+    case COLON:
+	LogSyntaxError(infoPtr, "unexpected ternary 'else' separator");
+	return TCL_ERROR;
+    case CLOSE_PAREN:
+	LogSyntaxError(infoPtr, "unexpected close parenthesis");
+	return TCL_ERROR;
+
+    default: {
+	char buf[64];
+
+	sprintf(buf, "unexpected operator %s", lexemeStrings[lexeme]);
+	LogSyntaxError(infoPtr, buf);
+	return TCL_ERROR;
+	}
     }
 
     /*
@@ -1414,10 +1519,6 @@ ParsePrimaryExpr(infoPtr)
     }
     parsePtr->term = infoPtr->next;
     return TCL_OK;
-
-    syntaxError:
-    LogSyntaxError(infoPtr);
-    return TCL_ERROR;
 }
 
 /*
@@ -1453,11 +1554,9 @@ GetLexeme(infoPtr)
     ParseInfo *infoPtr;		/* Holds state needed to parse the expr,
 				 * including the resulting lexeme. */
 {
-    register char *src;		/* Points to current source char. */
-    char *termPtr;		/* Points to char terminating a literal. */
-    double doubleValue;		/* Value of a scanned double literal. */
+    register CONST char *src;	/* Points to current source char. */
     char c;
-    int startsWithDigit, offset;
+    int offset, length, numBytes;
     Tcl_Parse *parsePtr = infoPtr->parsePtr;
     Tcl_Interp *interp = parsePtr->interp;
     Tcl_UniChar ch;
@@ -1471,26 +1570,18 @@ GetLexeme(infoPtr)
     infoPtr->prevEnd = infoPtr->next;
 
     /*
-     * Scan over leading white space at the start of a lexeme. Note that a
-     * backslash-newline is treated as a space.
+     * Scan over leading white space at the start of a lexeme. 
      */
 
     src = infoPtr->next;
-    c = *src;
-    while (isspace(UCHAR(c)) || (c == '\\')) { /* INTL: ISO space */
-	if (c == '\\') {
-	    if (src[1] == '\n') {
-		src += 2;
-	    } else {
-		break;	/* no longer white space */
-	    }
-	} else {
-	    src++;
-	}
-	c = *src;
-    }
+    numBytes = parsePtr->end - src;
+    do {
+	char type;
+	int scanned = TclParseWhiteSpace(src, numBytes, parsePtr, &type);
+	src += scanned; numBytes -= scanned;
+    } while  (numBytes && (*src == '\n') && (src++,numBytes--));
     parsePtr->term = src;
-    if (src >= infoPtr->lastChar) {
+    if (numBytes == 0) {
 	infoPtr->lexeme = END;
 	infoPtr->next = src;
 	return TCL_OK;
@@ -1503,59 +1594,48 @@ GetLexeme(infoPtr)
      * by mistake, which would eventually cause a syntax error.
      */
 
+    c = *src;
     if ((c != '+') && (c != '-')) {
-	startsWithDigit = isdigit(UCHAR(c)); /* INTL: digit */
-	if (startsWithDigit && TclLooksLikeInt(src, -1)) {
-	    errno = 0;
-	    (void) strtoul(src, &termPtr, 0);
-	    if (errno == ERANGE) {
-		if (interp != NULL) {
-		    char *s = "integer value too large to represent";
-		    Tcl_ResetResult(interp);
-		    Tcl_AppendToObj(Tcl_GetObjResult(interp), s, -1);
-		    Tcl_SetErrorCode(interp, "ARITH", "IOVERFLOW", s,
-			    (char *) NULL);
-		}
+	CONST char *end = infoPtr->lastChar;
+	if ((length = TclParseInteger(src, (end - src)))) {
+	    /*
+	     * First length bytes look like an integer.  Verify by
+	     * attempting the conversion to the largest integer we have.
+	     */
+	    int code;
+	    Tcl_WideInt wide;
+	    Tcl_Obj *value = Tcl_NewStringObj(src, length);
+
+	    Tcl_IncrRefCount(value);
+	    code = Tcl_GetWideIntFromObj(interp, value, &wide);
+	    Tcl_DecrRefCount(value);
+	    if (code == TCL_ERROR) {
 		parsePtr->errorType = TCL_PARSE_BAD_NUMBER;
 		return TCL_ERROR;
 	    }
- 	    if (termPtr != src) {
-                /*
-                 * src was the start of a valid integer, but was it
-		 * a bad octal?  Stopping at a digit would cause that.
-                 */
-		if (isdigit(UCHAR(*termPtr))) {	/* INTL: digit. */
-		    /*
-		     * We only want to report an error for the number,
-		     * but we may have something like "08+1"
-		     */
-		    if (interp != NULL) {
-			while (isdigit(UCHAR(*(++termPtr)))) {} /* INTL: digit. */
-			Tcl_ResetResult(interp);
-			offset = termPtr - src;
-			c = src[offset];
-			src[offset] = 0;
-			Tcl_AppendResult(interp, "\"", src,
-				"\" is an invalid octal number",
-				(char *) NULL);
-			src[offset] = c;
-		    }
-		    parsePtr->errorType = TCL_PARSE_BAD_NUMBER;
-		    return TCL_ERROR;
-		}
+            infoPtr->lexeme = LITERAL;
+	    infoPtr->start = src;
+	    infoPtr->size = length;
+            infoPtr->next = (src + length);
+	    parsePtr->term = infoPtr->next;
+            return TCL_OK;
+	} else if ((length = ParseMaxDoubleLength(src, end))) {
+	    /*
+	     * There are length characters that could be a double.
+	     * Let strtod() tells us for sure.  Need a writable copy
+	     * so we can set an terminating NULL to keep strtod from
+	     * scanning too far.
+	     */
+	    char *startPtr, *termPtr;
+	    double doubleValue;
+	    Tcl_DString toParse;
 
-                infoPtr->lexeme = LITERAL;
-		infoPtr->start = src;
-		infoPtr->size = (termPtr - src);
-                infoPtr->next = termPtr;
-		parsePtr->term = termPtr;
-                return TCL_OK;
-	    }
-	} else if (startsWithDigit || (c == '.')
-	        || (c == 'n') || (c == 'N')) {
 	    errno = 0;
-	    doubleValue = strtod(src, &termPtr);
-	    if (termPtr != src) {
+	    Tcl_DStringInit(&toParse);
+	    startPtr = Tcl_DStringAppend(&toParse, src, length);
+	    doubleValue = strtod(startPtr, &termPtr);
+	    Tcl_DStringFree(&toParse);
+	    if (termPtr != startPtr) {
 		if (errno != 0) {
 		    if (interp != NULL) {
 			TclExprFloatError(interp, doubleValue);
@@ -1565,14 +1645,19 @@ GetLexeme(infoPtr)
 		}
 		
 		/*
-                 * src was the start of a valid double.
+                 * startPtr was the start of a valid double, copied
+		 * from src.
                  */
 		
 		infoPtr->lexeme = LITERAL;
 		infoPtr->start = src;
-		infoPtr->size = (termPtr - src);
-		infoPtr->next = termPtr;
-		parsePtr->term = termPtr;
+		if ((termPtr - startPtr) > length) {
+		    infoPtr->size = length;
+		} else {
+		    infoPtr->size = (termPtr - startPtr);
+		}
+		infoPtr->next = src + infoPtr->size;
+		parsePtr->term = infoPtr->next;
 		return TCL_OK;
 	    }
 	}
@@ -1646,72 +1731,69 @@ GetLexeme(infoPtr)
 	    return TCL_OK;
 
 	case '<':
-	    switch (src[1]) {
-		case '<':
-		    infoPtr->lexeme = LEFT_SHIFT;
-		    infoPtr->size = 2;
-		    infoPtr->next = src+2;
-		    break;
-		case '=':
-		    infoPtr->lexeme = LEQ;
-		    infoPtr->size = 2;
-		    infoPtr->next = src+2;
-		    break;
-		default:
-		    infoPtr->lexeme = LESS;
-		    break;
+	    infoPtr->lexeme = LESS;
+	    if ((infoPtr->lastChar - src) > 1) {
+		switch (src[1]) {
+		    case '<':
+			infoPtr->lexeme = LEFT_SHIFT;
+			infoPtr->size = 2;
+			infoPtr->next = src+2;
+			break;
+		    case '=':
+			infoPtr->lexeme = LEQ;
+			infoPtr->size = 2;
+			infoPtr->next = src+2;
+			break;
+		}
 	    }
 	    parsePtr->term = infoPtr->next;
 	    return TCL_OK;
 
 	case '>':
-	    switch (src[1]) {
-		case '>':
-		    infoPtr->lexeme = RIGHT_SHIFT;
-		    infoPtr->size = 2;
-		    infoPtr->next = src+2;
-		    break;
-		case '=':
-		    infoPtr->lexeme = GEQ;
-		    infoPtr->size = 2;
-		    infoPtr->next = src+2;
-		    break;
-		default:
-		    infoPtr->lexeme = GREATER;
-		    break;
+	    infoPtr->lexeme = GREATER;
+	    if ((infoPtr->lastChar - src) > 1) {
+		switch (src[1]) {
+		    case '>':
+			infoPtr->lexeme = RIGHT_SHIFT;
+			infoPtr->size = 2;
+			infoPtr->next = src+2;
+			break;
+		    case '=':
+			infoPtr->lexeme = GEQ;
+			infoPtr->size = 2;
+			infoPtr->next = src+2;
+			break;
+		}
 	    }
 	    parsePtr->term = infoPtr->next;
 	    return TCL_OK;
 
 	case '=':
-	    if (src[1] == '=') {
+	    infoPtr->lexeme = UNKNOWN;
+	    if ((src[1] == '=') && ((infoPtr->lastChar - src) > 1)) {
 		infoPtr->lexeme = EQUAL;
 		infoPtr->size = 2;
 		infoPtr->next = src+2;
-	    } else {
-		infoPtr->lexeme = UNKNOWN;
 	    }
 	    parsePtr->term = infoPtr->next;
 	    return TCL_OK;
 
 	case '!':
-	    if (src[1] == '=') {
+	    infoPtr->lexeme = NOT;
+	    if ((src[1] == '=') && ((infoPtr->lastChar - src) > 1)) {
 		infoPtr->lexeme = NEQ;
 		infoPtr->size = 2;
 		infoPtr->next = src+2;
-	    } else {
-		infoPtr->lexeme = NOT;
 	    }
 	    parsePtr->term = infoPtr->next;
 	    return TCL_OK;
 
 	case '&':
-	    if (src[1] == '&') {
+	    infoPtr->lexeme = BIT_AND;
+	    if ((src[1] == '&') && ((infoPtr->lastChar - src) > 1)) {
 		infoPtr->lexeme = AND;
 		infoPtr->size = 2;
 		infoPtr->next = src+2;
-	    } else {
-		infoPtr->lexeme = BIT_AND;
 	    }
 	    parsePtr->term = infoPtr->next;
 	    return TCL_OK;
@@ -1721,12 +1803,11 @@ GetLexeme(infoPtr)
 	    return TCL_OK;
 
 	case '|':
-	    if (src[1] == '|') {
+	    infoPtr->lexeme = BIT_OR;
+	    if ((src[1] == '|') && ((infoPtr->lastChar - src) > 1)) {
 		infoPtr->lexeme = OR;
 		infoPtr->size = 2;
 		infoPtr->next = src+2;
-	    } else {
-		infoPtr->lexeme = BIT_OR;
 	    }
 	    parsePtr->term = infoPtr->next;
 	    return TCL_OK;
@@ -1735,14 +1816,52 @@ GetLexeme(infoPtr)
 	    infoPtr->lexeme = BIT_NOT;
 	    return TCL_OK;
 
+	case 'e':
+	    if ((src[1] == 'q') && ((infoPtr->lastChar - src) > 1)) {
+		infoPtr->lexeme = STREQ;
+		infoPtr->size = 2;
+		infoPtr->next = src+2;
+		parsePtr->term = infoPtr->next;
+		return TCL_OK;
+	    } else {
+		goto checkFuncName;
+	    }
+
+	case 'n':
+	    if ((src[1] == 'e') && ((infoPtr->lastChar - src) > 1)) {
+		infoPtr->lexeme = STRNEQ;
+		infoPtr->size = 2;
+		infoPtr->next = src+2;
+		parsePtr->term = infoPtr->next;
+		return TCL_OK;
+	    } else {
+		goto checkFuncName;
+	    }
+
 	default:
-	    offset = Tcl_UtfToUniChar(src, &ch);
+	checkFuncName:
+	    length = (infoPtr->lastChar - src);
+	    if (Tcl_UtfCharComplete(src, length)) {
+		offset = Tcl_UtfToUniChar(src, &ch);
+	    } else {
+		char utfBytes[TCL_UTF_MAX];
+		memcpy(utfBytes, src, (size_t) length);
+		utfBytes[length] = '\0';
+		offset = Tcl_UtfToUniChar(utfBytes, &ch);
+	    }
 	    c = UCHAR(ch);
 	    if (isalpha(UCHAR(c))) {	/* INTL: ISO only. */
 		infoPtr->lexeme = FUNC_NAME;
 		while (isalnum(UCHAR(c)) || (c == '_')) { /* INTL: ISO only. */
-		    src += offset;
-		    offset = Tcl_UtfToUniChar(src, &ch);
+		    src += offset; length -= offset;
+		    if (Tcl_UtfCharComplete(src, length)) {
+			offset = Tcl_UtfToUniChar(src, &ch);
+		    } else {
+			char utfBytes[TCL_UTF_MAX];
+			memcpy(utfBytes, src, (size_t) length);
+			utfBytes[length] = '\0';
+			offset = Tcl_UtfToUniChar(utfBytes, &ch);
+		    }
 		    c = UCHAR(ch);
 		}
 		infoPtr->size = (src - infoPtr->start);
@@ -1750,11 +1869,114 @@ GetLexeme(infoPtr)
 		parsePtr->term = infoPtr->next;
 		return TCL_OK;
 	    }
-	    infoPtr->lexeme = UNKNOWN;
+	    infoPtr->lexeme = UNKNOWN_CHAR;
 	    return TCL_OK;
     }
 }
 
+/*
+ *----------------------------------------------------------------------
+ *
+ * TclParseInteger --
+ *
+ *	Scans up to numBytes bytes starting at src, and checks whether
+ *	the leading bytes look like an integer's string representation.
+ *
+ * Results:
+ *	Returns 0 if the leading bytes do not look like an integer.
+ *	Otherwise, returns the number of bytes examined that look
+ *	like an integer.  This may be less than numBytes if the integer
+ *	is only the leading part of the string.
+ *
+ * Side effects:
+ *	None.
+ *
+ *----------------------------------------------------------------------
+ */
+
+int
+TclParseInteger(string, numBytes)
+    register CONST char *string;/* The string to examine. */
+    register int numBytes;	/* Max number of bytes to scan. */
+{
+    register CONST char *p = string;
+
+    /* Take care of introductory "0x" */
+    if ((numBytes > 1) && (p[0] == '0') && ((p[1] == 'x') || (p[1] == 'X'))) {
+	int scanned;
+	Tcl_UniChar ch;
+	p+=2; numBytes -= 2;
+ 	scanned = TclParseHex(p, numBytes, &ch);
+	if (scanned) {
+	    return scanned + 2;
+	}
+
+	/* Recognize the 0 as valid integer, but x is left behind */
+	return 1;
+    }
+    while (numBytes && isdigit(UCHAR(*p))) {	/* INTL: digit */
+	numBytes--; p++;
+    }
+    if (numBytes == 0) {
+        return (p - string);
+    }
+    if ((*p != '.') && (*p != 'e') && (*p != 'E')) {
+        return (p - string);
+    }
+    return 0;
+}
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * ParseMaxDoubleLength --
+ *
+ *      Scans a sequence of bytes checking that the characters could
+ *	be in a string rep of a double.
+ *
+ * Results:
+ *	Returns the number of bytes starting with string, runing to, but
+ *	not including end, all of which could be part of a string rep.
+ *	of a double.  Only character identity is used, no actual
+ *	parsing is done.
+ *
+ *	The legal bytes are '0' - '9', 'A' - 'F', 'a' - 'f', 
+ *	'.', '+', '-', 'i', 'I', 'n', 'N', 'p', 'P', 'x',  and 'X'.
+ *	This covers the values "Inf" and "Nan" as well as the
+ *	decimal and hexadecimal representations recognized by a
+ *	C99-compliant strtod().
+ *
+ * Side effects:
+ *	None.
+ *
+ *----------------------------------------------------------------------
+ */
+
+static int
+ParseMaxDoubleLength(string, end)
+    register CONST char *string;/* The string to examine. */
+    CONST char *end;		/* Point to the first character past the end
+				 * of the string we are examining. */
+{
+    CONST char *p = string;
+    while (p < end) {
+	switch (*p) {
+	    case '0': case '1': case '2': case '3': case '4': case '5':
+	    case '6': case '7': case '8': case '9': case 'A': case 'B':
+	    case 'C': case 'D': case 'E': case 'F': case 'I': case 'N':
+	    case 'P': case 'X': case 'a': case 'b': case 'c': case 'd':
+	    case 'e': case 'f': case 'i': case 'n': case 'p': case 'x':
+	    case '.': case '+': case '-':
+		p++;
+		break;
+	    default:
+		goto done;
+	}
+    }
+    done:
+    return (p - string);
+}
+
 /*
  *----------------------------------------------------------------------
  *
@@ -1777,10 +1999,10 @@ GetLexeme(infoPtr)
 
 static void
 PrependSubExprTokens(op, opBytes, src, srcBytes, firstIndex, infoPtr)
-    char *op;			/* Points to first byte of the operator
+    CONST char *op;		/* Points to first byte of the operator
 				 * in the source script. */
     int opBytes;		/* Number of bytes in the operator. */
-    char *src;			/* Points to first byte of the subexpression
+    CONST char *src;		/* Points to first byte of the subexpression
 				 * in the source script. */
     int srcBytes;		/* Number of bytes in subexpression's
 				 * source. */
@@ -1830,23 +2052,32 @@ PrependSubExprTokens(op, opBytes, src, srcBytes, firstIndex, infoPtr)
  *
  * Side effects:
  *	Sets the interpreter result to an error message describing the
- *	expression that was being parsed when the error occurred.
+ *	expression that was being parsed when the error occurred, and why
+ *	the parser considers that to be a syntax error at all.
  *
  *----------------------------------------------------------------------
  */
 
 static void
-LogSyntaxError(infoPtr)
+LogSyntaxError(infoPtr, extraInfo)
     ParseInfo *infoPtr;		/* Holds the parse state for the
 				 * expression being parsed. */
+    CONST char *extraInfo;	/* String to provide extra information
+				 * about the syntax error. */
 {
     int numBytes = (infoPtr->lastChar - infoPtr->originalExpr);
     char buffer[100];
 
-    sprintf(buffer, "syntax error in expression \"%.*s\"",
-	    ((numBytes > 60)? 60 : numBytes), infoPtr->originalExpr);
+    if (numBytes > 60) {
+	sprintf(buffer, "syntax error in expression \"%.60s...\"",
+		infoPtr->originalExpr);
+    } else {
+	sprintf(buffer, "syntax error in expression \"%.*s\"",
+		numBytes, infoPtr->originalExpr);
+    }
+    Tcl_ResetResult(infoPtr->parsePtr->interp);
     Tcl_AppendStringsToObj(Tcl_GetObjResult(infoPtr->parsePtr->interp),
-	    buffer, (char *) NULL);
+	    buffer, ": ", extraInfo, (char *) NULL);
     infoPtr->parsePtr->errorType = TCL_PARSE_SYNTAX;
     infoPtr->parsePtr->term = infoPtr->start;
 }
