@@ -9,8 +9,6 @@
  *
  * See the file "license.terms" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
- *
- * RCS: @(#) $Id: tclPipe.c,v 1.3 1999/04/16 00:46:51 stanton Exp $
  */
 
 #include "tclInt.h"
@@ -39,8 +37,9 @@ TCL_DECLARE_MUTEX(pipeMutex)		/* Guard access to detList. */
  */
 
 static TclFile	FileForRedirect _ANSI_ARGS_((Tcl_Interp *interp,
-	            char *spec, int atOk, char *arg, char *nextArg, 
-		    int flags, int *skipPtr, int *closePtr, int *releasePtr));
+	            CONST char *spec, int atOk, CONST char *arg, 
+		    CONST char *nextArg, int flags, int *skipPtr,
+		    int *closePtr, int *releasePtr));
 
 /*
  *----------------------------------------------------------------------
@@ -67,14 +66,14 @@ static TclFile
 FileForRedirect(interp, spec, atOK, arg, nextArg, flags, skipPtr, closePtr,
 	releasePtr)
     Tcl_Interp *interp;		/* Intepreter to use for error reporting. */
-    char *spec;			/* Points to character just after
+    CONST char *spec;		/* Points to character just after
 				 * redirection character. */
-    char *arg;			/* Pointer to entire argument containing 
-				 * spec:  used for error reporting. */
     int atOK;			/* Non-zero means that '@' notation can be 
 				 * used to specify a channel, zero means that
 				 * it isn't. */
-    char *nextArg;		/* Next argument in argc/argv array, if needed 
+    CONST char *arg;		/* Pointer to entire argument containing 
+				 * spec:  used for error reporting. */
+    CONST char *nextArg;	/* Next argument in argc/argv array, if needed 
 				 * for file name or channel name.  May be 
 				 * NULL. */
     int flags;			/* Flags to use for opening file or to 
@@ -106,9 +105,9 @@ FileForRedirect(interp, spec, atOK, arg, nextArg, flags, skipPtr, closePtr,
         }
 	file = TclpMakeFile(chan, writing ? TCL_WRITABLE : TCL_READABLE);
         if (file == NULL) {
-            Tcl_AppendResult(interp, "channel \"", Tcl_GetChannelName(chan),
-                    "\" wasn't opened for ",
-                    ((writing) ? "writing" : "reading"), (char *) NULL);
+	    Tcl_AppendResult(interp, "channel \"", Tcl_GetChannelName(chan),
+		    "\" wasn't opened for ",
+		    ((writing) ? "writing" : "reading"), (char *) NULL);
             return NULL;
         }
 	*releasePtr = 1;
@@ -123,7 +122,7 @@ FileForRedirect(interp, spec, atOK, arg, nextArg, flags, skipPtr, closePtr,
             Tcl_Flush(chan);
 	}
     } else {
-	char *name;
+	CONST char *name;
 	Tcl_DString nameString;
 
 	if (*spec == '\0') {
@@ -134,11 +133,10 @@ FileForRedirect(interp, spec, atOK, arg, nextArg, flags, skipPtr, closePtr,
 	    *skipPtr = 2;
 	}
 	name = Tcl_TranslateFileName(interp, spec, &nameString);
-	if (name != NULL) {
-	    file = TclpOpenFile(name, flags);
-	} else {
-	    file = NULL;
+	if (name == NULL) {
+	    return NULL;
 	}
+	file = TclpOpenFile(name, flags);
 	Tcl_DStringFree(&nameString);
 	if (file == NULL) {
 	    Tcl_AppendResult(interp, "couldn't ",
@@ -278,10 +276,18 @@ TclCleanupChildren(interp, numPids, pidPtr, errorChan)
     int i, abnormalExit, anyErrorInfo;
     Tcl_Pid pid;
     WAIT_STATUS_TYPE waitStatus;
-    char *msg;
+    CONST char *msg;
+    unsigned long resolvedPid;
 
     abnormalExit = 0;
     for (i = 0; i < numPids; i++) {
+	/*
+	 * We need to get the resolved pid before we wait on it as
+	 * the windows implementation of Tcl_WaitPid deletes the
+	 * information such that any following calls to TclpGetPid
+	 * fail.
+	 */
+	resolvedPid = TclpGetPid(pidPtr[i]);
         pid = Tcl_WaitPid(pidPtr[i], (int *) &waitStatus, 0);
 	if (pid == (Tcl_Pid) -1) {
 	    result = TCL_ERROR;
@@ -314,7 +320,7 @@ TclCleanupChildren(interp, numPids, pidPtr, errorChan)
 	    char msg1[TCL_INTEGER_SPACE], msg2[TCL_INTEGER_SPACE];
 
 	    result = TCL_ERROR;
-	    TclFormatInt(msg1, (long) TclpGetPid(pid));
+	    TclFormatInt(msg1, (long) resolvedPid);
 	    if (WIFEXITED(waitStatus)) {
                 if (interp != (Tcl_Interp *) NULL) {
 		    TclFormatInt(msg2, WEXITSTATUS(waitStatus));
@@ -324,7 +330,7 @@ TclCleanupChildren(interp, numPids, pidPtr, errorChan)
 		abnormalExit = 1;
 	    } else if (WIFSIGNALED(waitStatus)) {
                 if (interp != (Tcl_Interp *) NULL) {
-                    char *p;
+                    CONST char *p;
                     
                     p = Tcl_SignalMsg((int) (WTERMSIG(waitStatus)));
                     Tcl_SetErrorCode(interp, "CHILDKILLED", msg1,
@@ -335,7 +341,7 @@ TclCleanupChildren(interp, numPids, pidPtr, errorChan)
                 }
 	    } else if (WIFSTOPPED(waitStatus)) {
                 if (interp != (Tcl_Interp *) NULL) {
-                    char *p;
+                    CONST char *p;
 
                     p = Tcl_SignalMsg((int) (WSTOPSIG(waitStatus)));
                     Tcl_SetErrorCode(interp, "CHILDSUSP", msg1,
@@ -371,7 +377,7 @@ TclCleanupChildren(interp, numPids, pidPtr, errorChan)
 	    int count;
 	    Tcl_Obj *objPtr;
 	    
-	    Tcl_Seek(errorChan, 0L, SEEK_SET);
+	    Tcl_Seek(errorChan, (Tcl_WideInt)0, SEEK_SET);
 	    objPtr = Tcl_NewObj();
 	    count = Tcl_ReadChars(errorChan, objPtr, -1, 0);
 	    if (count < 0) {
@@ -439,7 +445,7 @@ TclCreatePipeline(interp, argc, argv, pidArrayPtr, inPipePtr,
 	outPipePtr, errFilePtr)
     Tcl_Interp *interp;		/* Interpreter to use for error reporting. */
     int argc;			/* Number of entries in argv. */
-    char **argv;		/* Array of strings describing commands in
+    CONST char **argv;		/* Array of strings describing commands in
 				 * pipeline plus I/O redirection with <,
 				 * <<,  >, etc.  Argv[argc] must be NULL. */
     Tcl_Pid **pidArrayPtr;	/* Word at *pidArrayPtr gets filled in with
@@ -476,7 +482,7 @@ TclCreatePipeline(interp, argc, argv, pidArrayPtr, inPipePtr,
 				 * at *pidPtr right now. */
     int cmdCount;		/* Count of number of distinct commands
 				 * found in argc/argv. */
-    char *inputLiteral = NULL;	/* If non-null, then this points to a
+    CONST char *inputLiteral = NULL;	/* If non-null, then this points to a
 				 * string containing input data (specified
 				 * via <<) to be piped to the first process
 				 * in the pipeline. */
@@ -498,8 +504,9 @@ TclCreatePipeline(interp, argc, argv, pidArrayPtr, inPipePtr,
     int errorClose = 0;		/* If non-zero, then errorFile should be 
     				 * closed when cleaning up. */
     int errorRelease = 0;
-    char *p;
-    int skip, lastBar, lastArg, i, j, atOK, flags, errorToOutput;
+    CONST char *p;
+    CONST char *nextArg;
+    int skip, lastBar, lastArg, i, j, atOK, flags, needCmd, errorToOutput = 0;
     Tcl_DString execBuffer;
     TclFile pipeIn;
     TclFile curInFile, curOutFile, curErrFile;
@@ -537,8 +544,10 @@ TclCreatePipeline(interp, argc, argv, pidArrayPtr, inPipePtr,
 
     lastBar = -1;
     cmdCount = 1;
+    needCmd = 1;
     for (i = 0; i < argc; i++) {
-        skip = 0;
+	errorToOutput = 0;
+	skip = 0;
 	p = argv[i];
 	switch (*p++) {
 	case '|':
@@ -555,6 +564,7 @@ TclCreatePipeline(interp, argc, argv, pidArrayPtr, inPipePtr,
 	    }
 	    lastBar = i;
 	    cmdCount++;
+	    needCmd = 1;
 	    break;
 
 	case '<':
@@ -571,7 +581,7 @@ TclCreatePipeline(interp, argc, argv, pidArrayPtr, inPipePtr,
 		inputLiteral = p + 1;
 		skip = 1;
 		if (*inputLiteral == '\0') {
-		    inputLiteral = argv[i + 1];
+		    inputLiteral = ((i + 1) == argc) ? NULL : argv[i + 1];
 		    if (inputLiteral == NULL) {
 			Tcl_AppendResult(interp, "can't specify \"", argv[i],
 				"\" as last word in command", (char *) NULL);
@@ -580,9 +590,10 @@ TclCreatePipeline(interp, argc, argv, pidArrayPtr, inPipePtr,
 		    skip = 2;
 		}
 	    } else {
+		nextArg = ((i + 1) == argc) ? NULL : argv[i + 1];
 		inputLiteral = NULL;
 		inputFile = FileForRedirect(interp, p, 1, argv[i], 
-			argv[i + 1], O_RDONLY, &skip, &inputClose, &inputRelease);
+			nextArg, O_RDONLY, &skip, &inputClose, &inputRelease);
 		if (inputFile == NULL) {
 		    goto error;
 		}
@@ -592,11 +603,16 @@ TclCreatePipeline(interp, argc, argv, pidArrayPtr, inPipePtr,
 	case '>':
 	    atOK = 1;
 	    flags = O_WRONLY | O_CREAT | O_TRUNC;
-	    errorToOutput = 0;
 	    if (*p == '>') {
 		p++;
 		atOK = 0;
-		flags = O_WRONLY | O_CREAT;
+
+		/*
+		 * Note that the O_APPEND flag only has an effect on POSIX
+		 * platforms. On Windows, we just have to carry on regardless.
+		 */
+
+		flags = O_WRONLY | O_CREAT | O_APPEND;
 	    }
 	    if (*p == '&') {
 		if (errorClose != 0) {
@@ -628,8 +644,9 @@ TclCreatePipeline(interp, argc, argv, pidArrayPtr, inPipePtr,
 		    TclpReleaseFile(outputFile);
 		}
 	    }
+	    nextArg = ((i + 1) == argc) ? NULL : argv[i + 1];
 	    outputFile = FileForRedirect(interp, p, atOK, argv[i], 
-		    argv[i + 1], flags, &skip, &outputClose, &outputRelease);
+		    nextArg, flags, &skip, &outputClose, &outputRelease);
 	    if (outputFile == NULL) {
 		goto error;
 	    }
@@ -666,12 +683,34 @@ TclCreatePipeline(interp, argc, argv, pidArrayPtr, inPipePtr,
 		errorRelease = 0;
 		TclpReleaseFile(errorFile);
 	    }
-	    errorFile = FileForRedirect(interp, p, atOK, argv[i], 
-		    argv[i + 1], flags, &skip, &errorClose, &errorRelease);
-	    if (errorFile == NULL) {
-		goto error;
+	    if (atOK && p[0] == '@' && p[1] == '1' && p[2] == '\0') {
+		/*
+		 * Special case handling of 2>@1 to redirect stderr to the
+		 * exec/open output pipe as well.  This is meant for the end
+		 * of the command string, otherwise use |& between commands.
+		 */
+		if (i != argc - 1) {
+		    Tcl_AppendResult(interp, "must specify \"", argv[i],
+			    "\" as last word in command", (char *) NULL);
+		    goto error;
+		}
+		errorFile = outputFile;
+		errorToOutput = 2;
+		skip = 1;
+	    } else {
+		nextArg = ((i + 1) == argc) ? NULL : argv[i + 1];
+		errorFile = FileForRedirect(interp, p, atOK, argv[i], 
+			nextArg, flags, &skip, &errorClose, &errorRelease);
+		if (errorFile == NULL) {
+		    goto error;
+		}
 	    }
 	    break;
+
+	default:
+	  /* Got a command word, not a redirection */
+	  needCmd = 0;
+	  break;
 	}
 
 	if (skip != 0) {
@@ -683,6 +722,15 @@ TclCreatePipeline(interp, argc, argv, pidArrayPtr, inPipePtr,
 	}
     }
 
+    if (needCmd) {
+        /* We had a bar followed only by redirections. */
+
+        Tcl_SetResult(interp,
+		      "illegal use of | or |& in command",
+		      TCL_STATIC);
+	goto error;
+    }
+
     if (inputFile == NULL) {
 	if (inputLiteral != NULL) {
 	    /*
@@ -756,7 +804,12 @@ TclCreatePipeline(interp, argc, argv, pidArrayPtr, inPipePtr,
     }
 
     if (errorFile == NULL) {
-	if (errFilePtr != NULL) {
+	if (errorToOutput == 2) {
+	    /*
+	     * Handle 2>@1 special case at end of cmd line
+	     */
+	    errorFile = outputFile;
+	} else if (errFilePtr != NULL) {
 	    /*
 	     * Set up the standard error output sink for the pipeline, if
 	     * requested.  Use a temporary file which is opened, then deleted.
@@ -802,7 +855,7 @@ TclCreatePipeline(interp, argc, argv, pidArrayPtr, inPipePtr,
     for (i = 0; i < argc; i = lastArg + 1) { 
 	int result, joinThisError;
 	Tcl_Pid pid;
-	char *oldName;
+	CONST char *oldName;
 
 	/*
 	 * Convert the program name into native form. 
@@ -828,7 +881,6 @@ TclCreatePipeline(interp, argc, argv, pidArrayPtr, inPipePtr,
 		}
 	    }
 	}
-	argv[lastArg] = NULL;
 
 	/*
 	 * If this is the last segment, use the specified outputFile.
@@ -836,9 +888,10 @@ TclCreatePipeline(interp, argc, argv, pidArrayPtr, inPipePtr,
 	 * curInFile for the next segment of the pipe.
 	 */
 
-	if (lastArg == argc) { 
+	if (lastArg == argc) {
 	    curOutFile = outputFile;
 	} else {
+	    argv[lastArg] = NULL;
 	    if (TclpCreatePipe(&pipeIn, &curOutFile) == 0) {
 		Tcl_AppendResult(interp, "couldn't create pipe: ",
 			Tcl_PosixError(interp), (char *) NULL);
@@ -992,7 +1045,7 @@ Tcl_OpenCommandChannel(interp, argc, argv, flags)
     Tcl_Interp *interp;		/* Interpreter for error reporting. Can
                                  * NOT be NULL. */
     int argc;			/* How many arguments. */
-    char **argv;		/* Array of arguments for command pipe. */
+    CONST char **argv;		/* Array of arguments for command pipe. */
     int flags;			/* Or'ed combination of TCL_STDIN, TCL_STDOUT,
 				 * TCL_STDERR, and TCL_ENFORCE_MODE. */
 {
