@@ -10,8 +10,6 @@
  *
  * See the file "license.terms" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
- *
- * RCS: @(#) $Id: tclPosixStr.c,v 1.3 1999/02/02 22:27:16 stanton Exp $
  */
 
 #include "tclInt.h"
@@ -35,11 +33,11 @@
  *----------------------------------------------------------------------
  */
 
-char *
+CONST char *
 Tcl_ErrnoId()
 {
     switch (errno) {
-#ifdef E2BIG
+#if defined(E2BIG) && (!defined(EOVERFLOW) || (E2BIG != EOVERFLOW))
 	case E2BIG: return "E2BIG";
 #endif
 #ifdef EACCES
@@ -63,7 +61,7 @@ Tcl_ErrnoId()
 #ifdef EALIGN
 	case EALIGN: return "EALIGN";
 #endif
-#if defined(EALREADY) && (!defined(EBUSY) || (EALREADY != EBUSY ))
+#if defined(EALREADY) && (!defined(EBUSY) || (EALREADY != EBUSY))
 	case EALREADY: return "EALREADY";
 #endif
 #ifdef EBADE
@@ -273,7 +271,7 @@ Tcl_ErrnoId()
 #ifdef ENOLCK
 	case ENOLCK: return "ENOLCK";
 #endif
-#ifdef ENOLINK
+#if defined(ENOLINK) && (!defined(ESOCKTNOSUPPORT) || (ESOCKTNOSUPPORT != ENOLINK))
 	case ENOLINK: return "ENOLINK";
 #endif
 #ifdef ENOMEM
@@ -288,7 +286,7 @@ Tcl_ErrnoId()
 #ifdef ENOPKG
 	case ENOPKG: return "ENOPKG";
 #endif
-#ifdef ENOPROTOOPT
+#if defined(ENOPROTOOPT) && (!defined(EPFNOSUPPORT) || (EPFNOSUPPORT != ENOPROTOOPT))
 	case ENOPROTOOPT: return "ENOPROTOOPT";
 #endif
 #ifdef ENOSPC
@@ -336,9 +334,12 @@ Tcl_ErrnoId()
 #ifdef ENXIO
 	case ENXIO: return "ENXIO";
 #endif
-#if defined(EOPNOTSUPP) && (!defined(ENOTSUP) || (ENOTSUP != EOPNOTSUPP))
+#if defined(EOPNOTSUPP) &&  (!defined(ENOTSUP) || (ENOTSUP != EOPNOTSUPP))
 	case EOPNOTSUPP: return "EOPNOTSUPP";
 #endif
+#if defined(EOVERFLOW) && (!defined(EFBIG) || (EOVERFLOW != EFBIG)) && (!defined(EINVAL) || (EOVERFLOW != EINVAL))
+        case EOVERFLOW: return "EOVERFLOW";
+#endif
 #ifdef EPERM
 	case EPERM: return "EPERM";
 #endif
@@ -480,12 +481,12 @@ Tcl_ErrnoId()
  *----------------------------------------------------------------------
  */
 
-char *
+CONST char *
 Tcl_ErrnoMsg(err)
     int err;			/* Error number (such as in errno variable). */
 {
     switch (err) {
-#ifdef E2BIG
+#if defined(E2BIG) && (!defined(EOVERFLOW) || (E2BIG != EOVERFLOW))
 	case E2BIG: return "argument list too long";
 #endif
 #ifdef EACCES
@@ -509,7 +510,7 @@ Tcl_ErrnoMsg(err)
 #ifdef EALIGN
 	case EALIGN: return "EALIGN";
 #endif
-#if defined(EALREADY) && (!defined(EBUSY) || (EALREADY != EBUSY ))
+#if defined(EALREADY) && (!defined(EBUSY) || (EALREADY != EBUSY))
 	case EALREADY: return "operation already in progress";
 #endif
 #ifdef EBADE
@@ -720,8 +721,8 @@ Tcl_ErrnoMsg(err)
 #ifdef ENOLCK
 	case ENOLCK: return "no locks available";
 #endif
-#ifdef ENOLINK
-	case ENOLINK: return "link has be severed";
+#if defined(ENOLINK) && (!defined(ESOCKTNOSUPPORT) || (ESOCKTNOSUPPORT != ENOLINK))
+	case ENOLINK: return "link has been severed";
 #endif
 #ifdef ENOMEM
 	case ENOMEM: return "not enough memory";
@@ -735,8 +736,8 @@ Tcl_ErrnoMsg(err)
 #ifdef ENOPKG
 	case ENOPKG: return "package not installed";
 #endif
-#ifdef ENOPROTOOPT
-	case ENOPROTOOPT: return "bad proocol option";
+#if defined(ENOPROTOOPT) && (!defined(EPFNOSUPPORT) || (EPFNOSUPPORT != ENOPROTOOPT))
+	case ENOPROTOOPT: return "bad protocol option";
 #endif
 #ifdef ENOSPC
 	case ENOSPC: return "no space left on device";
@@ -783,9 +784,12 @@ Tcl_ErrnoMsg(err)
 #ifdef ENXIO
 	case ENXIO: return "no such device or address";
 #endif
-#if defined(EOPNOTSUPP) && (!defined(ENOTSUP) || (ENOTSUP != EOPNOTSUPP))
+#if defined(EOPNOTSUPP) &&  (!defined(ENOTSUP) || (ENOTSUP != EOPNOTSUPP))
 	case EOPNOTSUPP: return "operation not supported on socket";
 #endif
+#if defined(EOVERFLOW) && (!defined(EFBIG) || (EOVERFLOW != EFBIG)) && (!defined(EINVAL) || (EOVERFLOW != EINVAL))
+        case EOVERFLOW: return "file too big";
+#endif
 #ifdef EPERM
 	case EPERM: return "not owner";
 #endif
@@ -811,7 +815,7 @@ Tcl_ErrnoMsg(err)
 	case EPROTO: return "protocol error";
 #endif
 #ifdef EPROTONOSUPPORT
-	case EPROTONOSUPPORT: return "protocol not suppored";
+	case EPROTONOSUPPORT: return "protocol not supported";
 #endif
 #ifdef EPROTOTYPE
 	case EPROTOTYPE: return "protocol wrong type for socket";
@@ -847,7 +851,7 @@ Tcl_ErrnoMsg(err)
 	case ERREMOTE: return "object is remote";
 #endif
 #ifdef ESHUTDOWN
-	case ESHUTDOWN: return "can't send afer socket shutdown";
+	case ESHUTDOWN: return "can't send after socket shutdown";
 #endif
 #ifdef ESOCKTNOSUPPORT
 	case ESOCKTNOSUPPORT: return "socket type not supported";
@@ -904,7 +908,7 @@ Tcl_ErrnoMsg(err)
 #ifdef NO_STRERROR
 	    return "unknown POSIX error";
 #else
-	    return strerror(errno);
+	    return strerror(err);
 #endif
     }
 }
@@ -927,7 +931,7 @@ Tcl_ErrnoMsg(err)
  *----------------------------------------------------------------------
  */
 
-char *
+CONST char *
 Tcl_SignalId(sig)
     int sig;			/* Number of signal. */
 {
@@ -986,13 +990,13 @@ Tcl_SignalId(sig)
 #ifdef SIGPROF
 	case SIGPROF: return "SIGPROF";
 #endif
-#if defined(SIGPWR) && (!defined(SIGXFSZ) || (SIGPWR != SIGXFSZ))
+#if defined(SIGPWR) && (!defined(SIGXFSZ) || (SIGPWR != SIGXFSZ)) && (!defined(SIGLOST) || (SIGPWR != SIGLOST))
 	case SIGPWR: return "SIGPWR";
 #endif
 #ifdef SIGQUIT
 	case SIGQUIT: return "SIGQUIT";
 #endif
-#ifdef SIGSEGV
+#if defined(SIGSEGV) && (!defined(SIGBUS) || (SIGSEGV != SIGBUS))
 	case SIGSEGV: return "SIGSEGV";
 #endif
 #ifdef SIGSTOP
@@ -1059,7 +1063,7 @@ Tcl_SignalId(sig)
  *----------------------------------------------------------------------
  */
 
-char *
+CONST char *
 Tcl_SignalMsg(sig)
     int sig;			/* Number of signal. */
 {
@@ -1118,13 +1122,13 @@ Tcl_SignalMsg(sig)
 #ifdef SIGPROF
 	case SIGPROF: return "profiling alarm";
 #endif
-#if defined(SIGPWR) && (!defined(SIGXFSZ) || (SIGPWR != SIGXFSZ))
+#if defined(SIGPWR) && (!defined(SIGXFSZ) || (SIGPWR != SIGXFSZ)) && (!defined(SIGLOST) || (SIGPWR != SIGLOST))
 	case SIGPWR: return "power-fail restart";
 #endif
 #ifdef SIGQUIT
 	case SIGQUIT: return "quit signal";
 #endif
-#ifdef SIGSEGV
+#if defined(SIGSEGV) && (!defined(SIGBUS) || (SIGSEGV != SIGBUS))
 	case SIGSEGV: return "segmentation violation";
 #endif
 #ifdef SIGSTOP
