@@ -2,8 +2,6 @@
 #
 # Implementation of the history command.
 #
-# RCS: @(#) $Id: history.tcl,v 1.3 1998/09/14 18:40:03 stanton Exp $
-#
 # Copyright (c) 1997 Sun Microsystems, Inc.
 #
 # See the file "license.terms" for information on usage and redistribution
@@ -166,10 +164,16 @@ proc history {args} {
 
  proc tcl::HistAdd {command {exec {}}} {
     variable history
+
+    # Do not add empty commands to the history
+    if {[string trim $command] eq ""} {
+	return ""
+    }
+
     set i [incr history(nextid)]
     set history($i) $command
     set j [incr history(oldest)]
-    if {[info exists history($j)]} {unset history($j)}
+    unset -nocomplain history($j)
     if {[string match e* $exec]} {
 	return [uplevel #0 $command]
     } else {
@@ -192,13 +196,13 @@ proc history {args} {
 
  proc tcl::HistKeep {{limit {}}} {
     variable history
-    if {[string length $limit] == 0} {
+    if {$limit eq ""} {
 	return $history(keep)
     } else {
 	set oldold $history(oldest)
 	set history(oldest) [expr {$history(nextid) - $limit}]
 	for {} {$oldold <= $history(oldest)} {incr oldold} {
-	    if {[info exists history($oldold)]} {unset history($oldold)}
+	    unset -nocomplain history($oldold)
 	}
 	set history(keep) $limit
     }
@@ -240,7 +244,7 @@ proc history {args} {
 
  proc tcl::HistInfo {{num {}}} {
     variable history
-    if {$num == {}} {
+    if {$num eq ""} {
 	set num [expr {$history(keep) + 1}]
     }
     set result {}
@@ -250,8 +254,7 @@ proc history {args} {
 	if {![info exists history($i)]} {
 	    continue
 	}
-	set cmd [string trimright $history($i) \ \n]
-	regsub -all \n $cmd "\n\t" cmd
+	set cmd [string map [list \n \n\t] [string trimright $history($i) \ \n]]
 	append result $newline[format "%6d  %s" $i $cmd]
 	set newline \n
     }
@@ -275,7 +278,7 @@ proc history {args} {
 
  proc tcl::HistRedo {{event -1}} {
     variable history
-    if {[string length $event] == 0} {
+    if {$event eq ""} {
 	set event -1
     }
     set i [HistIndex $event]
@@ -305,7 +308,8 @@ proc history {args} {
  proc tcl::HistIndex {event} {
     variable history
     if {[catch {expr {~$event}}]} {
-	for {set i $history(nextid)} {[info exists history($i)]} {incr i -1} {
+	for {set i [expr {$history(nextid)-1}]} {[info exists history($i)]} \
+		{incr i -1} {
 	    if {[string match $event* $history($i)]} {
 		return $i;
 	    }
