@@ -7,13 +7,10 @@
  *
  * Copyright (c) 1995-1997 Sun Microsystems, Inc.
  * Copyright 2001, Apple Computer, Inc.
- * Copyright (c) 2005 Tcl Core Team.
- * Copyright (c) 2005-2006 Daniel A. Steffen <das@users.sourceforge.net>
+ * Copyright (c) 2005-2008 Daniel A. Steffen <das@users.sourceforge.net>
  *
  * See the file "license.terms" for information on usage and redistribution of
  * this file, and for a DISCLAIMER OF ALL WARRANTIES.
- *
- * RCS: @(#) $Id: tclMacOSXNotify.c,v 1.1.2.9 2006/08/21 03:50:53 das Exp $
  */
 
 #include "tclInt.h"
@@ -150,11 +147,12 @@ static int triggerPipe = -1;
 static int receivePipe = -1; /* Output end of triggerPipe */
 
 /*
- * We use Darwin-native spinlocks instead of pthread mutexes for notifier
- * locking: this radically simplifies the implementation and lowers overhead.
- * Note that these are not pure spinlocks, they employ various strategies to
- * back off, making them immune to most priority-inversion livelocks (c.f. man
- * 3 OSSpinLockLock).
+ * We use the Darwin-native spinlock API rather than pthread mutexes for
+ * notifier locking: this radically simplifies the implementation and lowers
+ * overhead. Note that these are not pure spinlocks, they employ various
+ * strategies to back off and relinquish the processor, making them immune to
+ * most priority-inversion livelocks (c.f. 'man 3 OSSpinLockLock' and Darwin
+ * sources: xnu/osfmk/{ppc,i386}/commpage/spinlocks.s).
  */
 
 #if defined(HAVE_LIBKERN_OSATOMIC_H) && defined(HAVE_OSSPINLOCKLOCK)
@@ -169,26 +167,42 @@ static int receivePipe = -1; /* Output end of triggerPipe */
  * Support for weakly importing spinlock API.
  */
 #define WEAK_IMPORT_SPINLOCKLOCK
-extern void	OSSpinLockLock(OSSpinLock *lock) WEAK_IMPORT_ATTRIBUTE;
-extern void	OSSpinLockUnlock(OSSpinLock *lock) WEAK_IMPORT_ATTRIBUTE;
-extern void	_spin_lock(OSSpinLock *lock) WEAK_IMPORT_ATTRIBUTE;
-extern void	_spin_unlock(OSSpinLock *lock) WEAK_IMPORT_ATTRIBUTE;
-static void (* lockLock)(OSSpinLock *lock) = NULL;
-static void (* lockUnlock)(OSSpinLock *lock) = NULL;
+#if MAC_OS_X_VERSION_MAX_ALLOWED >= 1050
+#define VOLATILE volatile
+#else
+#define VOLATILE
+#endif
+#ifndef bool
+#define bool int
+#endif
+extern void OSSpinLockLock(VOLATILE OSSpinLock *lock) WEAK_IMPORT_ATTRIBUTE;
+extern void OSSpinLockUnlock(VOLATILE OSSpinLock *lock) WEAK_IMPORT_ATTRIBUTE;
+extern bool OSSpinLockTry(VOLATILE OSSpinLock *lock) WEAK_IMPORT_ATTRIBUTE;
+extern void _spin_lock(VOLATILE OSSpinLock *lock) WEAK_IMPORT_ATTRIBUTE;
+extern void _spin_unlock(VOLATILE OSSpinLock *lock) WEAK_IMPORT_ATTRIBUTE;
+extern bool _spin_lock_try(VOLATILE OSSpinLock *lock) WEAK_IMPORT_ATTRIBUTE;
+static void (* lockLock)(VOLATILE OSSpinLock *lock) = NULL;
+static void (* lockUnlock)(VOLATILE OSSpinLock *lock) = NULL;
+static bool (* lockTry)(VOLATILE OSSpinLock *lock) = NULL;
+#undef VOLATILE
 static pthread_once_t spinLockLockInitControl = PTHREAD_ONCE_INIT;
 static void SpinLockLockInit(void) {
     lockLock   = OSSpinLockLock   != NULL ? OSSpinLockLock   : _spin_lock;
     lockUnlock = OSSpinLockUnlock != NULL ? OSSpinLockUnlock : _spin_unlock;
+    lockTry    = OSSpinLockTry    != NULL ? OSSpinLockTry    : _spin_lock_try;
     if (lockLock == NULL || lockUnlock == NULL) {
 	Tcl_Panic("SpinLockLockInit: no spinlock API available");
     }
 }
 #define SpinLockLock(p) 	lockLock(p)
 #define SpinLockUnlock(p)	lockUnlock(p)
+#define SpinLockTry(p)  	lockTry(p)
 #else
 #define SpinLockLock(p) 	OSSpinLockLock(p)
 #define SpinLockUnlock(p)	OSSpinLockUnlock(p)
+#define SpinLockTry(p)  	OSSpinLockTry(p)
 #endif /* HAVE_WEAK_IMPORT */
+#define SPINLOCK_INIT   	OS_SPINLOCK_INIT
 
 #else
 /*
@@ -196,10 +210,13 @@ static void SpinLockLockInit(void) {
  */
 
 typedef uint32_t OSSpinLock;
-extern void	_spin_lock(OSSpinLock *lock);
-extern void	_spin_unlock(OSSpinLock *lock);
+extern void _spin_lock(OSSpinLock *lock);
+extern void _spin_unlock(OSSpinLock *lock);
+extern int  _spin_lock_try(OSSpinLock *lock);
 #define SpinLockLock(p) 	_spin_lock(p)
 #define SpinLockUnlock(p)	_spin_unlock(p)
+#define SpinLockTry(p)  	_spin_lock_try(p)
+#define SPINLOCK_INIT   	0
 
 #endif /* HAVE_LIBKERN_OSATOMIC_H && HAVE_OSSPINLOCKLOCK */
 
@@ -207,8 +224,8 @@ extern void	_spin_unlock(OSSpinLock *lock);
  * These spinlocks lock access to the global notifier state.
  */
 
-static OSSpinLock notifierInitLock = 0;
-static OSSpinLock notifierLock = 0;
+static OSSpinLock notifierInitLock = SPINLOCK_INIT;
+static OSSpinLock notifierLock     = SPINLOCK_INIT;
 
 /*
  * Macros abstracting notifier locking/unlocking
@@ -255,7 +272,8 @@ static CFStringRef tclEventsOnlyRunLoopMode = NULL;
  * Static routines defined in this file.
  */
 
-static void	NotifierThreadProc(ClientData clientData);
+static void	NotifierThreadProc(ClientData clientData)
+	__attribute__ ((__noreturn__));
 static int	FileHandlerEventProc(Tcl_Event *evPtr, int flags);
 
 #ifdef HAVE_PTHREAD_ATFORK
@@ -269,6 +287,17 @@ static void	AtForkChild(void);
 extern int pthread_atfork(void (*prepare)(void), void (*parent)(void),
                           void (*child)(void)) WEAK_IMPORT_ATTRIBUTE;
 #endif /* HAVE_WEAK_IMPORT */
+/*
+ * On Darwin 9 and later, it is not possible to call CoreFoundation after
+ * a fork.
+ */
+#if !defined(MAC_OS_X_VERSION_MIN_REQUIRED) || \
+	MAC_OS_X_VERSION_MIN_REQUIRED < 1050
+MODULE_SCOPE long tclMacOSXDarwinRelease;
+#define noCFafterFork (tclMacOSXDarwinRelease >= 9)
+#else /* MAC_OS_X_VERSION_MIN_REQUIRED */
+#define noCFafterFork 1
+#endif /* MAC_OS_X_VERSION_MIN_REQUIRED */
 #endif /* HAVE_PTHREAD_ATFORK */
 
 /*
@@ -407,8 +436,8 @@ Tcl_InitNotifier(void)
  */
 
 void
-Tcl_FinalizeNotifier(clientData)
-    ClientData clientData;		/* Not used. */
+Tcl_FinalizeNotifier(
+    ClientData clientData)		/* Not used. */
 {
     ThreadSpecificData *tsdPtr = TCL_TSD_INIT(&dataKey);
 
@@ -489,8 +518,8 @@ Tcl_FinalizeNotifier(clientData)
  */
 
 void
-Tcl_AlertNotifier(clientData)
-    ClientData clientData;
+Tcl_AlertNotifier(
+    ClientData clientData)
 {
     ThreadSpecificData *tsdPtr = (ThreadSpecificData *) clientData;
 
@@ -522,8 +551,8 @@ Tcl_AlertNotifier(clientData)
  */
 
 void
-Tcl_SetTimer(timePtr)
-    Tcl_Time *timePtr;		/* Timeout value, may be NULL. */
+Tcl_SetTimer(
+    Tcl_Time *timePtr)		/* Timeout value, may be NULL. */
 {
     /*
      * The interval timer doesn't do anything in this implementation, because
@@ -553,8 +582,8 @@ Tcl_SetTimer(timePtr)
  */
 
 void
-Tcl_ServiceModeHook(mode)
-    int mode;			/* Either TCL_SERVICE_ALL, or
+Tcl_ServiceModeHook(
+    int mode)			/* Either TCL_SERVICE_ALL, or
 				 * TCL_SERVICE_NONE. */
 {
 }
@@ -576,15 +605,15 @@ Tcl_ServiceModeHook(mode)
  */
 
 void
-Tcl_CreateFileHandler(fd, mask, proc, clientData)
-    int fd;			/* Handle of stream to watch. */
-    int mask;			/* OR'ed combination of TCL_READABLE,
+Tcl_CreateFileHandler(
+    int fd,			/* Handle of stream to watch. */
+    int mask,			/* OR'ed combination of TCL_READABLE,
 				 * TCL_WRITABLE, and TCL_EXCEPTION: indicates
 				 * conditions under which proc should be
 				 * called. */
-    Tcl_FileProc *proc;		/* Function to call for each selected
+    Tcl_FileProc *proc,		/* Function to call for each selected
 				 * event. */
-    ClientData clientData;	/* Arbitrary data to pass to proc. */
+    ClientData clientData)	/* Arbitrary data to pass to proc. */
 {
     ThreadSpecificData *tsdPtr = TCL_TSD_INIT(&dataKey);
     FileHandler *filePtr;
@@ -653,8 +682,8 @@ Tcl_CreateFileHandler(fd, mask, proc, clientData)
  */
 
 void
-Tcl_DeleteFileHandler(fd)
-    int fd;			/* Stream id for which to remove callback
+Tcl_DeleteFileHandler(
+    int fd)			/* Stream id for which to remove callback
 				 * function. */
 {
     FileHandler *filePtr, *prevPtr;
@@ -746,9 +775,9 @@ Tcl_DeleteFileHandler(fd)
  */
 
 static int
-FileHandlerEventProc(evPtr, flags)
-    Tcl_Event *evPtr;		/* Event to service. */
-    int flags;			/* Flags that indicate what events to handle,
+FileHandlerEventProc(
+    Tcl_Event *evPtr,		/* Event to service. */
+    int flags)			/* Flags that indicate what events to handle,
 				 * such as TCL_FILE_EVENTS. */
 {
     int mask;
@@ -815,8 +844,8 @@ FileHandlerEventProc(evPtr, flags)
  */
 
 int
-Tcl_WaitForEvent(timePtr)
-    Tcl_Time *timePtr;		/* Maximum block time, or NULL. */
+Tcl_WaitForEvent(
+    Tcl_Time *timePtr)		/* Maximum block time, or NULL. */
 {
     FileHandler *filePtr;
     FileHandlerEvent *fileEvPtr;
@@ -833,6 +862,9 @@ Tcl_WaitForEvent(timePtr)
      */
 
     LOCK_NOTIFIER_INIT;
+    if (!notifierCount) {
+        Tcl_Panic("Tcl_WaitForEvent: notifier not initialized");
+    }
     if (!notifierThread) {
 	int result;
 	pthread_attr_t attr;
@@ -856,7 +888,9 @@ Tcl_WaitForEvent(timePtr)
      */
 
     LOCK_NOTIFIER;
-
+    if (!tsdPtr->runLoop) {
+        Tcl_Panic("Tcl_WaitForEvent: CFRunLoop not initialized");
+    }
     waitForFiles = (tsdPtr->numFdBits > 0);
     if (timePtr != NULL && timePtr->sec == 0 && timePtr->usec == 0) {
 	/*
@@ -1008,8 +1042,8 @@ Tcl_WaitForEvent(timePtr)
  */
 
 static void
-NotifierThreadProc(clientData)
-    ClientData clientData;	/* Not used. */
+NotifierThreadProc(
+    ClientData clientData)	/* Not used. */
 {
     ThreadSpecificData *tsdPtr;
     fd_set readableMask;
@@ -1157,7 +1191,7 @@ NotifierThreadProc(clientData)
 	    }
 	}
     }
-    pthread_exit (0);
+    pthread_exit(0);
 }
 
 #ifdef HAVE_PTHREAD_ATFORK
@@ -1232,7 +1266,9 @@ AtForkChild(void)
     UNLOCK_NOTIFIER_INIT;
     if (tsdPtr->runLoop) {
 	tsdPtr->runLoop = NULL;
-	CFRunLoopSourceInvalidate(tsdPtr->runLoopSource);
+	if (!noCFafterFork) {
+	    CFRunLoopSourceInvalidate(tsdPtr->runLoopSource);
+	}
 	CFRelease(tsdPtr->runLoopSource);
 	tsdPtr->runLoopSource = NULL;
     }
@@ -1248,7 +1284,9 @@ AtForkChild(void)
 	 * Tcl_AlertNotifier may break in the child.
 	 */
 
-	Tcl_InitNotifier();
+	if (!noCFafterFork) {
+	    Tcl_InitNotifier();
+	}
     }
 }
 #endif /* HAVE_PTHREAD_ATFORK */
