@@ -9,8 +9,6 @@
  *
  * See the file "license.terms" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
- *
- * RCS: @(#) $Id: tclLoadShl.c,v 1.3.12.1 2001/09/12 21:26:36 dgp Exp $
  */
 
 #include <dl.h>
@@ -28,17 +26,14 @@
 /*
  *----------------------------------------------------------------------
  *
- * TclpLoadFile --
+ * TclpDlopen --
  *
  *	Dynamically loads a binary code file into memory and returns
- *	the addresses of two procedures within that file, if they
- *	are defined.
+ *	a handle to the new code.
  *
  * Results:
  *	A standard Tcl completion code.  If an error occurs, an error
- *	message is left in the interp's result.  *proc1Ptr and *proc2Ptr
- *	are filled in with the addresses of the symbols given by
- *	*sym1 and *sym2, or NULL if those symbols can't be found.
+ *	message is left in the interp's result.
  *
  * Side effects:
  *	New code suddenly appears in memory.
@@ -47,21 +42,21 @@
  */
 
 int
-TclpLoadFile(interp, fileName, sym1, sym2, proc1Ptr, proc2Ptr, clientDataPtr)
+TclpDlopen(interp, pathPtr, loadHandle, unloadProcPtr)
     Tcl_Interp *interp;		/* Used for error reporting. */
-    char *fileName;		/* Name of the file containing the desired
-				 * code. */
-    char *sym1, *sym2;		/* Names of two procedures to look up in
-				 * the file's symbol table. */
-    Tcl_PackageInitProc **proc1Ptr, **proc2Ptr;
-				/* Where to return the addresses corresponding
-				 * to sym1 and sym2. */
-    ClientData *clientDataPtr;	/* Filled with token for dynamically loaded
+    Tcl_Obj *pathPtr;		/* Name of the file containing the desired
+				 * code (UTF-8). */
+    Tcl_LoadHandle *loadHandle;	/* Filled with token for dynamically loaded
 				 * file which will be passed back to 
-				 * TclpUnloadFile() to unload the file. */
+				 * (*unloadProcPtr)() to unload the file. */
+    Tcl_FSUnloadFileProc **unloadProcPtr;	
+				/* Filled with address of Tcl_FSUnloadFileProc
+				 * function which should be used for
+				 * this file. */
 {
     shl_t handle;
-    Tcl_DString newName;
+    CONST char *native;
+    char *fileName = Tcl_GetString(pathPtr);
 
     /*
      * The flags below used to be BIND_IMMEDIATE; they were changed at
@@ -73,43 +68,80 @@ TclpLoadFile(interp, fileName, sym1, sym2, proc1Ptr, proc2Ptr, clientDataPtr)
      * when they are build."
      */
 
-    handle = shl_load(fileName, BIND_DEFERRED|BIND_VERBOSE, 0L);
+
+    /*
+     * First try the full path the user gave us.  This is particularly
+     * important if the cwd is inside a vfs, and we are trying to load
+     * using a relative path.
+     */
+    native = Tcl_FSGetNativePath(pathPtr);
+    handle = shl_load(native, BIND_DEFERRED|BIND_VERBOSE, 0L);
+
+    if (handle == NULL) {
+	/*
+	 * Let the OS loader examine the binary search path for
+	 * whatever string the user gave us which hopefully refers
+	 * to a file on the binary path
+	 */
+	Tcl_DString ds;
+	native = Tcl_UtfToExternalDString(NULL, fileName, -1, &ds);
+	handle = shl_load(native,
+			  BIND_DEFERRED|BIND_VERBOSE|DYNAMIC_PATH, 0L);
+	Tcl_DStringFree(&ds);
+    }
+
     if (handle == NULL) {
 	Tcl_AppendResult(interp, "couldn't load file \"", fileName,
 		"\": ", Tcl_PosixError(interp), (char *) NULL);
 	return TCL_ERROR;
     }
-    *clientDataPtr = (ClientData) handle;
-
+    *loadHandle = (Tcl_LoadHandle) handle;
+    *unloadProcPtr = &TclpUnloadFile;
+    return TCL_OK;
+}
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * TclpFindSymbol --
+ *
+ *	Looks up a symbol, by name, through a handle associated with
+ *	a previously loaded piece of code (shared library).
+ *
+ * Results:
+ *	Returns a pointer to the function associated with 'symbol' if
+ *	it is found.  Otherwise returns NULL and may leave an error
+ *	message in the interp's result.
+ *
+ *----------------------------------------------------------------------
+ */
+Tcl_PackageInitProc*
+TclpFindSymbol(interp, loadHandle, symbol) 
+    Tcl_Interp *interp;
+    Tcl_LoadHandle loadHandle;
+    CONST char *symbol;
+{
+    Tcl_DString newName;
+    Tcl_PackageInitProc *proc=NULL;
+    shl_t handle = (shl_t)loadHandle;
     /*
      * Some versions of the HP system software still use "_" at the
      * beginning of exported symbols while others don't;  try both
      * forms of each name.
      */
 
-    if (shl_findsym(&handle, sym1, (short) TYPE_PROCEDURE, (void *) proc1Ptr)
+    if (shl_findsym(&handle, symbol, (short) TYPE_PROCEDURE, (void *) &proc)
 	    != 0) {
 	Tcl_DStringInit(&newName);
 	Tcl_DStringAppend(&newName, "_", 1);
-	Tcl_DStringAppend(&newName, sym1, -1);
+	Tcl_DStringAppend(&newName, symbol, -1);
 	if (shl_findsym(&handle, Tcl_DStringValue(&newName),
-		(short) TYPE_PROCEDURE, (void *) proc1Ptr) != 0) {
-	    *proc1Ptr = NULL;
+		(short) TYPE_PROCEDURE, (void *) &proc) != 0) {
+	    proc = NULL;
 	}
 	Tcl_DStringFree(&newName);
     }
-    if (shl_findsym(&handle, sym2, (short) TYPE_PROCEDURE, (void *) proc2Ptr)
-	    != 0) {
-	Tcl_DStringInit(&newName);
-	Tcl_DStringAppend(&newName, "_", 1);
-	Tcl_DStringAppend(&newName, sym2, -1);
-	if (shl_findsym(&handle, Tcl_DStringValue(&newName),
-		(short) TYPE_PROCEDURE, (void *) proc2Ptr) != 0) {
-	    *proc2Ptr = NULL;
-	}
-	Tcl_DStringFree(&newName);
-    }
-    return TCL_OK;
+    return proc;
 }
 
 /*
@@ -131,15 +163,15 @@ TclpLoadFile(interp, fileName, sym1, sym2, proc1Ptr, proc2Ptr, clientDataPtr)
  */
 
 void
-TclpUnloadFile(clientData)
-    ClientData clientData;	/* ClientData returned by a previous call
-				 * to TclpLoadFile().  The clientData is 
+TclpUnloadFile(loadHandle)
+    Tcl_LoadHandle loadHandle;	/* loadHandle returned by a previous call
+				 * to TclpDlopen().  The loadHandle is 
 				 * a token that represents the loaded 
 				 * file. */
 {
     shl_t handle;
 
-    handle = (shl_t) clientData;
+    handle = (shl_t) loadHandle;
     shl_unload(handle);
 }
 
@@ -165,7 +197,7 @@ TclpUnloadFile(clientData)
 
 int
 TclGuessPackageName(fileName, bufPtr)
-    char *fileName;		/* Name of file containing package (already
+    CONST char *fileName;	/* Name of file containing package (already
 				 * translated to local form if needed). */
     Tcl_DString *bufPtr;	/* Initialized empty dstring.  Append
 				 * package name to this if possible. */
