@@ -8,8 +8,6 @@
  *
  * See the file "license.terms" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
- *
- * RCS: @(#) $Id: tclXtNotify.c,v 1.2 1998/09/14 18:40:18 stanton Exp $
  */
 
 #include <X11/Intrinsic.h>
@@ -79,11 +77,16 @@ static int		FileHandlerEventProc _ANSI_ARGS_((Tcl_Event *evPtr,
 			    int flags));
 static void		FileProc _ANSI_ARGS_((caddr_t clientData,
 			    int *source, XtInputId *id));
-static void		InitNotifier _ANSI_ARGS_((void));
+void			InitNotifier _ANSI_ARGS_((void));
 static void		NotifierExitHandler _ANSI_ARGS_((
 			    ClientData clientData));
 static void		TimerProc _ANSI_ARGS_((caddr_t clientData,
 			    XtIntervalId *id));
+static void		CreateFileHandler _ANSI_ARGS_((int fd, int mask, 
+				Tcl_FileProc * proc, ClientData clientData));
+static void		DeleteFileHandler _ANSI_ARGS_((int fd));
+static void		SetTimer _ANSI_ARGS_((Tcl_Time * timePtr));
+static int		WaitForEvent _ANSI_ARGS_((Tcl_Time * timePtr));
 
 /*
  * Functions defined in this file for use by users of the Xt Notifier:
@@ -180,9 +183,10 @@ TclSetAppContext(appContext)
  *----------------------------------------------------------------------
  */
 
-static void
-InitNotifier(void)
+void
+InitNotifier()
 {
+    Tcl_NotifierProcs notifier;
     /*
      * Only reinitialize if we are not in exit handling. The notifier
      * can get reinitialized after its own exit handler has run, because
@@ -193,6 +197,12 @@ InitNotifier(void)
         return;
     }
 
+    notifier.createFileHandlerProc = CreateFileHandler;
+    notifier.deleteFileHandlerProc = DeleteFileHandler;
+    notifier.setTimerProc = SetTimer;
+    notifier.waitForEventProc = WaitForEvent;
+    Tcl_SetNotifier(&notifier);
+
     /*
      * DO NOT create the application context yet; doing so would prevent
      * external applications from setting it for us to their own ones.
@@ -241,7 +251,7 @@ NotifierExitHandler(
 /*
  *----------------------------------------------------------------------
  *
- * Tcl_SetTimer --
+ * SetTimer --
  *
  *	This procedure sets the current notifier timeout value.
  *
@@ -254,8 +264,8 @@ NotifierExitHandler(
  *----------------------------------------------------------------------
  */
 
-void
-Tcl_SetTimer(timePtr)
+static void
+SetTimer(timePtr)
     Tcl_Time *timePtr;		/* Timeout value, may be NULL. */
 {
     long timeout;
@@ -311,7 +321,7 @@ TimerProc(data, id)
 /*
  *----------------------------------------------------------------------
  *
- * Tcl_CreateFileHandler --
+ * CreateFileHandler --
  *
  *	This procedure registers a file handler with the Xt notifier.
  *
@@ -325,8 +335,8 @@ TimerProc(data, id)
  *----------------------------------------------------------------------
  */
 
-void
-Tcl_CreateFileHandler(fd, mask, proc, clientData)
+static void
+CreateFileHandler(fd, mask, proc, clientData)
     int fd;			/* Handle of stream to watch. */
     int mask;			/* OR'ed combination of TCL_READABLE,
 				 * TCL_WRITABLE, and TCL_EXCEPTION:
@@ -407,7 +417,7 @@ Tcl_CreateFileHandler(fd, mask, proc, clientData)
 /*
  *----------------------------------------------------------------------
  *
- * Tcl_DeleteFileHandler --
+ * DeleteFileHandler --
  *
  *	Cancel a previously-arranged callback arrangement for
  *	a file.
@@ -421,8 +431,8 @@ Tcl_CreateFileHandler(fd, mask, proc, clientData)
  *----------------------------------------------------------------------
  */
 
-void
-Tcl_DeleteFileHandler(fd)
+static void
+DeleteFileHandler(fd)
     int fd;			/* Stream id for which to remove
 				 * callback procedure. */
 {
@@ -609,7 +619,7 @@ FileHandlerEventProc(evPtr, flags)
 /*
  *----------------------------------------------------------------------
  *
- * Tcl_WaitForEvent --
+ * WaitForEvent --
  *
  *	This function is called by Tcl_DoOneEvent to wait for new
  *	events on the message queue.  If the block time is 0, then
@@ -626,8 +636,8 @@ FileHandlerEventProc(evPtr, flags)
  *----------------------------------------------------------------------
  */
 
-int
-Tcl_WaitForEvent(
+static int
+WaitForEvent(
     Tcl_Time *timePtr)		/* Maximum block time, or NULL. */
 {
     int timeout;
