@@ -8,8 +8,6 @@
  *
  * See the file "license.terms" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
- *
- * RCS: @(#) $Id: tclWinChan.c,v 1.30.2.4 2005/06/23 15:05:22 kennykb Exp $
  */
 
 #include "tclWinInt.h"
@@ -100,8 +98,8 @@ static void		FileWatchProc _ANSI_ARGS_((ClientData instanceData,
 		            int mask));
 static void             FileThreadActionProc _ANSI_ARGS_ ((
 			   ClientData instanceData, int action));
+static DWORD		FileGetType _ANSI_ARGS_((HANDLE handle));
 
-			    
 /*
  * This structure describes the channel type structure for file based IO.
  */
@@ -476,9 +474,8 @@ FileSeekProc(instanceData, offset, mode, errorCodePtr)
     int *errorCodePtr;		/* To store error code. */
 {
     FileInfo *infoPtr = (FileInfo *) instanceData;
-    DWORD moveMethod;
-    DWORD newPos, newPosHigh;
-    DWORD oldPos, oldPosHigh;
+    DWORD moveMethod, newPos, oldPos;
+    LONG newPosHigh, oldPosHigh;
 
     *errorCodePtr = 0;
     if (mode == SEEK_SET) {
@@ -492,7 +489,7 @@ FileSeekProc(instanceData, offset, mode, errorCodePtr)
     /*
      * Save our current place in case we need to roll-back the seek.
      */
-    oldPosHigh = (DWORD)0;
+    oldPosHigh = (LONG)0;
     oldPos = SetFilePointer(infoPtr->handle, (LONG)0, &oldPosHigh,
 	    FILE_CURRENT);
     if (oldPos == INVALID_SET_FILE_POINTER) {
@@ -504,7 +501,7 @@ FileSeekProc(instanceData, offset, mode, errorCodePtr)
 	}
     }
 
-    newPosHigh = (DWORD)(offset < 0 ? -1 : 0);
+    newPosHigh = (LONG)(offset < 0 ? -1 : 0);
     newPos = SetFilePointer(infoPtr->handle, (LONG) offset, &newPosHigh,
 			    moveMethod);
     if (newPos == INVALID_SET_FILE_POINTER) {
@@ -553,8 +550,8 @@ FileWideSeekProc(instanceData, offset, mode, errorCodePtr)
     int *errorCodePtr;		/* To store error code. */
 {
     FileInfo *infoPtr = (FileInfo *) instanceData;
-    DWORD moveMethod;
-    DWORD newPos, newPosHigh;
+    DWORD moveMethod, newPos;
+    LONG newPosHigh;
 
     *errorCodePtr = 0;
     if (mode == SEEK_SET) {
@@ -576,7 +573,7 @@ FileWideSeekProc(instanceData, offset, mode, errorCodePtr)
 	    return -1;
 	}
     }
-    return ((Tcl_WideInt) newPos) | (((Tcl_WideInt) newPosHigh) << 32);
+    return (((Tcl_WideInt)((unsigned)newPos)) | (Tcl_LongAsWide(newPosHigh) << 32));
 }
 
 /*
@@ -783,10 +780,9 @@ TclpOpenFileChannel(interp, pathPtr, mode, permissions)
                                          * it? */
 {
     Tcl_Channel channel = 0;
-    int channelPermissions;
-    DWORD accessMode, createMode, shareMode, flags, consoleParams, type;
+    int channelPermissions = 0;
+    DWORD accessMode = 0, createMode, shareMode, flags;
     CONST TCHAR *nativeName;
-    DCB dcb;
     HANDLE handle;
     char channelName[16 + TCL_INTEGER_SPACE];
     TclFile readFile = NULL;
@@ -885,29 +881,9 @@ TclpOpenFileChannel(interp, pathPtr, mode, permissions)
         return NULL;
     }
     
-    type = GetFileType(handle);
-
-    /*
-     * If the file is a character device, we need to try to figure out
-     * whether it is a serial port, a console, or something else.  We
-     * test for the console case first because this is more common.
-     */
-
-    if (type == FILE_TYPE_CHAR) {
-	if (GetConsoleMode(handle, &consoleParams)) {
-	    type = FILE_TYPE_CONSOLE;
-	} else {
-	    dcb.DCBlength = sizeof( DCB ) ;
-	    if (GetCommState(handle, &dcb)) {
-		type = FILE_TYPE_SERIAL;
-	    }
-		    
-	}
-    }
-
     channel = NULL;
 
-    switch (type) {
+    switch ( FileGetType(handle) ) {
     case FILE_TYPE_SERIAL:
 	/*
 	 * Reopen channel for OVERLAPPED operation
@@ -986,15 +962,13 @@ Tcl_MakeFileChannel(rawHandle, mode)
     int mode;			/* ORed combination of TCL_READABLE and
                                  * TCL_WRITABLE to indicate file mode. */
 {
-#ifdef HAVE_NO_SEH
+#if defined(HAVE_NO_SEH) && !defined(_WIN64)
     EXCEPTION_REGISTRATION registration;
 #endif
     char channelName[16 + TCL_INTEGER_SPACE];
     Tcl_Channel channel = NULL;
     HANDLE handle = (HANDLE) rawHandle;
     HANDLE dupedHandle;
-    DCB dcb;
-    DWORD consoleParams, type;
     TclFile readFile = NULL;
     TclFile writeFile = NULL;
     BOOL result;
@@ -1003,30 +977,7 @@ Tcl_MakeFileChannel(rawHandle, mode)
 	return NULL;
     }
 
-    /*
-     * GetFileType() returns FILE_TYPE_UNKNOWN for invalid handles.
-     */
-
-    type = GetFileType(handle);
-
-    /*
-     * If the file is a character device, we need to try to figure out
-     * whether it is a serial port, a console, or something else.  We
-     * test for the console case first because this is more common.
-     */
-
-    if (type == FILE_TYPE_CHAR) {
-	if (GetConsoleMode(handle, &consoleParams)) {
-	    type = FILE_TYPE_CONSOLE;
-	} else {
-	    dcb.DCBlength = sizeof( DCB ) ;
-	    if (GetCommState(handle, &dcb)) {
-		type = FILE_TYPE_SERIAL;
-	    }
-	}
-    }
-
-    switch (type)
+    switch (FileGetType(handle))
     {
     case FILE_TYPE_SERIAL:
 	channel = TclWinOpenSerialChannel(handle, channelName, mode);
@@ -1081,12 +1032,7 @@ Tcl_MakeFileChannel(rawHandle, mode)
 	 */
 
 	result = 0;
-#ifndef HAVE_NO_SEH
-	__try {
-	    CloseHandle(dupedHandle);
-	    result = 1;
-	} __except (EXCEPTION_EXECUTE_HANDLER) {}
-#else
+#if defined(HAVE_NO_SEH) && !defined(_WIN64)
 	/*
 	 * Don't have SEH available, do things the hard way.
 	 * Note that this needs to be one block of asm, to avoid stack
@@ -1163,6 +1109,15 @@ Tcl_MakeFileChannel(rawHandle, mode)
 	    );
 	result = registration.status;
 
+#else
+#ifndef HAVE_NO_SEH
+	__try {
+#endif
+	    CloseHandle(dupedHandle);
+	    result = 1;
+#ifndef HAVE_NO_SEH
+	} __except (EXCEPTION_EXECUTE_HANDLER) {}
+#endif
 #endif
 	if (result == FALSE) {
 	    return NULL;
@@ -1204,10 +1159,10 @@ TclpGetDefaultStdChannel(type)
 {
     Tcl_Channel channel;
     HANDLE handle;
-    int mode;
-    char *bufMode;
-    DWORD handleId;		/* Standard handle to retrieve. */
-
+    int mode = -1;
+    char *bufMode = NULL;
+    DWORD handleId = (DWORD)INVALID_HANDLE_VALUE;
+				/* Standard handle to retrieve. */
 
     switch (type) {
 	case TCL_STDIN:
@@ -1430,3 +1385,50 @@ FileThreadActionProc (instanceData, action)
 	}
     }
 }
+
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * FileGetType --
+ *
+ *	Given a file handle, return its type
+ *
+ * Results:
+ *	None.
+ *
+ * Side effects:
+ *	None.
+ *
+ *----------------------------------------------------------------------
+ */
+
+DWORD
+FileGetType(handle)
+    HANDLE handle; /* Opened file handle */
+{ 
+    DWORD type;
+    DWORD consoleParams;
+    DCB dcb;
+
+    type = GetFileType(handle);
+
+    /*
+     * If the file is a character device, we need to try to figure out
+     * whether it is a serial port, a console, or something else.  We
+     * test for the console case first because this is more common.
+     */
+    
+    if (type == FILE_TYPE_CHAR || (type == FILE_TYPE_UNKNOWN && !GetLastError())) {
+	    if (GetConsoleMode(handle, &consoleParams)) {
+	      type = FILE_TYPE_CONSOLE;
+      } else {
+	      dcb.DCBlength = sizeof( DCB ) ;
+	      if (GetCommState(handle, &dcb)) {
+		      type = FILE_TYPE_SERIAL;
+        }
+      }
+    }
+
+    return type;
+}
