@@ -10,15 +10,15 @@
  *
  * See the file "license.terms" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
- *
- * RCS: @(#) $Id: tclWinFile.c,v 1.44.2.13 2005/11/15 16:41:41 kennykb Exp $
  */
 
-//#define _WIN32_WINNT  0x0500
-
+#ifndef _WIN64
+/* See [Bug 2935503]: file mtime sets wrong time */
+#   define _USE_32BIT_TIME_T
+#endif
+#include <sys/stat.h>
 #include "tclWinInt.h"
 #include <winioctl.h>
-#include <sys/stat.h>
 #include <shlobj.h>
 #include <lmaccess.h>		/* For TclpGetUserHome(). */
 
@@ -82,6 +82,9 @@
 #  define FSCTL_GET_REPARSE_POINT    CTL_CODE(FILE_DEVICE_FILE_SYSTEM, 42, METHOD_BUFFERED, FILE_ANY_ACCESS) 
 #  define FSCTL_DELETE_REPARSE_POINT CTL_CODE(FILE_DEVICE_FILE_SYSTEM, 43, METHOD_BUFFERED, FILE_SPECIAL_ACCESS) 
 #endif
+#ifndef INVALID_FILE_ATTRIBUTES
+#define INVALID_FILE_ATTRIBUTES ((DWORD)-1)
+#endif
 
 /* 
  * Maximum reparse buffer info size. The max user defined reparse
@@ -114,6 +117,7 @@ typedef struct _REPARSE_DATA_BUFFER {
             WORD   SubstituteNameLength;
             WORD   PrintNameOffset;
             WORD   PrintNameLength;
+            ULONG Flags;
             WCHAR PathBuffer[1];
         } SymbolicLinkReparseBuffer;
         struct {
@@ -169,8 +173,6 @@ typedef NET_API_STATUS NET_API_FUNCTION NETAPIBUFFERFREEPROC
 typedef NET_API_STATUS NET_API_FUNCTION NETGETDCNAMEPROC
 	(LPWSTR servername, LPWSTR domainname, LPBYTE *bufptr);
 
-extern Tcl_FSDupInternalRepProc NativeDupInternalRep;
-
 /*
  * Declarations for local procedures defined in this file:
  */
@@ -180,7 +182,7 @@ static int NativeStat(CONST TCHAR *path, Tcl_StatBuf *statPtr, int checkLinks);
 static unsigned short NativeStatMode(DWORD attr, int checkLinks, int isExec);
 static int NativeIsExec(CONST TCHAR *path);
 static int NativeReadReparse(CONST TCHAR* LinkDirectory, 
-			     REPARSE_DATA_BUFFER* buffer);
+			     REPARSE_DATA_BUFFER* buffer, DWORD desiredAccess);
 static int NativeWriteReparse(CONST TCHAR* LinkDirectory, 
 			      REPARSE_DATA_BUFFER* buffer);
 static int NativeMatchType(int isDrive, DWORD attr, CONST TCHAR* nativeName, 
@@ -222,7 +224,7 @@ WinLink(LinkSource, LinkTarget, linkAction)
 
     /* Make sure source file doesn't exist */
     attr = (*tclWinProcs->getFileAttributesProc)(LinkSource);
-    if (attr != 0xffffffff) {
+    if (attr != -1) {
 	Tcl_SetErrno(EEXIST);
 	return -1;
     }
@@ -236,7 +238,7 @@ WinLink(LinkSource, LinkTarget, linkAction)
     }
     /* Check the target */
     attr = (*tclWinProcs->getFileAttributesProc)(LinkTarget);
-    if (attr == 0xffffffff) {
+    if (attr == -1) {
 	/* The target doesn't exist */
 	TclWinConvertError(GetLastError());
 	return -1;
@@ -300,7 +302,7 @@ WinReadLink(LinkSource)
 
     /* Make sure source file does exist */
     attr = (*tclWinProcs->getFileAttributesProc)(LinkSource);
-    if (attr == 0xffffffff) {
+    if (attr == -1) {
 	/* The source doesn't exist */
 	TclWinConvertError(GetLastError());
 	return NULL;
@@ -358,18 +360,18 @@ WinSymLinkDirectory(LinkDirectory, LinkTarget)
     /* Build the reparse info */
     memset(reparseBuffer, 0, sizeof(DUMMY_REPARSE_BUFFER));
     reparseBuffer->ReparseTag = IO_REPARSE_TAG_MOUNT_POINT;
-    reparseBuffer->SymbolicLinkReparseBuffer.SubstituteNameLength = 
+    reparseBuffer->MountPointReparseBuffer.SubstituteNameLength =
       wcslen(nativeTarget) * sizeof(WCHAR);
     reparseBuffer->Reserved = 0;
-    reparseBuffer->SymbolicLinkReparseBuffer.PrintNameLength = 0;
-    reparseBuffer->SymbolicLinkReparseBuffer.PrintNameOffset = 
-      reparseBuffer->SymbolicLinkReparseBuffer.SubstituteNameLength 
+    reparseBuffer->MountPointReparseBuffer.PrintNameLength = 0;
+    reparseBuffer->MountPointReparseBuffer.PrintNameOffset =
+      reparseBuffer->MountPointReparseBuffer.SubstituteNameLength
       + sizeof(WCHAR);
-    memcpy(reparseBuffer->SymbolicLinkReparseBuffer.PathBuffer, nativeTarget, 
+    memcpy(reparseBuffer->MountPointReparseBuffer.PathBuffer, nativeTarget,
       sizeof(WCHAR) 
-      + reparseBuffer->SymbolicLinkReparseBuffer.SubstituteNameLength);
+      + reparseBuffer->MountPointReparseBuffer.SubstituteNameLength);
     reparseBuffer->ReparseDataLength = 
-      reparseBuffer->SymbolicLinkReparseBuffer.SubstituteNameLength + 12;
+      reparseBuffer->MountPointReparseBuffer.SubstituteNameLength + 12;
 	
     return NativeWriteReparse(LinkDirectory, reparseBuffer);
 }
@@ -394,7 +396,7 @@ TclWinSymLinkCopyDirectory(LinkOriginal, LinkCopy)
     DUMMY_REPARSE_BUFFER dummy;
     REPARSE_DATA_BUFFER *reparseBuffer = (REPARSE_DATA_BUFFER*)&dummy;
     
-    if (NativeReadReparse(LinkOriginal, reparseBuffer)) {
+    if (NativeReadReparse(LinkOriginal, reparseBuffer, GENERIC_READ)) {
 	return -1;
     }
     return NativeWriteReparse(LinkCopy, reparseBuffer);
@@ -478,7 +480,7 @@ WinReadLinkDirectory(LinkDirectory)
 	Tcl_SetErrno(EINVAL);
 	return NULL;
     }
-    if (NativeReadReparse(LinkDirectory, reparseBuffer)) {
+    if (NativeReadReparse(LinkDirectory, reparseBuffer, 0)) {
         return NULL;
     }
     
@@ -504,10 +506,10 @@ WinReadLinkDirectory(LinkDirectory)
 	     * that changes in the future, this code will have to be
 	     * generalised.
 	     */
-	    if (reparseBuffer->SymbolicLinkReparseBuffer.PathBuffer[0] 
+	    if (reparseBuffer->MountPointReparseBuffer.PathBuffer[0]
 		                                                 == L'\\') {
 		/* Check whether this is a mounted volume */
-		if (wcsncmp(reparseBuffer->SymbolicLinkReparseBuffer.PathBuffer, 
+		if (wcsncmp(reparseBuffer->MountPointReparseBuffer.PathBuffer,
 			    L"\\??\\Volume{",11) == 0) {
 		    char drive;
 		    /* 
@@ -515,14 +517,14 @@ WinReadLinkDirectory(LinkDirectory)
 		     * we have to fix here.  It doesn't seem very well
 		     * documented.
 		     */
-		    reparseBuffer->SymbolicLinkReparseBuffer
+		    reparseBuffer->MountPointReparseBuffer
 		                                      .PathBuffer[1] = L'\\';
 		    /* 
 		     * Check if a corresponding drive letter exists, and
 		     * use that if it is found
 		     */
 		    drive = TclWinDriveLetterForVolMountPoint(reparseBuffer
-					->SymbolicLinkReparseBuffer.PathBuffer);
+					->MountPointReparseBuffer.PathBuffer);
 		    if (drive != -1) {
 			char driveSpec[3] = {
 			    drive, ':', '\0'
@@ -543,11 +545,11 @@ WinReadLinkDirectory(LinkDirectory)
 		     */
 		    Tcl_SetErrno(EINVAL);
 		    return NULL;
-		} else if (wcsncmp(reparseBuffer->SymbolicLinkReparseBuffer
+		} else if (wcsncmp(reparseBuffer->MountPointReparseBuffer
 				   .PathBuffer, L"\\\\?\\",4) == 0) {
 		    /* Strip off the prefix */
 		    offset = 4;
-		} else if (wcsncmp(reparseBuffer->SymbolicLinkReparseBuffer
+		} else if (wcsncmp(reparseBuffer->MountPointReparseBuffer
 				   .PathBuffer, L"\\??\\",4) == 0) {
 		    /* Strip off the prefix */
 		    offset = 4;
@@ -555,8 +557,8 @@ WinReadLinkDirectory(LinkDirectory)
 	    }
 	    
 	    Tcl_WinTCharToUtf(
-		(CONST char*)reparseBuffer->SymbolicLinkReparseBuffer.PathBuffer, 
-		(int)reparseBuffer->SymbolicLinkReparseBuffer
+		(CONST char*)reparseBuffer->MountPointReparseBuffer.PathBuffer,
+		(int)reparseBuffer->MountPointReparseBuffer
 		.SubstituteNameLength, &ds);
 	
 	    copy = Tcl_DStringValue(&ds)+offset;
@@ -584,14 +586,15 @@ WinReadLinkDirectory(LinkDirectory)
  *--------------------------------------------------------------------
  */
 static int 
-NativeReadReparse(LinkDirectory, buffer)
+NativeReadReparse(LinkDirectory, buffer, desiredAccess)
     CONST TCHAR* LinkDirectory;   /* The junction to read */
     REPARSE_DATA_BUFFER* buffer;  /* Pointer to buffer. Cannot be NULL */
+    DWORD desiredAccess;
 {
     HANDLE hFile;
     DWORD returnedLength;
    
-    hFile = (*tclWinProcs->createFileProc)(LinkDirectory, GENERIC_READ, 0,
+    hFile = (*tclWinProcs->createFileProc)(LinkDirectory, desiredAccess, 0,
 	NULL, OPEN_EXISTING, 
 	FILE_FLAG_OPEN_REPARSE_POINT|FILE_FLAG_BACKUP_SEMANTICS, NULL);
     if (hFile == INVALID_HANDLE_VALUE) {
@@ -1074,7 +1077,7 @@ static int WinIsReserved(
  * volume, because for NTFS root volumes, the getFileAttributesProc
  * returns a 'hidden' attribute when it should not.
  * 
- * We never make any calss to a 'get attributes' routine here,
+ * We never make any calls to a 'get attributes' routine here,
  * since we have arranged things so that our caller already knows
  * such information.
  * 
@@ -1334,16 +1337,35 @@ NativeAccess(
 
     if (attr == 0xffffffff) {
 	/*
-	 * File doesn't exist. 
+	 * File might not exist.
 	 */
 
-	TclWinConvertError(GetLastError());
-	return -1;
+	DWORD lasterror = GetLastError();
+	if (lasterror != ERROR_SHARING_VIOLATION) {
+	    TclWinConvertError(lasterror);
+	    return -1;
+	}
     }
 
-    if ((mode & W_OK) && (attr & FILE_ATTRIBUTE_READONLY)) {
+    if (mode == F_OK) {
 	/*
-	 * File is not writable.
+	 * File exists, nothing else to check.
+	 */
+
+	return 0;
+    }
+
+    if ((mode & W_OK)
+	&& (attr & FILE_ATTRIBUTE_READONLY)
+	&& !(attr & FILE_ATTRIBUTE_DIRECTORY)) {
+	/*
+	 * The attributes say the file is not writable.	 If the file is a
+	 * regular file (i.e., not a directory), then the file is not
+	 * writable, full stop.	 For directories, the read-only bit is
+	 * (mostly) ignored by Windows, so we can't ascertain anything about
+	 * directory access from the attrib data.  However, if we have the
+	 * advanced 'getFileSecurityProc', then more robust ACL checks
+	 * will be done below.
 	 */
 
 	Tcl_SetErrno(EACCES);
@@ -1351,20 +1373,186 @@ NativeAccess(
     }
 
     if (mode & X_OK) {
-	if (attr & FILE_ATTRIBUTE_DIRECTORY) {
+	if (!(attr & FILE_ATTRIBUTE_DIRECTORY) && !NativeIsExec(nativePath)) {
 	    /*
-	     * Directories are always executable. 
+	     * It's not a directory and doesn't have the correct extension.
+	     * Therefore it can't be executable
 	     */
-	    
-	    return 0;
+
+	    Tcl_SetErrno(EACCES);
+	    return -1;
 	}
-	if (NativeIsExec(nativePath)) {
-	    return 0;
-	}
-	Tcl_SetErrno(EACCES);
-	return -1;
     }
 
+    /*
+     * It looks as if the permissions are ok, but if we are on NT, 2000 or XP,
+     * we have a more complex permissions structure so we try to check that.
+     * The code below is remarkably complex for such a simple thing as finding
+     * what permissions the OS has set for a file.
+     */
+
+    if (tclWinProcs->getFileSecurityProc != NULL) {
+	SECURITY_DESCRIPTOR *sdPtr = NULL;
+	unsigned long size;
+	PSID pSid = 0;
+	BOOL SidDefaulted;
+	SID_IDENTIFIER_AUTHORITY samba_unmapped = {{0, 0, 0, 0, 0, 22}};
+	GENERIC_MAPPING genMap;
+	HANDLE hToken = NULL;
+	DWORD desiredAccess = 0;
+	DWORD grantedAccess = 0;
+	BOOL accessYesNo = FALSE;
+	PRIVILEGE_SET privSet;
+	DWORD privSetSize = sizeof(PRIVILEGE_SET);
+	int error;
+
+	/*
+	 * First find out how big the buffer needs to be
+	 */
+
+	size = 0;
+	(*tclWinProcs->getFileSecurityProc)(nativePath,
+		OWNER_SECURITY_INFORMATION | GROUP_SECURITY_INFORMATION
+		| DACL_SECURITY_INFORMATION | LABEL_SECURITY_INFORMATION,
+		0, 0, &size);
+
+	/*
+	 * Should have failed with ERROR_INSUFFICIENT_BUFFER
+	 */
+
+	error = GetLastError();
+	if (error != ERROR_INSUFFICIENT_BUFFER) {
+	    /*
+	     * Most likely case is ERROR_ACCESS_DENIED, which we will convert
+	     * to EACCES - just what we want!
+	     */
+
+	    TclWinConvertError((DWORD)error);
+	    return -1;
+	}
+
+	/*
+	 * Now size contains the size of buffer needed
+	 */
+
+	sdPtr = (SECURITY_DESCRIPTOR *) HeapAlloc(GetProcessHeap(), 0, size);
+
+	if (sdPtr == NULL) {
+	    goto accessError;
+	}
+
+	/*
+	 * Call GetFileSecurity() for real
+	 */
+
+	if (!(*tclWinProcs->getFileSecurityProc)(nativePath,
+		OWNER_SECURITY_INFORMATION | GROUP_SECURITY_INFORMATION
+		| DACL_SECURITY_INFORMATION | LABEL_SECURITY_INFORMATION,
+		sdPtr, size, &size)) {
+	    /*
+	     * Error getting owner SD
+	     */
+
+	    goto accessError;
+	}
+
+	/*
+	 * As of Samba 3.0.23 (10-Jul-2006), unmapped users and groups are
+	 * assigned to SID domains S-1-22-1 and S-1-22-2, where "22" is the
+	 * top-level authority.	 If the file owner and group is unmapped then
+	 * the ACL access check below will only test against world access,
+	 * which is likely to be more restrictive than the actual access
+	 * restrictions.  Since the ACL tests are more likely wrong than
+	 * right, skip them.  Moreover, the unix owner access permissions are
+	 * usually mapped to the Windows attributes, so if the user is the
+	 * file owner then the attrib checks above are correct (as far as they
+	 * go).
+	 */
+
+	if(!GetSecurityDescriptorOwner(sdPtr,&pSid,&SidDefaulted) ||
+	   memcmp(GetSidIdentifierAuthority(pSid),&samba_unmapped,
+		  sizeof(SID_IDENTIFIER_AUTHORITY))==0) {
+	    HeapFree(GetProcessHeap(), 0, sdPtr);
+	    return 0; /* Attrib tests say access allowed. */
+	}
+
+	/*
+	 * Perform security impersonation of the user and open the
+	 * resulting thread token.
+	 */
+
+	if (!(*tclWinProcs->impersonateSelfProc)(SecurityImpersonation)) {
+	    /*
+	     * Unable to perform security impersonation.
+	     */
+	    
+	    goto accessError;
+	}
+	if (!(*tclWinProcs->openThreadTokenProc)(GetCurrentThread (),
+		TOKEN_DUPLICATE | TOKEN_QUERY, FALSE, &hToken)) {
+	    /*
+	     * Unable to get current thread's token.
+	     */
+	    
+	    goto accessError;
+	}
+	
+	(*tclWinProcs->revertToSelfProc)();
+	
+	/*
+	 * Setup desiredAccess according to the access priveleges we are
+	 * checking.
+	 */
+
+	if (mode & R_OK) {
+	    desiredAccess |= FILE_GENERIC_READ;
+	}
+	if (mode & W_OK) {
+	    desiredAccess |= FILE_GENERIC_WRITE;
+	}
+	if (mode & X_OK) {
+	    desiredAccess |= FILE_GENERIC_EXECUTE;
+	}
+
+	memset (&genMap, 0x0, sizeof (GENERIC_MAPPING));
+	genMap.GenericRead = FILE_GENERIC_READ;
+	genMap.GenericWrite = FILE_GENERIC_WRITE;
+	genMap.GenericExecute = FILE_GENERIC_EXECUTE;
+	genMap.GenericAll = FILE_ALL_ACCESS;
+	
+	/*
+	 * Perform access check using the token.
+	 */
+
+	if (!(*tclWinProcs->accessCheckProc)(sdPtr, hToken, desiredAccess,
+		&genMap, &privSet, &privSetSize, &grantedAccess,
+		&accessYesNo)) {
+	    /*
+	     * Unable to perform access check.
+	     */
+
+	accessError:
+	    TclWinConvertError(GetLastError());
+	    if (sdPtr != NULL) {
+		HeapFree(GetProcessHeap(), 0, sdPtr);
+	    }
+	    if (hToken != NULL) {
+		CloseHandle(hToken);
+	    }
+	    return -1;
+	}
+
+	/*
+	 * Clean up.
+	 */
+
+	HeapFree(GetProcessHeap (), 0, sdPtr);
+	CloseHandle(hToken);
+	if (!accessYesNo) {
+	    Tcl_SetErrno(EACCES);
+	    return -1;
+	}
+    }
     return 0;
 }
 
@@ -1457,24 +1645,9 @@ TclpObjChdir(pathPtr)
 {
     int result;
     CONST TCHAR *nativePath;
-#ifdef __CYGWIN__
-    extern int cygwin_conv_to_posix_path 
-	_ANSI_ARGS_((CONST char *, char *));
-    char posixPath[MAX_PATH+1];
-    CONST char *path;
-    Tcl_DString ds;
-#endif /* __CYGWIN__ */
 
     nativePath = (CONST TCHAR *) Tcl_FSGetNativePath(pathPtr);
-#ifdef __CYGWIN__
-    /* Cygwin chdir only groks POSIX path. */
-    path = Tcl_WinTCharToUtf(nativePath, -1, &ds);
-    cygwin_conv_to_posix_path(path, posixPath);
-    result = (chdir(posixPath) == 0 ? 1 : 0);
-    Tcl_DStringFree(&ds);
-#else /* __CYGWIN__ */
     result = (*tclWinProcs->setCurrentDirectoryProc)(nativePath);
-#endif /* __CYGWIN__ */
 
     if (result == 0) {
 	TclWinConvertError(GetLastError());
@@ -1483,51 +1656,6 @@ TclpObjChdir(pathPtr)
     return 0;
 }
 
-#ifdef __CYGWIN__
-/*
- *---------------------------------------------------------------------------
- *
- * TclpReadlink --
- *
- *     This function replaces the library version of readlink().
- *
- * Results:
- *     The result is a pointer to a string specifying the contents
- *     of the symbolic link given by 'path', or NULL if the symbolic
- *     link could not be read.  Storage for the result string is
- *     allocated in bufferPtr; the caller must call Tcl_DStringFree()
- *     when the result is no longer needed.
- *
- * Side effects:
- *     See readlink() documentation.
- *
- *---------------------------------------------------------------------------
- */
-
-char *
-TclpReadlink(path, linkPtr)
-    CONST char *path;          /* Path of file to readlink (UTF-8). */
-    Tcl_DString *linkPtr;      /* Uninitialized or free DString filled
-                                * with contents of link (UTF-8). */
-{
-    char link[MAXPATHLEN];
-    int length;
-    char *native;
-    Tcl_DString ds;
-
-    native = Tcl_UtfToExternalDString(NULL, path, -1, &ds);
-    length = readlink(native, link, sizeof(link));     /* INTL: Native. */
-    Tcl_DStringFree(&ds);
-    
-    if (length < 0) {
-	return NULL;
-    }
-
-    Tcl_ExternalToUtfDString(NULL, link, length, linkPtr);
-    return Tcl_DStringValue(linkPtr);
-}
-#endif /* __CYGWIN__ */
-
 /*
  *----------------------------------------------------------------------
  *
@@ -1691,7 +1819,7 @@ NativeStat(nativePath, statPtr, checkLinks)
 	     */
 
 	    attr = (*tclWinProcs->getFileAttributesProc)(nativePath);
-	    if (attr == 0xffffffff) {
+	    if (attr == INVALID_FILE_ATTRIBUTES) {
 		Tcl_SetErrno(ENOENT);
 		return -1;
 	    }
@@ -1764,8 +1892,21 @@ NativeStat(nativePath, statPtr, checkLinks)
 	if((*tclWinProcs->getFileAttributesExProc)(nativePath,
 						   GetFileExInfoStandard,
 						   &data) != TRUE) {
-	    Tcl_SetErrno(ENOENT);
-	    return -1;
+	    HANDLE hFind;
+	    WIN32_FIND_DATAT ffd;
+	    DWORD lasterror = GetLastError();
+
+	    if (lasterror != ERROR_SHARING_VIOLATION) {
+		TclWinConvertError(lasterror);
+		return -1;
+		}
+	    hFind = (*tclWinProcs->findFirstFileProc)(nativePath, &ffd);
+	    if (hFind == INVALID_HANDLE_VALUE) {
+		TclWinConvertError(GetLastError());
+		return -1;
+	    }
+	    memcpy(&data, &ffd, sizeof(data));
+	    FindClose(hFind);
 	}
 
     
@@ -1863,8 +2004,8 @@ NativeStatMode(DWORD attr, int checkLinks, int isExec)
      * other positions.
      */
 
-    mode |= (mode & 0x0700) >> 3;
-    mode |= (mode & 0x0700) >> 6;
+    mode |= (mode & (S_IREAD|S_IWRITE|S_IEXEC)) >> 3;
+    mode |= (mode & (S_IREAD|S_IWRITE|S_IEXEC)) >> 6;
     return (unsigned short)mode;
 }
 
@@ -2163,6 +2304,7 @@ TclpObjNormalizePath(interp, pathPtr, nextCheckpoint)
     Tcl_DString dsNorm;
     char *path;
     char *currentPathEndPosition;
+    Tcl_Obj *temp = NULL;
 
     Tcl_DStringInit(&dsNorm);
     path = Tcl_GetString(pathPtr);
@@ -2195,7 +2337,7 @@ TclpObjNormalizePath(interp, pathPtr, nextCheckpoint)
 		 * the current normalized path, if the file exists.
 		 */
 		if (isDrive) {
-		    if (GetFileAttributesA(nativePath) == 0xffffffff) {
+		    if (GetFileAttributesA(nativePath) == INVALID_FILE_ATTRIBUTES) {
 			/* File doesn't exist */
 			if (isDrive) {
 			    int len = WinIsReserved(path);
@@ -2209,6 +2351,12 @@ TclpObjNormalizePath(interp, pathPtr, nextCheckpoint)
 				}
 				Tcl_DStringAppend(&dsNorm, nativePath, len);
 				lastValidPathEnd = currentPathEndPosition;
+			    } else if (nextCheckpoint == 0) {
+				/* Path starts with a drive designation
+				 * that's not actually on the system.
+				 * We still must normalize up past the
+				 * first separator.  [Bug 3603434] */
+				currentPathEndPosition++;
 			    }
 			}
 			Tcl_DStringFree(&ds);
@@ -2225,7 +2373,7 @@ TclpObjNormalizePath(interp, pathPtr, nextCheckpoint)
 		    handle = FindFirstFileA(nativePath, &fData);
 		    if (handle == INVALID_HANDLE_VALUE) {
 			if (GetFileAttributesA(nativePath) 
-			    == 0xffffffff) {
+			    == INVALID_FILE_ATTRIBUTES) {
 			    /* File doesn't exist */
 			    Tcl_DStringFree(&ds);
 			    break;
@@ -2259,7 +2407,6 @@ TclpObjNormalizePath(interp, pathPtr, nextCheckpoint)
 	}
     } else {
 	/* We're on WinNT or 2000 or XP */
-	Tcl_Obj *temp = NULL;
 	int isDrive = 1;
 	Tcl_DString ds;
 
@@ -2292,6 +2439,12 @@ TclpObjNormalizePath(interp, pathPtr, nextCheckpoint)
 			    Tcl_DStringAppend(&dsNorm, nativePath,
 					      sizeof(WCHAR)*len);
 			    lastValidPathEnd = currentPathEndPosition;
+			} else if (nextCheckpoint == 0) {
+			    /* Path starts with a drive designation
+			     * that's not actually on the system.
+			     * We still must normalize up past the
+			     * first separator.  [Bug 3603434] */
+			    currentPathEndPosition++;
 			}
 		    }
 		    Tcl_DStringFree(&ds);
@@ -2448,10 +2601,19 @@ TclpObjNormalizePath(interp, pathPtr, nextCheckpoint)
 	Tcl_DStringFree(&dsTemp);
     }
     Tcl_DStringFree(&dsNorm);
+
+    /*
+     * This must be done after we are totally finished with 'path' as we are
+     * sharing the same underlying string.
+     */
+
+    if (temp != NULL) {
+	Tcl_DecrRefCount(temp);
+    }
+
     return nextCheckpoint;
 }
 
-
 /*
  *---------------------------------------------------------------------------
  *
@@ -2476,20 +2638,30 @@ TclpUtime(
 {
     int res = 0;
     HANDLE fileHandle;
+    CONST TCHAR *native;
+    DWORD attr = 0;
+    DWORD flags = FILE_ATTRIBUTE_NORMAL;
     FILETIME lastAccessTime, lastModTime;
 
     FromCTime(tval->actime, &lastAccessTime);
     FromCTime(tval->modtime, &lastModTime);
 
+    native = (CONST TCHAR *)Tcl_FSGetNativePath(pathPtr);
+
+    attr = (*tclWinProcs->getFileAttributesProc)(native);
+
+    if (attr != INVALID_FILE_ATTRIBUTES && attr & FILE_ATTRIBUTE_DIRECTORY) {
+	flags = FILE_FLAG_BACKUP_SEMANTICS;
+    }
+
     /*
      * We use the native APIs (not 'utime') because there are some daylight
      * savings complications that utime gets wrong.
      */
 
     fileHandle = (tclWinProcs->createFileProc) (
-	    (CONST TCHAR *) Tcl_FSGetNativePath(pathPtr),
-	    FILE_WRITE_ATTRIBUTES, 0, NULL, OPEN_EXISTING,
-	    FILE_ATTRIBUTE_NORMAL, NULL);
+	    native, FILE_WRITE_ATTRIBUTES, 0, NULL,
+	    OPEN_EXISTING, flags, NULL);
 
     if (fileHandle == INVALID_HANDLE_VALUE ||
 	    !SetFileTime(fileHandle, NULL, &lastAccessTime, &lastModTime)) {
