@@ -8,7 +8,41 @@
  * See the file "license.terms" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
  *
- * RCS: @(#) $Id: tclWinSock.c,v 1.36.2.2 2004/05/06 01:03:56 davygrvy Exp $
+ * -----------------------------------------------------------------------
+ *
+ * General information on how this module works.
+ *
+ * - Each Tcl-thread with its sockets maintains an internal window to receive
+ *   socket messages from the OS.
+ *
+ * - To ensure that message reception is always running this window is
+ *   actually owned and handled by an internal thread. This we call the
+ *   co-thread of Tcl's thread.
+ *
+ * - The whole structure is set up by InitSockets() which is called for each
+ *   Tcl thread. The implementation of the co-thread is in SocketThread(),
+ *   and the messages are handled by SocketProc(). The connection between
+ *   both is not directly visible, it is done through a Win32 window class.
+ *   This class is initialized by InitSockets() as well, and used in the
+ *   creation of the message receiver windows.
+ *
+ * - An important thing to note is that *both* thread and co-thread have
+ *   access to the list of sockets maintained in the private TSD data of the
+ *   thread. The co-thread was given access to it upon creation through the
+ *   new thread's client-data.
+ *
+ *   Because of this dual access the TSD data contains an OS mutex, the
+ *   "socketListLock", to mediate exclusion between thread and co-thread.
+ *
+ *   The co-thread's access is all in SocketProc(). The thread's access is
+ *   through SocketEventProc() (1) and the functions called by it.
+ *
+ *   (Ad 1) This is the handler function for all queued socket events, which
+ *          all the OS messages are translated to through the EventSource (2)
+ *          driven by the OS messages.
+ *
+ *   (Ad 2) The main functions for this are SocketSetupProc() and
+ *          SocketCheckProc().
  */
 
 #include "tclWinInt.h"
@@ -62,7 +96,7 @@ TCL_DECLARE_MUTEX(socketMutex)
             struct sockaddr FAR *name, int FAR *namelen);
     typedef int (PASCAL FAR *LPFN_GETSOCKOPT)(SOCKET s, int level,
             int optname, char FAR * optval, int FAR *optlen);
-    typedef u_short (PASCAL FAR *LPFN_HTONS)(u_short hostshort);
+    typedef unsigned short (PASCAL FAR *LPFN_HTONS)(unsigned short hostshort);
     typedef unsigned long (PASCAL FAR *LPFN_INET_ADDR)
             (const char FAR * cp);
     typedef char FAR * (PASCAL FAR *LPFN_INET_NTOA)
@@ -70,7 +104,7 @@ TCL_DECLARE_MUTEX(socketMutex)
     typedef int (PASCAL FAR *LPFN_IOCTLSOCKET)(SOCKET s,
             long cmd, u_long FAR *argp);
     typedef int (PASCAL FAR *LPFN_LISTEN)(SOCKET s, int backlog);
-    typedef u_short (PASCAL FAR *LPFN_NTOHS)(u_short netshort);
+    typedef unsigned short (PASCAL FAR *LPFN_NTOHS)(unsigned short netshort);
     typedef int (PASCAL FAR *LPFN_RECV)(SOCKET s, char FAR * buf,
             int len, int flags);
     typedef int (PASCAL FAR *LPFN_SELECT)(int nfds,
@@ -249,7 +283,6 @@ static LRESULT CALLBACK	    SocketProc _ANSI_ARGS_((HWND hwnd,
 				    UINT message, WPARAM wParam,
 				    LPARAM lParam));
 static Tcl_EventSetupProc   SocketSetupProc;
-static Tcl_ExitProc	    SocketThreadExitHandler;
 static int		    SocketsEnabled _ANSI_ARGS_((void));
 static void		    TcpAccept _ANSI_ARGS_((SocketInfo *infoPtr));
 static Tcl_DriverBlockModeProc	TcpBlockProc;
@@ -265,6 +298,10 @@ static int		    WaitForSocketEvent _ANSI_ARGS_((
 				int *errorCodePtr));
 static DWORD WINAPI	    SocketThread _ANSI_ARGS_((LPVOID arg));
 
+static void             TcpThreadActionProc _ANSI_ARGS_ ((
+			   ClientData instanceData, int action));
+
+
 /*
  * This structure describes the channel type structure for TCP socket
  * based IO.
@@ -272,7 +309,7 @@ static DWORD WINAPI	    SocketThread _ANSI_ARGS_((LPVOID arg));
 
 static Tcl_ChannelType tcpChannelType = {
     "tcp",		    /* Type name. */
-    TCL_CHANNEL_VERSION_2,  /* v2 channel */
+    TCL_CHANNEL_VERSION_4,  /* v4 channel */
     TcpCloseProc,	    /* Close proc. */
     TcpInputProc,	    /* Input proc. */
     TcpOutputProc,	    /* Output proc. */
@@ -285,6 +322,8 @@ static Tcl_ChannelType tcpChannelType = {
     TcpBlockProc,	    /* Set socket into (non-)blocking mode. */
     NULL,		    /* flush proc. */
     NULL,		    /* handler proc. */
+    NULL,                   /* wide seek proc */
+    TcpThreadActionProc,    /* thread action proc */
 };
 
 
@@ -297,7 +336,7 @@ static Tcl_ChannelType tcpChannelType = {
  *	library and set up the winSock function table.  If successful,
  *	registers the event window for the socket notifier code.
  *
- *	Assumes Mutex is held.
+ *	Assumes socketMutex is held.
  *
  * Results:
  *	None.
@@ -321,7 +360,7 @@ InitSockets()
 
     if (!initialized) {
 	initialized = 1;
-	Tcl_CreateExitHandler(SocketExitHandler, (ClientData) NULL);
+	TclCreateLateExitHandler(SocketExitHandler, (ClientData) NULL);
 
 	winSock.hModule = LoadLibraryA("wsock32.dll");
 
@@ -485,45 +524,41 @@ InitSockets()
     if (tsdPtr == NULL) {
 	tsdPtr = TCL_TSD_INIT(&dataKey);
 	tsdPtr->socketList = NULL;
-	tsdPtr->hwnd = NULL;
-
-	tsdPtr->threadId = Tcl_GetCurrentThread();
-	
+	tsdPtr->hwnd       = NULL;
+	tsdPtr->threadId   = Tcl_GetCurrentThread();
 	tsdPtr->readyEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
+	if (tsdPtr->readyEvent == NULL) {
+	    goto unloadLibrary;
+	}
 	tsdPtr->socketListLock = CreateEvent(NULL, FALSE, TRUE, NULL);
+	if (tsdPtr->socketListLock == NULL) {
+	    goto unloadLibrary;
+	}
 	tsdPtr->socketThread = CreateThread(NULL, 256, SocketThread,
 		tsdPtr, 0, &id);
-	SetThreadPriority(tsdPtr->socketThread, THREAD_PRIORITY_HIGHEST);
-
 	if (tsdPtr->socketThread == NULL) {
 	    goto unloadLibrary;
 	}
-	
+
+	SetThreadPriority(tsdPtr->socketThread, THREAD_PRIORITY_HIGHEST);
 
 	/*
-	 * Wait for the thread to signal that the window has
-	 * been created and is ready to go.  Timeout after twenty
-	 * seconds.
+	 * Wait for the thread to signal when the window has
+	 * been created and if it is ready to go.
 	 */
-	
-	if (WaitForSingleObject(tsdPtr->readyEvent, 20000)
-		== WAIT_TIMEOUT) {
-	    goto unloadLibrary;
-	}
+
+	WaitForSingleObject(tsdPtr->readyEvent, INFINITE);
 
 	if (tsdPtr->hwnd == NULL) {
-	    goto unloadLibrary;
+	    goto unloadLibrary; /* Trouble creating the window */
 	}
-	
+
 	Tcl_CreateEventSource(SocketSetupProc, SocketCheckProc, NULL);
-	Tcl_CreateThreadExitHandler(SocketThreadExitHandler, NULL);
     }
     return;
 
 unloadLibrary:
-    if (tsdPtr != NULL && tsdPtr->hwnd != NULL) {
-	SocketThreadExitHandler(0);
-    }
+    TclpFinalizeSockets();
     FreeLibrary(winSock.hModule);
     winSock.hModule = NULL;
     return;
@@ -562,7 +597,7 @@ SocketsEnabled()
  *
  * SocketExitHandler --
  *
- *	Callback invoked during exit clean up to delete the socket
+ *	Callback invoked during app exit clean up to delete the socket
  *	communication window and to release the WinSock DLL.
  *
  * Results:
@@ -585,7 +620,7 @@ SocketExitHandler(clientData)
 	 * Make sure the socket event handling window is cleaned-up
 	 * for, at most, this thread.
 	 */
-	SocketThreadExitHandler(clientData);
+	TclpFinalizeSockets();
 	UnregisterClass("TclSocket", TclWinGetTclInstance());
 	winSock.WSACleanup();
 	FreeLibrary(winSock.hModule);
@@ -599,49 +634,51 @@ SocketExitHandler(clientData)
 /*
  *----------------------------------------------------------------------
  *
- * SocketThreadExitHandler --
+ * TclpFinalizeSockets --
  *
- *	Callback invoked during thread clean up to delete the socket
- *	event source.
+ *	This function is called from Tcl_FinalizeThread to finalize
+ *	the platform specific socket subsystem.
+ *	Also, it may be called from within this module to cleanup
+ *	the state if unable to initialize the sockets subsystem.
  *
  * Results:
  *	None.
  *
  * Side effects:
- *	Delete the event source.
+ *	Deletes the event source and destroys the socket thread.
  *
  *----------------------------------------------------------------------
  */
 
-    /* ARGSUSED */
-static void
-SocketThreadExitHandler(clientData)
-    ClientData clientData;              /* Not used. */
+void
+TclpFinalizeSockets()
 {
-    ThreadSpecificData *tsdPtr = 
-	(ThreadSpecificData *)TclThreadDataKeyGet(&dataKey);
+    ThreadSpecificData *tsdPtr;
 
-    if (tsdPtr != NULL && tsdPtr->socketThread != NULL) {
-	DWORD exitCode;
-
-	GetExitCodeThread(tsdPtr->socketThread, &exitCode);
-	if (exitCode == STILL_ACTIVE) {
-	    PostMessage(tsdPtr->hwnd, SOCKET_TERMINATE, 0, 0);
-	    /*
-	     * Wait for the thread to close.  This ensures that we are
-	     * completely cleaned up before we leave this function.
-	     * If Tcl_Finalize was called from DllMain, the thread
-	     * is in a paused state so we need to timeout and continue.
-	     */
-
-	    WaitForSingleObject(tsdPtr->socketThread, 100);
+    tsdPtr = (ThreadSpecificData *)TclThreadDataKeyGet(&dataKey);
+    if (tsdPtr != NULL) {
+	if (tsdPtr->socketThread != NULL) {
+	    if (tsdPtr->hwnd != NULL) {
+		if (PostMessage(tsdPtr->hwnd, SOCKET_TERMINATE, 0, 0)) {
+		    /*
+		     * Wait for the thread to exit. This ensures that we are
+		     * completely cleaned up before we leave this function.
+		     */
+		    WaitForSingleObject(tsdPtr->readyEvent, INFINITE);
+		}
+		tsdPtr->hwnd = NULL;
+	    }
+	    CloseHandle(tsdPtr->socketThread);
+	    tsdPtr->socketThread = NULL;
+	}
+	if (tsdPtr->readyEvent != NULL) {
+	    CloseHandle(tsdPtr->readyEvent);
+	    tsdPtr->readyEvent = NULL;
+	}
+	if (tsdPtr->socketListLock != NULL) {
+	    CloseHandle(tsdPtr->socketListLock);
+	    tsdPtr->socketListLock = NULL;
 	}
-	CloseHandle(tsdPtr->socketThread);
-	tsdPtr->socketThread = NULL;
-	CloseHandle(tsdPtr->readyEvent);
-	CloseHandle(tsdPtr->socketListLock);
-
-	Tcl_DeleteThreadExitHandler(SocketThreadExitHandler, NULL);
 	Tcl_DeleteEventSource(SocketSetupProc, SocketCheckProc, NULL);
     }
 }
@@ -970,9 +1007,8 @@ TcpCloseProc(instanceData, interp)
     Tcl_Interp *interp;		/* Unused. */
 {
     SocketInfo *infoPtr = (SocketInfo *) instanceData;
-    SocketInfo **nextPtrPtr;
+    /* TIP #218 */
     int errorCode = 0;
-    ThreadSpecificData *tsdPtr = TCL_TSD_INIT(&dataKey);
 
     /*
      * Check that WinSock is initialized; do not call it if not, to
@@ -995,20 +1031,12 @@ TcpCloseProc(instanceData, interp)
         }
     }
 
-    /*
-     * Remove the socket from socketList.
+    /* TIP #218. Removed the code removing the structure
+     * from the global socket list. This is now done by
+     * the thread action callbacks, and only there. This
+     * happens before this code is called. We can free
+     * without fear of damanging the list.
      */
-
-    WaitForSingleObject(tsdPtr->socketListLock, INFINITE);
-    for (nextPtrPtr = &(tsdPtr->socketList); (*nextPtrPtr) != NULL;
-	 nextPtrPtr = &((*nextPtrPtr)->nextPtr)) {
-	if ((*nextPtrPtr) == infoPtr) {
-	    (*nextPtrPtr) = infoPtr->nextPtr;
-	    break;
-	}
-    }
-    
-    SetEvent(tsdPtr->socketListLock);
     ckfree((char *) infoPtr);
     return errorCode;
 }
@@ -1025,7 +1053,7 @@ TcpCloseProc(instanceData, interp)
  *	Returns a newly allocated SocketInfo.
  *
  * Side effects:
- *	Adds the socket to the global socket list.
+ *	None, except for allocation of memory.
  *
  *----------------------------------------------------------------------
  */
@@ -1035,7 +1063,6 @@ NewSocketInfo(socket)
     SOCKET socket;
 {
     SocketInfo *infoPtr;
-    ThreadSpecificData *tsdPtr = TCL_TSD_INIT(&dataKey);
 
     infoPtr = (SocketInfo *) ckalloc((unsigned) sizeof(SocketInfo));
     infoPtr->socket = socket;
@@ -1047,10 +1074,11 @@ NewSocketInfo(socket)
     infoPtr->acceptProc = NULL;
     infoPtr->lastError = 0;
 
-    WaitForSingleObject(tsdPtr->socketListLock, INFINITE);
-    infoPtr->nextPtr = tsdPtr->socketList;
-    tsdPtr->socketList = infoPtr;
-    SetEvent(tsdPtr->socketListLock);
+    /* TIP #218. Removed the code inserting the new structure
+     * into the global list. This is now handled in the thread
+     * action callbacks, and only there.
+     */
+    infoPtr->nextPtr = NULL;
     
     return infoPtr;
 }
@@ -1067,7 +1095,7 @@ NewSocketInfo(socket)
  *	Returns a new SocketInfo, or NULL with an error in interp.
  *
  * Side effects:
- *	Adds a new socket to the socketList.
+ *	None, except for allocation of memory.
  *
  *----------------------------------------------------------------------
  */
@@ -1089,7 +1117,7 @@ CreateSocket(interp, port, host, server, myaddr, myport, async)
 				 * in progress. */
     SOCKADDR_IN sockaddr;	/* Socket address */
     SOCKADDR_IN mysockaddr;	/* Socket address for client */
-    SOCKET sock;
+    SOCKET sock = INVALID_SOCKET;
     SocketInfo *infoPtr;	/* The returned value. */
     ThreadSpecificData *tsdPtr = 
 	(ThreadSpecificData *)TclThreadDataKeyGet(&dataKey);
@@ -1129,7 +1157,7 @@ CreateSocket(interp, port, host, server, myaddr, myport, async)
      * Set kernel space buffering
      */
 
-    TclSockMinimumBuffers((int) sock, TCP_BUFFER_SIZE);
+    TclSockMinimumBuffers((void *)sock, TCP_BUFFER_SIZE);
 
     if (server) {
 	/*
@@ -1293,7 +1321,7 @@ CreateSocketAddress(sockaddrPtr, host, port)
 
     ZeroMemory(sockaddrPtr, sizeof(SOCKADDR_IN));
     sockaddrPtr->sin_family = AF_INET;
-    sockaddrPtr->sin_port = winSock.htons((u_short) (port & 0xFFFF));
+    sockaddrPtr->sin_port = winSock.htons((unsigned short) (port & 0xFFFF));
     if (host == NULL) {
 	addr.s_addr = INADDR_ANY;
     } else {
@@ -1490,7 +1518,7 @@ Tcl_MakeTcpClientChannel(sock)
      * Set kernel space buffering and non-blocking.
      */
 
-    TclSockMinimumBuffers((int) sock, TCP_BUFFER_SIZE);
+    TclSockMinimumBuffers(sock, TCP_BUFFER_SIZE);
 
     infoPtr = NewSocketInfo((SOCKET) sock);
 
@@ -1606,6 +1634,12 @@ TcpAccept(infoPtr)
     newSocket = winSock.accept(infoPtr->socket, (SOCKADDR *)&addr,
 	    &len);
 
+    /*
+     * Protect access to sockets (acceptEventCount, readyEvents) in socketList
+     * by the lock.  Fix for SF Tcl Bug 3056775.
+     */
+    WaitForSingleObject(tsdPtr->socketListLock, INFINITE);
+
     /*
      * Clear the ready mask so we can detect the next connection request.
      * Note that connection requests are level triggered, so if there is
@@ -1615,6 +1649,8 @@ TcpAccept(infoPtr)
     if (newSocket == INVALID_SOCKET) {
 	infoPtr->acceptEventCount = 0;
 	infoPtr->readyEvents &= ~(FD_ACCEPT);
+
+	SetEvent(tsdPtr->socketListLock);
 	return;
     }
 
@@ -1630,6 +1666,8 @@ TcpAccept(infoPtr)
 	infoPtr->readyEvents &= ~(FD_ACCEPT);
     }
 
+    SetEvent(tsdPtr->socketListLock);
+
     /*
      * Win-NT has a misfeature that sockets are inherited in child
      * processes by default.  Turn off the inherit bit.
@@ -1775,11 +1813,23 @@ TcpInputProc(instanceData, buf, toRead, errorCodePtr)
 	    break;
 	}
   
+ 	error = winSock.WSAGetLastError();
+ 
+ 	/*
+ 	 * If an RST comes, then ignore the error and report an EOF just like
+ 	 * on unix.
+ 	 */
+ 
+ 	if (error == WSAECONNRESET) {
+ 	    infoPtr->flags |= SOCKET_EOF;
+ 	    bytesRead = 0;
+ 	    break;
+ 	}
+ 
 	/*
 	 * Check for error condition or underflow in non-blocking case.
 	 */
   
-	error = winSock.WSAGetLastError();
 	if ((infoPtr->flags & SOCKET_ASYNC) || (error != WSAEWOULDBLOCK)) {
 	    TclWinConvertWSAError(error);
 	    *errorCodePtr = Tcl_GetErrno();
@@ -1938,9 +1988,9 @@ TcpSetOptionProc (
     CONST char *optionName,	/* Name of the option to set. */
     CONST char *value)		/* New value for option. */
 {
+/*
     SocketInfo *infoPtr;
     SOCKET sock;
-/*
     BOOL val = FALSE;
     int boolVar, rtn;
 */
@@ -1958,10 +2008,10 @@ TcpSetOptionProc (
         return TCL_ERROR;
     }
 
+/*
     infoPtr = (SocketInfo *) instanceData;
     sock = infoPtr->socket;
 
-/*
     if (!stricmp(optionName, "-keepalive")) {
 	if (Tcl_GetBoolean(interp, value, &boolVar) != TCL_OK) {
 	    return TCL_ERROR;
@@ -2325,29 +2375,44 @@ SocketThread(LPVOID arg)
     MSG msg;
     ThreadSpecificData *tsdPtr = (ThreadSpecificData *)(arg);
 
+    /*
+     * Create a dummy window receiving socket events.
+     */
+
     tsdPtr->hwnd = CreateWindow("TclSocket", "TclSocket", 
 	    WS_TILED, 0, 0, 0, 0, NULL, NULL, windowClass.hInstance, arg);
 
     /*
-     * Signal the main thread that the window has been created
-     * and that the socket thread is ready to go.
+     * Signalize thread creator that we are done creating the window.
      */
-    
+
     SetEvent(tsdPtr->readyEvent);
-    
+
+    /*
+     * If unable to create the window, exit this thread immediately.
+     */
+
     if (tsdPtr->hwnd == NULL) {
 	return 1;
     }
 
     /*
      * Process all messages on the socket window until WM_QUIT.
+     * This threads exits only when instructed to do so by the
+     * call to PostMessage(SOCKET_TERMINATE) in TclpFinalizeSockets().
      */
 
     while (GetMessage(&msg, NULL, 0, 0) > 0) {
 	DispatchMessage(&msg);
     }
 
-    return msg.wParam;
+    /*
+     * This releases waiters on thread exit in TclpFinalizeSockets()
+     */
+
+    SetEvent(tsdPtr->readyEvent);
+
+    return (DWORD)msg.wParam;
 }
 
 
@@ -2534,41 +2599,28 @@ Tcl_GetHostName()
     Tcl_MutexLock(&socketMutex);
     InitSockets();
 
-    if (hostnameInitialized) {
-	Tcl_MutexUnlock(&socketMutex);
-        return hostname;
-    }
-    Tcl_MutexUnlock(&socketMutex);
-	
-    if (TclpHasSockets(NULL) == TCL_OK) {
+    if (!hostnameInitialized) {
 	/*
-	 * INTL: bug
+	 * Convert hostname from native to UTF then change to lowercase.
 	 */
-
-	if (winSock.gethostname(hostname, sizeof(hostname)) == 0) {
-	    Tcl_MutexLock(&socketMutex);
-	    hostnameInitialized = 1;
-	    Tcl_MutexUnlock(&socketMutex);
-	    return hostname;
-	}
-    }
-    Tcl_MutexLock(&socketMutex);
-    length = sizeof(hostname);
-    if ((*tclWinProcs->getComputerNameProc)(wbuf, &length) != 0) {
-	/*
-	 * Convert string from native to UTF then change to lowercase.
-	 */
-
 	Tcl_DString ds;
 
-	lstrcpynA(hostname, Tcl_WinTCharToUtf((TCHAR *) wbuf, -1, &ds),
-		sizeof(hostname));
+	length = sizeof(hostname);
+	/* same as SocketsEnabled without the socketMutex lock */
+	if ((winSock.hModule != NULL)
+		&& (winSock.gethostname(hostname, length) == 0)) {
+	    Tcl_ExternalToUtfDString(NULL, hostname, -1, &ds);
+	} else if ((*tclWinProcs->getComputerNameProc)(wbuf, &length) != 0) {
+	    Tcl_WinTCharToUtf((TCHAR *) wbuf, -1, &ds);
+	} else {
+	    Tcl_DStringInit(&ds);
+	    Tcl_DStringSetLength(&ds, 0);
+	}
+	lstrcpynA(hostname, Tcl_DStringValue(&ds), sizeof(hostname));
 	Tcl_DStringFree(&ds);
 	Tcl_UtfToLower(hostname);
-    } else {
-	hostname[0] = '\0';
+	hostnameInitialized = 1;
     }
-    hostnameInitialized = 1;
     Tcl_MutexUnlock(&socketMutex);
     return hostname;
 }
@@ -2594,7 +2646,7 @@ Tcl_GetHostName()
 
 int
 TclWinGetSockOpt(SOCKET s, int level, int optname, char * optval,
-	int FAR *optlen)
+	int *optlen)
 {
     /*
      * Check that WinSock is initialized; do not call it if not, to
@@ -2627,8 +2679,8 @@ TclWinSetSockOpt(SOCKET s, int level, int optname, const char * optval,
     return winSock.setsockopt(s, level, optname, optval, optlen);
 }
 
-u_short
-TclWinNToHS(u_short netshort)
+unsigned short
+TclWinNToHS(unsigned short netshort)
 {
     /*
      * Check that WinSock is initialized; do not call it if not, to
@@ -2638,12 +2690,29 @@ TclWinNToHS(u_short netshort)
      */
 
     if (!SocketsEnabled()) {
-        return (u_short) -1;
+        return (unsigned short) -1;
     }
 
     return winSock.ntohs(netshort);
 }
 
+char *
+TclpInetNtoa(struct in_addr addr)
+{
+    /*
+     * Check that WinSock is initialized; do not call it if not, to
+     * prevent system crashes. This can happen at exit time if the exit
+     * handler for WinSock ran before other exit handlers that want to
+     * use sockets.
+     */
+
+    if (!SocketsEnabled()) {
+        return NULL;
+    }
+
+    return winSock.inet_ntoa(addr);
+}
+
 struct servent *
 TclWinGetServByName(const char * name, const char * proto)
 {
@@ -2659,5 +2728,85 @@ TclWinGetServByName(const char * name, const char * proto)
 
     return winSock.getservbyname(name, proto);
 }
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * TcpThreadActionProc --
+ *
+ *	Insert or remove any thread local refs to this channel.
+ *
+ * Results:
+ *	None.
+ *
+ * Side effects:
+ *	Changes thread local list of valid channels.
+ *
+ *----------------------------------------------------------------------
+ */
 
+static void
+TcpThreadActionProc (instanceData, action)
+     ClientData instanceData;
+     int action;
+{
+    ThreadSpecificData *tsdPtr;
+    SocketInfo *infoPtr = (SocketInfo *) instanceData;
+    int      notifyCmd;
 
+    if (action == TCL_CHANNEL_THREAD_INSERT) {
+        /*
+	 * Ensure that socket subsystem is initialized in this thread, or
+	 * else sockets will not work.
+	 */
+
+        Tcl_MutexLock(&socketMutex);
+	InitSockets();
+	Tcl_MutexUnlock(&socketMutex);
+
+	tsdPtr = TCL_TSD_INIT(&dataKey);
+
+	WaitForSingleObject(tsdPtr->socketListLock, INFINITE);
+	infoPtr->nextPtr = tsdPtr->socketList;
+	tsdPtr->socketList = infoPtr;
+	SetEvent(tsdPtr->socketListLock);
+
+	notifyCmd = SELECT;
+    } else {
+	SocketInfo **nextPtrPtr;
+	int removed = 0;
+
+	tsdPtr  = TCL_TSD_INIT(&dataKey);
+
+	/* TIP #218, Bugfix: All access to socketList has to be protected by the lock */
+	WaitForSingleObject(tsdPtr->socketListLock, INFINITE);
+	for (nextPtrPtr = &(tsdPtr->socketList); (*nextPtrPtr) != NULL;
+	     nextPtrPtr = &((*nextPtrPtr)->nextPtr)) {
+	    if ((*nextPtrPtr) == infoPtr) {
+	        (*nextPtrPtr) = infoPtr->nextPtr;
+		removed = 1;
+		break;
+	    }
+	}
+	SetEvent(tsdPtr->socketListLock);
+
+	/*
+	 * This could happen if the channel was created in one thread
+	 * and then moved to another without updating the thread
+	 * local data in each thread.
+	 */
+
+	if (!removed) {
+	    Tcl_Panic("file info ptr not on thread channel list");
+	}
+
+	notifyCmd = UNSELECT;
+    }
+
+    /*
+     * Ensure that, or stop, notifications for the socket occur in this thread.
+     */
+
+    SendMessage(tsdPtr->hwnd, SOCKET_SELECT,
+		(WPARAM) notifyCmd, (LPARAM) infoPtr);
+}
