@@ -8,10 +8,11 @@
  *
  * See the file "license.terms" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
- *
- * RCS: @(#) $Id: tclWinTime.c,v 1.14.2.5 2004/06/05 17:25:40 kennykb Exp $
  */
 
+#if defined(_WIN32) && !defined(_WIN64)
+#   define _USE_32BIT_TIME_T
+#endif
 #include "tclWinInt.h"
 
 #define SECSPERDAY (60L * 60L * 24L)
@@ -28,11 +29,11 @@
  * each month, where index 1 is January.
  */
 
-static int normalDays[] = {
+static const int normalDays[] = {
     -1, 30, 58, 89, 119, 150, 180, 211, 242, 272, 303, 333, 364
 };
 
-static int leapDays[] = {
+static const int leapDays[] = {
     -1, 30, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334, 365
 };
 
@@ -121,7 +122,7 @@ static TimeInfo timeInfo = {
     0
 };
 
-CONST static FILETIME posixEpoch = { 0xD53E8000, 0x019DB1DE };
+static CONST FILETIME posixEpoch = { 0xD53E8000, 0x019DB1DE };
     
 /*
  * Declarations for functions defined later in this file.
@@ -223,7 +224,7 @@ TclpGetClicks()
 
 int
 TclpGetTimeZone (currentTime)
-    unsigned long  currentTime;
+    Tcl_WideInt currentTime;
 {
     int timeZone;
 
@@ -260,8 +261,7 @@ void
 Tcl_GetTime(timePtr)
     Tcl_Time *timePtr;		/* Location to store time information. */
 {
-	
-    struct timeb t;
+    struct _timeb t;
 
     int useFtime = 1;		/* Flag == TRUE if we need to fall back
 				 * on ftime rather than using the perf
@@ -379,7 +379,6 @@ Tcl_GetTime(timePtr)
     }
 
     if ( timeInfo.perfCounterAvailable ) {
-	
 	/*
 	 * Query the performance counter and use it to calculate the
 	 * current time.
@@ -417,27 +416,26 @@ Tcl_GetTime(timePtr)
 	 */
 	if ( curCounter.QuadPart - timeInfo.perfCounterLastCall.QuadPart
 	     < 11 * timeInfo.curCounterFreq.QuadPart / 10 ) {
-	    
+
 	    curFileTime = timeInfo.fileTimeLastCall.QuadPart
 		+ ( ( curCounter.QuadPart - timeInfo.perfCounterLastCall.QuadPart )
 		    * 10000000 / timeInfo.curCounterFreq.QuadPart );
 	    timeInfo.fileTimeLastCall.QuadPart = curFileTime;
 	    timeInfo.perfCounterLastCall.QuadPart = curCounter.QuadPart;
 	    usecSincePosixEpoch = ( curFileTime - posixEpoch.QuadPart ) / 10;
-	    timePtr->sec = (time_t) ( usecSincePosixEpoch / 1000000 );
+	    timePtr->sec = (long) ( usecSincePosixEpoch / 1000000 );
 	    timePtr->usec = (unsigned long ) ( usecSincePosixEpoch % 1000000 );
 	    useFtime = 0;
 	}
 
 	LeaveCriticalSection( &timeInfo.cs );
     }
-	
+
     if ( useFtime ) {
-	
 	/* High resolution timer is not available.  Just use ftime */
 
-	ftime(&t);
-	timePtr->sec = t.time;
+	_ftime(&t);
+	timePtr->sec = (long)t.time;
 	timePtr->usec = t.millitm * 1000;
     }
 }
@@ -489,7 +487,7 @@ StopCalibration( ClientData unused )
 char *
 TclpGetTZName(int dst)
 {
-    int len;
+    size_t len;
     char *zone, *p;
     TIME_ZONE_INFORMATION tz;
     Tcl_Encoding encoding;
@@ -536,7 +534,7 @@ TclpGetTZName(int dst)
 		}
 	    }
 	}
-	Tcl_ExternalToUtf(NULL, NULL, zone, len, 0, NULL, name,
+	Tcl_ExternalToUtf(NULL, NULL, zone, (int)len, 0, NULL, name,
 		sizeof(tsdPtr->tzName), NULL, NULL, NULL);
     }
     if (name[0] == '\0') {
@@ -581,7 +579,7 @@ TclpGetDate(t, useGMT)
 {
     const time_t *tp = (const time_t *) t;
     struct tm *tmPtr;
-    long time;
+    time_t time;
 
     if (!useGMT) {
 	tzset();
@@ -597,7 +595,7 @@ TclpGetDate(t, useGMT)
 	}
 
 	time = *tp - _timezone;
-	
+
 	/*
 	 * If we aren't near to overflowing the long, just add the bias and
 	 * use the normal calculation.  Otherwise we will need to adjust
@@ -623,7 +621,7 @@ TclpGetDate(t, useGMT)
 		tmPtr->tm_sec += 60;
 		time -= 60;
 	    }
-    
+
 	    time = tmPtr->tm_min + time/60;
 	    tmPtr->tm_min = (int)(time % 60);
 	    if (tmPtr->tm_min < 0) {
@@ -639,9 +637,9 @@ TclpGetDate(t, useGMT)
 	    }
 
 	    time /= 24;
-	    tmPtr->tm_mday += time;
-	    tmPtr->tm_yday += time;
-	    tmPtr->tm_wday = (tmPtr->tm_wday + time) % 7;
+	    tmPtr->tm_mday += (int)time;
+	    tmPtr->tm_yday += (int)time;
+	    tmPtr->tm_wday = (tmPtr->tm_wday + (int)time) % 7;
 	}
     } else {
 	tmPtr = ComputeGMT(tp);
@@ -673,7 +671,7 @@ ComputeGMT(tp)
     struct tm *tmPtr;
     long tmp, rem;
     int isLeap;
-    int *days;
+    const int *days;
     ThreadSpecificData *tsdPtr = TCL_TSD_INIT(&dataKey);
 
     tmPtr = &tsdPtr->tm;
@@ -682,8 +680,8 @@ ComputeGMT(tp)
      * Compute the 4 year span containing the specified time.
      */
 
-    tmp = *tp / SECSPER4YEAR;
-    rem = *tp % SECSPER4YEAR;
+    tmp = (long)(*tp / SECSPER4YEAR);
+    rem = (LONG)(*tp % SECSPER4YEAR);
 
     /*
      * Correct for weird mod semantics so the remainder is always positive.
@@ -749,7 +747,7 @@ ComputeGMT(tp)
      * Compute day of week.  Epoch started on a Thursday.
      */
 
-    tmPtr->tm_wday = (*tp / SECSPERDAY) + 4;
+    tmPtr->tm_wday = (long)(*tp / SECSPERDAY) + 4;
     if ((*tp % SECSPERDAY) < 0) {
 	tmPtr->tm_wday--;
     }
@@ -1103,7 +1101,7 @@ AccumulateSample( Tcl_WideInt perfCounter,
 
 struct tm *
 TclpGmtime( tt )
-    TclpTime_t tt;
+    TclpTime_t_CONST tt;
 {
     CONST time_t *timePtr = (CONST time_t *) tt;
 				/* Pointer to the number of seconds
@@ -1135,7 +1133,7 @@ TclpGmtime( tt )
 
 struct tm *
 TclpLocaltime( tt )
-    TclpTime_t tt;
+    TclpTime_t_CONST tt;
 {
     CONST time_t *timePtr = (CONST time_t *) tt;
 				/* Pointer to the number of seconds
